function [remaining,notfound] = set_chanid(data,lookup)
% [Indices,Not-Found] = set_chanid(Dataset/Chanlocs,Lookup)
% Translate channel names into (valid) indices in a dataset's channels, or in a chanlocs structure.
%
% In:
%   Dataset/Chanlocs :   EEG dataset, or EEGLAB chanlocs structure, or cell array of channel labels
%
%   Lookup           :   cell-string array of channel names to look up, or numeric indices
%
% Out:
%   Indices          :   The indices of the specified channel names, in the order of appearance in
%                        the supplied dataset/chanlocs.
%
%   Not-Found        :   indices, into lookup, of channels that were not found in the dataset
%
% Note:
%  Differs from eeg_chaninds in that it does not abort when some of the channels are not present.
%  Also, the channel indices are ordered as they appear in Lookup.
%
%                       Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                       2010-05-08

% get the channel labels from the data
data = exp_eval(data);
if isfield(data,'chanlocs') && isfield(data.chanlocs,'labels')
    labels = {data.chanlocs.labels};
elseif isfield(data,'labels')
    labels = {data.labels};
elseif iscell(data)
    labels = data;
else
    error('chanlocs data structure or EEGLAB data expected');
end

if iscell(lookup)
	% got channel names: look them up from the chanlocs, but ordered according to lookup
    [x,a,b] = intersect(lookup,labels); %#ok<ASGLU>
    [x,I] = sort(a); remaining = b(I); %#ok<ASGLU>
    notfound = setdiff(1:length(lookup),a);
elseif isnumeric(lookup)
	% got channel indices: filter out the invalid channels
    if isempty(lookup)
        lookup = 1:size(data.data,1); end;
	remaining = lookup(1:min(length(lookup),size(data.data,1)));
    notfound = lookup((1+size(data.data,1)):end);
end