function eeg = set_derive_target(eeg, eventtypes, default)
% Result-Set = set_derive_target(Input-Set, Event-Types, Default-Value)
% (Re-)derives a categorical target field for an epoched EEG set from events, 
% based on a given set of time-locking event types.
%
% An epoched dataset which is lacking a target variable (e.g., created with EEGLAB's pop_epoch), 
% but whose time-locking event types are indicative of the target values, can be transformed into
% a dataset which has a proper target variable, using this function. 
%
% In:
%   Input-Set     : epoched EEG set
%   Event-Types   : types of time-locking events that determine target group
%   Default-Value : default target value, if no matching time-locking event type was found (default: 0)
%
% Out:
%   Result-Set  : epoched EEG set with epoch.target set so that the target
%                 equals the index among the eventtypes
%
% Notes:
%   This function can not derive the target value when the epoch's time range (xmin-xmax) does not cross time zero,
%   since then, the epochs do not contain time-locking events.
%
%                       Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                       2010-04-01

if ~exp_begindef('editing') return; end
if ~exist('default','var')
    default = 0; end

% get time-locking event indices
events = find(abs(eeg_point2lat([eeg.event.latency],[eeg.event.epoch],eeg.srate,[eeg.xmin eeg.xmax])) < 0.0005);
% ... their corresponding epoch indices
epochs = [eeg.event(events).epoch];
% ... and their types
types = {eeg.event(events).type};

eeg.epoch(1).target = default;
if ~isequal(length(eeg.epoch), length(events), size(eeg.data,3))
    error('Not the same number of time-locking events, epochs and trials.'); end;

% init the target variable (one value per epoch) to the default value (using cell arrays to later assign it to a struct field)
target = repmat({default},1,length(epochs));
% for every specified eventtype, assign target value t to all entries that match eventtype #t
for t=1:length(eventtypes)
    target(strcmp(types,eventtypes{t})) = {t}; end
% write into the struct
[eeg.epoch.target] = target{:};

exp_enddef;