function [eeg,sample_mask] = flt_clean_windows(eeg,flag_quantile,window_len,ignored_chans, keep_metadata)
% [Out-Set,Mask] = flt_clean_windows(In-Set,Flagged-Quantile,Window-Length,Max-Ignoredchans)
% remove periods of abnormal (extreme magnitude) data from a (sufficiently high-passed) continuous dataset. Non-causal, not applied online.
%
% This is an autmated artifact rejection function which cuts segments of artifacts (characterized by their signal power) from the data.
%
% In:
%   In-Set           : continuous data dataset, assumed to be appropriately high-passed (e.g. >0.5Hz or 0.5Hz - 2.0Hz transition band)
%
%   Flagged-Quantile : upper quantile of the per-channel windows that should be flagged for potential removal (removed if flagged in all except 
%                      for some possibly bad channels); controls the aggressiveness of the rejection; if two numbers are specified,
%                      the first is the lower quantile and the second is the upper quantile to be removed (default: 0.1)
%
%   Window-Length    : length of the windows (in seconds) which are inspected for artifact content; ideally as long as the expected
%                      time scale of the artifacts (e.g. chewing) (default: 1)
% 
%   Max-Ignoredchans : maximum number of ignored channels in the data (these can deliver arbitrary data without affecting the outcome)
%                      may also be a ratio (default: 0.3)
%
%`  keep_metadata   : boolean; whether meta data of EEG struct (such as events, ICA decomposition etc.) should be returned. If true, 
%                      meta data is returned. Returning meta data is quite slow. (default: false)
%
% Out:
%   Out-Set : EEG dataset with bad time periods (and all events) removed, if keep_metadata is false.
%
%   Mask    : mask of retained samples (logical array)
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-07-06

sample_mask = [];
if ~exp_begindef('editing') return; end;
%opts = hlp_varargin2struct(varargin, 'window_len',1, 'flag_quantile',0.15, 'ignored_chans',0.3, 'keep_metadata',false);
if ~exist('window_len','var') || isempty(window_len) window_len = 1; end
if ~exist('flag_quantile','var') || isempty(flag_quantile) flag_quantile = 0.15; end
if ~exist('ignored_chans','var') || isempty(ignored_chans) ignored_chans = 0.3; end
if ~exist('keep_metadata','var') keep_metadata = false; end

if ignored_chans > 0 && ignored_chans < 1
    ignored_chans = size(eeg.data,1)*ignored_chans; end
if isscalar(flag_quantile)
    flag_quantile = [0 flag_quantile]; end

[C,S] = size(eeg.data);
window_len = window_len*eeg.srate;
wnd = 0:window_len-1;
offsets = 1:window_len/2:S-window_len;
W = length(offsets);1:98

wpwr = zeros(C,W);
% for each channel
for c = 1:C
    % for each window
    for o=1:W
        % extract data
        x = eeg.data(c,offsets(o)+wnd);
        % compute windowed power (measures both mean deviations, i.e. jumps, and large oscillations)
        wpwr(c,o) = sqrt(sum(x.^2)/window_len);
    end
end

[dummy,i] = sort(wpwr'); %#ok<TRSRT,ASGLU>

% find retained window indices per channel
retained_quantiles = i(1+floor(W*flag_quantile(1)):round(W*(1-flag_quantile(2))),:)';

% flag them in a Channels x Windows matrix (this can be neatly visualized)
window_mask = zeros(C,W);
for c = 1:C
    window_mask(c,retained_quantiles(c,:)) = 1; end

% find retained windows
retained_windows = find(sum(window_mask) > ignored_chans);
% find retained samples
retained_samples = repmat(offsets(retained_windows)',1,length(wnd))+repmat(wnd,length(retained_windows),1);
% mask them out
sample_mask = false(1,S); sample_mask(retained_samples(:)) = true;
fprintf('Removing %.1f%% (%.0f seconds) of the data.\n',100*(1-mean(sample_mask)),nnz(~sample_mask)/eeg.srate);

if(keep_metadata)
    %{
    sample_mask_pushed_left = NaN(1, S); sample_mask_pushed_left(1:(S-1)) = sample_mask(2:S);
    sample_mask_pushed_right = NaN(1, S); sample_mask_pushed_right(2:S) = sample_mask(1:(S-1));

    % 1 in the sample_mask indicates retention, so we find the indices at which a 1 is encountered after a 0.
    min_point_values = find(sample_mask > sample_mask_pushed_right); 
    max_point_values = find(sample_mask > sample_mask_pushed_left);

    % take care of extreme cases where samples at either extreme end of data should be kept
    if(sample_mask(1) == 1); min_point_values = [1 min_point_values]; end
    if(sample_mask(S) == 1); max_point_values = [max_point_values S]; end

    retain_data_intervals = [min_point_values; max_point_values]';
    %}
    retain_data_intervals = reshape(find(diff([false sample_mask false])),2,[])';
    retain_data_intervals(:,2) = retain_data_intervals(:,2)-1;  

    eeg = pop_select(eeg, 'point', retain_data_intervals);
else
    
    % retain the masked data, clear all events or other aggregated data
    eeg = exp_eval(set_new(eeg,'data',eeg.data(:,sample_mask),'icaact',[],'event',eeg.event([]),'urevent',eeg.urevent([]), ...
        'epoch',[],'reject',[],'stats',[],'specdata',[],'specicaact',[]));
end



    
exp_enddef;