function data = flt_default(data, varargin)
% Out-Data = flt_default(In-Data, Options...)
% The default configurable preprocessing pipeline for most BCI paradigms.
%
% Most BCI paradigms contain a sequence of signal processing steps, many of which are shared among paradigms (for example, spatial and spectral filtering).
% The default Signal Processing pipeline allows to restrict the raw sensor signals to the contents of interest, such as, for example, specific frequency bands 
% (e.g., the alpha band), or specific spatial areas (e.g. signals emitted from the motor cortex). This can be viewed as a way to improve the signal-to-noise 
% ratio (for some notion of what is the signal) of the data, or as a way to express and incorporate prior knowledge into the paradigm.
%
% With no options, flt_default returns the input signal unprocessed. By giving options, various stages can be enabled (their relative order is pre-determined).
% Most paradigms (the paradigms/para_*) functions have their own defaults for the options in flt_default, e.g. 'srate',200, 'epoch',[0 3], but allow their
% parameters to be overridden by the user; therefore, the user can specify, e.g., 'srate',500, to override the sampling rate of a paradigm. Further, most paradigms
% forward all user-supplied parameters to the default pipeline (and not just those for the stages that they enable by default), so that all the following 
% options can be set, unset, or altered for most paradigms, by specifying the appropriate parameter name in the paradigm's parameter list (which is, usually,
% when using the paradigm through bci_train -- see bci_train for the full explanation of paradigms and their parameterizations).
%
% In:
%   In-Data    : a dataset to be processed (as, e.g., loaded with io_loadset)
%   
%   Options... : optional name-value pairs that enable various stages of the default preprocessing pipeline:
%
%                Most of the filters/flt_* and some of the dataset_ops/set_* functions are in the pipeline and their parameters can be set
%                by using their name (without the prefix), or one of the listed alternative names (e.g., 'baseline' instead of 'rmbase').
%                In general, the value is a cell array of the arguments that are to be passed to the respective function (excluding the 
%                dataset and state arguments of the filter functions, which are not user-accessible) or, in case of a single argument, 
%                alternatively the argument without the enclosing cell can be passed (e.g. 500 instead of {500} for 'resample'/'srate'). 
%                Functions that do not require arguments are enabled by passing 'on' after the name.
%
%                * 'resample'/'srate': resample to the given sampling rate, in Hz (see flt_resample)
%                                      example: flt_default(X,...,'resample',200,...)
%
%                * 'reref'/'ref': channel(s) for re-referencing or {[]} for average reference (see flt_reref)
%                                 example: flt_default(X,...,'ref',{'TP7','TP8'},...)
%
%                * 'ica': derive an ICA decomposition with options (see flt_ica)
%                         example: flt_default(X,...,'ica','on',...), flt_default(X,...,'ica',{'amica','max_iter',300},...)
%
%                * 'selchans'/'channels': select a channel subset, typically a cell-string array (see flt_selchans)
%                                         example: flt_default(X,...,'channels',{'C3' 'C4'},...) 
%
%                * 'laplace': simple surface laplacian (see flt_laplace); by passing a cell array of channels,
%                             only those channels are retained after applying the laplacian
%                             example: flt_default(X,...,'laplace','on',...), flt_default(X,...,'laplace',{'C3' 'C4'}, ...
%
%                * 'iir': apply an IIR-based frequency filter (see flt_iir)
%                             example: flt_default(X,...,'iir',[7 8 29 31],...), flt_default(X,...,'iir',{[10 15],'lp'},...) 
%
%                * 'bandpass','lowpass','highpass','bandstop': apply an IIR-based frequency filter with mode predefined (see flt_iir);
%                                                              value is as in 'iir', except that the mode argument is omitted
%                                                              example: flt_default(X,...,'highpass',[2.5 10],...) 
%
%                * 'fir': apply an FIR-based frequency filter (see flt_fir)
%                             example: flt_default(X,...,'fir',{[0.5 2],'hp','minimum-phase'}, ...)
%
%                * 'standardize' : standardize the channels using a window of past data, length optionally specified in seconds (see flt_standardize)
%                                  example: flt_default(X,...,'standardize','on',...)
%
%                * 'makepos'/'epoch': Extract epochs from the data (and deduce a target variable from the time-locking event types) (see set_makepos).
%                                     In most cases specified as {[begin end], {evt1,evt2,evt3,...}} where begin/end are in seconds and evt1/... are 
%                                     event types (or cell arrays thereof) around which the epochs are extracted; the target variable is then set to K for evtK.
%                                     Note: If events are not specified here, they must be specified separately in an 'events' parameter.
%                                     example: flt_default(X,...,'epoch',{[-2 0],'keypress1','keypress2'},...) 
%
%                * 'events': cell array of events to be used for makepos/epoch (takes precedence over events that are specified in the makepos/epoch clause)
%                            example: flt_default(X,...,'events',{'keypress1','keypress2'},...,'epoch',[-2 0],...) 
%
%                * 'rmbase'/'baseline': remove a baseline window of the given epoch(s) (see flt_rmbase)
%                                       example: flt_default(X,...,'rmbase',[-2 -1.7],...) 
%
%                * 'select': apply an epoch-based selection of a data portion (in time and/or frequency) (see flt_select)
%                            example: flt_default(X,...,'select',{'time',{'hann' [-2 2]}},...) 
%
%                * 'window': apply a window function (value as in flt_select(eeg,'time',value), i.e. 'hann', {'gauss',alpha}, etc.)
%                            example: flt_default(X,...,'window','hann',...), flt_default(X,...,'window',{'gauss',1.5,[0 4]},...)
%
%                * 'spectrum': apply a free-form spectral filter (value as in flt_select(eeg,'freq',value), i.e. [7 13; 20 25], @(f)f>30, etc.)
%                              note: if no epochs are being extracted, this gives a non-causal filtering of the data
%                              example: flt_default(X,...,'spectrum',[6.5 7.5 27 33],...), flt_default(X,...,'spectrum',@(f)1./exp(0.2*(f-20).^2),...) 
%
%                * 'reconstruct': reconstruct the signal in terms of a new (possibly overcomplete) basis (see flt_reconstruct)
%                                 example: flt_default(X,...,'reconstruct',[icaweights1,icaweights2,icaweights3],...)
%
%                * 'project': apply a custom spatial projection matrix to the data (see flt_project)
%                             example: flt_default(X,...,'project',eye(128),...)
%
%                * 'fourier': transform (usually epoched) data into a Fourier representation (see flt_fourier)
%                             example: flt_default(X,...,'fourier','amplitude',...)
%
%                Advanced Options: In general, custom filters can be hooked into the default pipeline, before and/or after any
%                of the default stages (inrrespective of whether that stage is enabled or not), by specifying an option of the following form:
%
%                * 'pre_<defaultstage>' / 'post_<defaultstage>': execute a custom filter step before or after the point where one of the above 
%                                                                default stages would be applied (they are listed in the order of their their application above)
%
%                                                                the value is a cell array of the form {functionhandle, argument2, argument3, ...} and leads
%                                                                to the given function handle being called with the current data as its first argument,
%                                                                and the following arguments taken as the user-supplied argument2, argument3, etc.
%                                                                example: flt_default(X,...,'post_ica',{@mycalcact,10,'test'}, ...)
%                                                                         ... executes eeg = mycalcact(eeg,10,'test') directly following the ICA step
%                                                                note: stateful filters require that the state argument is supplied as argument2 (usually,
%                                                                      with that argument specified as [])
%
%                                                                if the value is a cell array of cell arrays, they are executed as a sequence of custom steps
%
% Out:
%   Out-Data   : the processed dataset
%
%                                                 Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                                                 2010-04-29

% set defaults
opts = hlp_varargin2struct(varargin, {'resample','srate'},[], {'reref','ref'},[], 'ica',[], 'deblink',[], {'selchans','channels'},[], ...
    'laplace',[], 'iir',[], 'bandpass',[], 'lowpass',[], 'highpass',[], 'bandstop',[], 'standardize',[], {'makepos','epoch'},[], {'rmbase','baseline'},[], ...
    'select',[], 'window',[], 'spectrum',[], 'reconstruct',[], 'project',[], 'events',[], 'fourier',[],'fir',[]);

% make sure that all enabled options are wrapped in cells (disabled ones remain [])
% 'on' is translated into an empty arguments cell array
% (for functions that take a cell array as first argument, we must additionally make sure that the first cell is a cell array itself)
for fn = fieldnames(opts)'
    fname = fn{1};
    if strcmp(opts.(fname),'on')
        opts.(fname) = {}; end
    if strcmp(opts.(fname),'off')
        opts.(fname) = []; end
    if ~isempty(opts.(fname)) && ~iscell(opts.(fname))
        opts.(fname) = {opts.(fname)}; end
end

exp_declare_symbols partitioner

% resampling
data = flt_step(data,{'pre_resample','pre_srate'},opts);
if iscell(opts.resample)
    data = flt_resample(data,[],opts.resample{:}); end
data = flt_step(data,{'post_resample','post_srate'},opts);

% re-referencing
data = flt_step(data,{'pre_reref','pre_ref'},opts);
if iscell(opts.reref)
    % single channels may be specified without the {}
    if ~isempty(opts.reref) && ischar(opts.reref{1})
        opts.reref{1} = opts.reref(1); end
    data = flt_reref(data,opts.reref{:}); 
end
data = flt_step(data,{'post_reref','post_ref'},opts);

% ICA derivation (with implicit partitioning)
data = flt_step(data,'pre_ica',opts);
if iscell(opts.ica)
    data = flt_ica((data),[],opts.ica{:}); end
data = flt_step(data,'post_ica',opts);

% blink correction
data = flt_step(data,'pre_deblink',opts);
if iscell(opts.deblink)
    data = flt_deblink(data,opts.deblink{:}); end
data = flt_step(data,'post_deblink',opts);

% channel subsetting
data = flt_step(data,{'pre_selchans','pre_channels'},opts);
if iscell(opts.selchans)
    if ischar(opts.selchans{1})
        opts.selchans = {opts.selchans}; end
    data = flt_selchans(data,opts.selchans{:}); end
data = flt_step(data,{'post_selchans','post_channels'},opts);

% surface laplacian
data = flt_step(data,'pre_laplace',opts);
if iscell(opts.laplace)
    data = flt_laplace(data);
    % do a post-laplacian channel selection
    if ~isempty(opts.laplace) && iscell(opts.laplace{1})
        data = flt_selchans(data, opts.laplace{1}); end
end
data = flt_step(data,'post_laplace',opts);


% IIR filter
data = flt_step(data,'pre_iir',opts);
if iscell(opts.iir)
    data = flt_iir(data,[],opts.iir{:}); end
data = flt_step(data,'post_iir',opts);

% lowpass, highpass, bandpass, bandstop filter
data = flt_step(data,'pre_bandpass',opts);
data = flt_step(data,'pre_bandstop',opts);
data = flt_step(data,'pre_lowpass',opts);
data = flt_step(data,'pre_highpass',opts);
if iscell(opts.bandpass)
    data = flt_iir(data,[],opts.bandpass{1},'bp',opts.bandpass{2:end}); end
if iscell(opts.lowpass)
    data = flt_iir(data,[],opts.lowpass{1},'lp',opts.lowpass{2:end}); end
if iscell(opts.highpass)
    data = flt_iir(data,[],opts.highpass{1},'hp',opts.highpass{2:end}); end
if iscell(opts.bandstop)
    data = flt_iir(data,[],opts.bandstop{1},'bs',opts.bandstop{2:end}); end
data = flt_step(data,'post_bandpass',opts);
data = flt_step(data,'post_lowpass',opts);
data = flt_step(data,'post_highpass',opts);
data = flt_step(data,'post_bandstop',opts);

% FIR filter
data = flt_step(data,'pre_fir',opts);
if iscell(opts.fir)
    data = flt_fir(data,[],opts.fir{:}); end
data = flt_step(data,'post_fir',opts);

% standardization
data = flt_step(data,'pre_standardize',opts);
if iscell(opts.standardize)
    data = flt_standardize(data,[],opts.standardize{:}); end
data = flt_step(data,'post_standardize',opts);

% epoch generation
data = flt_step(data,{'pre_makepos','pre_epoch'},opts);
if iscell(opts.makepos)
    if iscell(opts.events)
        opts.makepos{2} = opts.events; end
    % special window-generation heuristic
    if strcmp(opts.makepos{1},'blankertz-erd')
        opts.makepos = [utl_select_timeival(partitioning(data),opts.opts.makepos{2}) opts.makepos{2:end}]; end
    data = set_makepos(data,opts.makepos{:}); 
end
data = flt_step(data,{'post_makepos','post_epoch'},opts);

% baseline correction
data = flt_step(data,{'pre_rmbase','pre_baseline'},opts);
if iscell(opts.rmbase)
    data = flt_rmbase(data,opts.rmbase{:}); end
data = flt_step(data,{'post_rmbase','post_baseline'},opts);

% epoch data selection
data = flt_step(data,'pre_select',opts);
if iscell(opts.select)
    data = flt_select(data,opts.select{:}); end
data = flt_step(data,'post_select',opts);

% window function
data = flt_step(data,'pre_window',opts);
if iscell(opts.window)
    data = flt_select(data,'time',opts.window); end
data = flt_step(data,'post_window',opts);

% free-form spectral filter
data = flt_step(data,'pre_spectrum',opts);
if iscell(opts.spectrum)
    data = flt_select(data,'freq',opts.spectrum); end
data = flt_step(data,'post_spectrum',opts);

% sparse reconstruction
data = flt_step(data,'pre_reconstruct',opts);
if iscell(opts.reconstruct)
    data = flt_reconstruct(data,opts.reconstruct{:}); end
data = flt_step(data,'post_reconstruct',opts);

% projection
data = flt_step(data,'pre_project',opts);
if iscell(opts.project)
	data = flt_project(data,opts.project{:}); end
data = flt_step(data,'post_project',opts);

% Fourier transform
data = flt_step(data,'pre_fourier',opts);
if iscell(opts.fourier)
	data = flt_fourier(data,opts.fourier{:}); end
data = flt_step(data,'post_fourier',opts);



% convenience: if an option named 'tag' is given, treat it as a cell array of the form
% {head, arg2, arg3, arg4, ...} and evaluate it as a filter stage of the form
% eeg = head(eeg, arg2, arg3, arg3, arg4, ...);
function data = flt_step(data,tag,opts)
% uniformize tag format
if ~iscell(tag) 
    tag = {tag}; end
% check if any tag matches one of the supplied options
match = isfield(opts,tag);
if any(match)
    if nnz(match)>1 
        % if there were multiple options using valid names, we take the one that was specified last (i.e. comes last in the opts fieldnames)
        [names,idx2fnidx] = intersect(fieldnames(opts),tag(match));
        option = opts.(names{argmax(idx2fnidx)});
    else
        option = opts.(tag{match});
    end
    % uniformize option
    if ~iscell(option{1})
        option = {option}; end
    % execute all steps in sequence
    for c=1:length(option)
        data = option{c}{1}(data,option{c}{2:end});  end
end
