function eeg = flt_fourier(eeg, rep, normalized)
% Output-Set = flt_fourier(Input-Set, Representation, Normalized)
% Transform an epoched dataset into a fourier representation.
%
% The fourier representation is practical if there is a highly complex (relevant) spectral structure in the data. In any representation except for 'complex',
% it is necessary that an appropriate spatial filter has been applied to the data beforehand (e.g., ICA with the 'transform' option set to 1, or the 
% surface laplacian). The complex representation can be spatially transformed after running the Fourier filter, since it is a linear operator. If classifiers 
% operate directly on the fourier representation, it is usually necessary to chose a representation that can be linearly mapped to a prediction, such as 
% 'amplitude', 'polar' or 'phase'; 'complex' requires a non-linear classification, which is almost certain to overfit any other random non-linear dependencies 
% in the data. Some interesting assumptions that can be imposed on Fourier data include sparsity (l1) and group sparsity (if independent component activity was 
% transformed).
%
% In:
%   Input-Set      :   Epoched eeg set to be processed
%
%   Representation : desired representation; can be one of the following:
%                    'complex'  : the epoch contains the complex spectrum (default)
%                    'power'    : the epoch contains the power spectrum
%                    'logpower' : the epoch contains the logarithmic power spectrum
%                    'amplitude': the epoch contains the amplitude spectrum (sqrt() of the power spectrum)
%                    'phase'    : the epoch contains the phase spectrum
%                    'polar'    : the epoch contains the amplitude/phase spectrum (first N channels: amplitude, remaining N channels: phase)
%
%   Normalized     : whether the spectrum should be normalized by 1/f (default: 1)
%
% Out: 
%   Output-Set  :   baseline-corrected eeg set
%
%                   Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                   2010-03-28

if ~exp_begindef('filter') return; end
if ~exist('rep','var') || isempty(rep) rep = 'complex'; end
if ~exist('normalized','var') || isempty(normalized) normalized = 1; end

% first transform into the spectrum
eeg.data = fft(eeg.data,[],2);
if normalized
    eeg.data = bsxfun(@times,eeg.data,(0:size(eeg.data,2)-1)/size(eeg.data,2)); end

% then map to the desired representation
switch rep
    case 'power'
        eeg.data = abs(eeg.data);
    case 'logpower'
        eeg.data = log(abs(eeg.data));
    case 'amplitude'
        eeg.data = real(sqrt(abs(eeg.data)));
    case 'phase'
        eeg.data = real(angle(eeg.data));
    case 'polar'
        eeg.data = real([sqrt(abs(eeg.data)); angle(eeg.data)]);
        % replicate & rename channels
        labels = [cellfun(@(s) [s '_a'], {eeg.chanlocs.labels},'UniformOutput',false) cellfun(@(s) [s '_p'], {eeg.chanlocs.labels},'UniformOutput',false)];        
        eeg.chanlocs = eeg.chanlocs([1:end 1:end]);
        [eeg.chanlocs.labels] = labels{:};
    case 'complex'
        % nothing to do
    otherwise
        error('unsupported representation selected.');        
end
% restrict to the first half of the spectrum
eeg.data = eeg.data(:,1:size(eeg.data,2)/2,:);

exp_enddef;