function [eeg,newstate] = flt_iir(eeg,oldstate,f,fmode,ftype,atten,ripple)
% [Result-Set,State] = flt_iir(Input-Set, State, Freqs, Mode, Type, Attenuation, Ripple)
% Filter a continuous eeg set using a digital IIR lowpass/highpass/bandpass/bandstop filter.
%
% Digital IIR filters [1] are efficient for both offline and online analysis. They distort the signal, but introduce relatively low delay (comparable to minimum-phase FIR 
% filters), so that the latency of a BCI is only marginally increased by adding an IIR filter. However, IIR filters are numerically sensitive (they can "blow up"),
% and therefore, extreme frequency responses (e.g., low-frequency 'brickwall' filters) can often not be satisfactorily implemented. In these cases, FIR or FFT filters 
% (flt_fir or flt_select, respectively) can be used as a fall-back.
%
% In:
%   Input-Set    :   continuous eeg set to be filtered
%
%   State        :   previous filter state, as obtained by a previous execution of flt_iir on an immediately preceding dataset (default: [])
%
%   Freqs        :   frequency specification: 
%                    * for a low/high-pass filter, this is: [transition-start, transition-end], in Hz
%                    * for a band-pass/stop filter, this is: [low-transition-start, low-transition-end, hi-transition-start, hi-transition-end], in Hz
%
%   Mode         :   filter mode, 'bp' for bandpass, 'hp' for highpass, 'lp' for lowpass, 'bs' for bandstop (default: 'bp')
%
%   Type         :   'butter' for a Butterworth filter -- pro: flat response overall; con: slow attenuation (default)
%                    'cheb_type1' for a Chebychev Type I filter -- pro: steep attennuation; con: strong passband ripple
%                    'cheb_type2' for a Chebychev Type II filter -- pro: flat passband response; con: slower 
%                                 attenuation than cheb_type1
%                    'elliptic' for an Elliptic filter -- pro: steepest rolloff, lowest latency; con: passband ripple
%
%   Attenuation  :   stop-band attenuation, in db, default: 50
%
%   Ripple       :   maximum allowed pass-band ripple, in db, default: 3
%
% Out: 
%   Result-Set   :  filtered, continuous EEGLAB dataset
%
%   State        :  state of the filter, can be used to continue on a subsequent portion of the data
%
% Example:
%   filtered = flt_iir(raw,[],[5 10 25 35]);  % Apply a digital IIR band-pass filter to the raw data set.
%
% References:
%  [1] T. W. Parks and C. S. Burrus, "Digital Filter Design",
%      John Wiley & Sons, 1987, chapter 7.
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-17

if ~exp_begindef('filter') return; end
if ~exist('fmode','var') || isempty(fmode) fmode = 'bp'; end;
if ~exist('ftype','var') || isempty(ftype) ftype = 'butter'; end;
if ~exist('atten','var') || isempty(atten) atten = 50; end;
if ~exist('ripple','var') || isempty(ripple) ripple = 3; end;

if ~isnumeric(f)
    % online case
    [b,a] = f{:};
else
    % numeric cutoff freqs given, need to design filters first:
    f = [2*f/eeg.srate 0 0];
    % define order-calculating function for each filter type (takes frequencies, gives n,Wx)
    ordfunc = struct('butter',{@buttord},'cheb_type1',{@cheb1ord},'cheb_type2',{@cheb2ord},'elliptic',{@ellipord});
    % define cutoff parameter order (in the order-calculating function) for each filter mode
    ordparams = struct('bp',{{f([2,3]),f([1,4])}},'bs',{{f([1,4]),f([2,3])}},'lp',{{f(1),f(2)}},'hp',{{f(2),f(1)}});
    % define filter coefficient function for each filter type (takes n,Wx, gives coeffs)
    coeffunc = struct('butter',{@(n,Wx)params(@butter,n,Wx,fmode)},'cheb_type1',{@(n,Wx)params(@cheby1,n,0.5,Wx,fmode)}, ...
                      'cheb_type2',{@(n,Wx)params(@cheby2,n,20,Wx,fmode)},'elliptic',{@(n,Wx)params(@ellip,n,0.5,20,Wx,fmode)});
    % compute order & coeffs
    [n,Wn] = ordfunc.(ftype)(ordparams.(fmode){:},ripple,atten);
    [b,a] = coeffunc.(ftype)(n,Wn);
end

% apply the filter
[eeg.data,newstate] = filter(double(b),double(a),double(eeg.data),oldstate,2);
% check for divergence
naninf = find(~isfinite(eeg.data(:)));
if any(naninf)
    if ~isempty(oldstate)
        % filter divergence online: zero-out and reset filter
        eeg.data(naninf) = 0;
        newstate = [];
        warning('BCILAB:flt_iir:divergence','filter reset due to divergence');
    else
        % offline we throw an error, as this situation should be fixed by the user
        error('filter divergence'); 
    end
end

% finish
f = {b,a}; %#ok<NASGU>
exp_enddef;


% helper: adapt the butter/cheby1/cheby2/ellip functions to accept 'bp','bs','lp','hp' as filter type strings
function [b,a] = params(func,varargin)
if any(strcmp(varargin,'bp'))
    [b,a] = func(varargin{~strcmp(varargin,'bp')});
else
    varargin(strcmp(varargin,'bs')) = {'stop'};
    varargin(strcmp(varargin,'lp')) = {'low'};
    varargin(strcmp(varargin,'hp')) = {'high'};
    [b,a] = func(varargin{:});
end
