function eeg = flt_rmbase(eeg, wnd)
% Output-Set = flt_rmbase(Input-Set, Window)
% Subtract a baseline from an data set, computed over the given baseline window.
%
% Baseline correction is a commonly used method in the analysis of Event-Related Potentials, to factor out irrelevant variations in a signal's baseline.
% It is applied to an epoched dataset, by specifying a sub-window that (typically) lies before the phenomena of interest, where the mean signal value in that window
% is subtracted from the entire signal (per channel or component). Baseline considerations are relevant for BCIs operating on slow cortical potentials, and can be 
% implemented in a variety of other ways, too. One is to use a highpass filter (flt_iir, flt_fir, flt_select) to subtract low-frequency drifts in the signal.
% Another one is to add features from the signal which measure the baseline, for example one or a collection of windows of various lengths prior to the phenomenon of 
% interest (in this case, the baseline correction is done by the machine learning algorithm that operates on these features). Paradigms which assign different weights 
% for every time point in an epoch typically do baseline correction implicitly.
%
% In:
%   Input-Set   :   Epoched eeg set to be processed.
%
%   Window      :   baseline window in s, e.g. [-0.5 -0.3]
%
% Out: 
%   Output-Set  :   baseline-corrected eeg set
%
%                   Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                   2010-03-28

if ~exp_begindef('filter') return; end
if ~exist('wnd','var') || isempty(wnd) wnd = [-Inf Inf]; end

eeg.data = double(eeg.data);
wnd = round(max(1,min(size(eeg.data,2),(wnd-eeg.xmin)*eeg.srate+1)));
eeg.data = eeg.data - repmat(mean(eeg.data(:,wnd(1):wnd(2),:),2),[1,size(eeg.data,2),1]);

exp_enddef;