function eeg = flt_select(eeg,varargin)
% Output-Set = flt_select(Input-Set, Options...)
% Select a portion of the data in an epoched data set.
%
% flt_select can be used to implement temporal and/or spectral filtering of the data, i.e. selection of a temporal and/or spectral portion of the data.
% Temporal selection is done by specifying a time window (within the epoch) and/or a smooth window function, to which the data should be reduced.
% Likewise, spectral selection is done by specifying one or more frequency windows (or a function of frequency), to which the data should be restricted.
% Temporal filtering is usually relevant for the analysis of event-locked processes, which are usually found to be localized in some time window near the event.
% It can also be useful to increase the precision of spectral estimates of the data by using a window function with a desirable frequency response [1].
% Spectral filtering is critical, on one hand, to separate oscillatory processes of interest (e.g. thalamo-cortical loops) from other signal contents (e.g.,
% muscle artifacts, slow cortical potentials) and on the other hand to isolate slow cortical potentials (e.g. some event-related negativity) from higher-frequency 
% signal content.
%
% In:
%   Input-Set  : epoched data set
%
%   Options... : name-value pairs of selection options:
%                'time': time-domain selection, allows for the specification of an epoch sub-interval and/or
%                        a window function, in one of the formats: 
%                          * {'windowfunction' windowparameter, [begin end]}
%                          * {'windowfunction',windowparameter}
%                          * {'windowfunction' [begin end]}
%                          * 'windowfunction'
%                          * [begin end]
%                        where begin/end are in seconds, 'windowfunction' is one of the MATLAB-supplied window functions:
%                         'barthann','bartlett','blackman',blackmanharris',bohman','cheb','flattop','gauss','hamming','hann',
%                         'kaiser','nuttall','parzen','rect','taylor','triang','tukey'
%                        cheb,kaiser and tukey require a window parameter, for other windows, this is optional (see MATLAB's window() for reference)
%                        'hann' and 'gauss' windows are supported without the Signal Processing toolbox.
%
%                'freq': frequency-domain selection, can be specified as one of the following
%                         * [low high] or [low high; low high; low high; ...] for a brickwall filter or sum of brickwall filters (numbers in Hz)
%                         * [low_begin low_end high_begin high_end; ...] for a linearly sloped filter or sum of sloped filters (numbers in Hz)
%                         * function_handle for an arbitrary function of frequency (in Hz)
%                        low and high frequencies are inclusive
%
% Out:
%   Output-Set      :   subset of the data set
%
% References:
%   [1] Oppenheim, A.V., and R.W. Schafer, "Discrete-Time Signal Processing", 
%       Prentice-Hall, 1989.
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-06-28

if ~exp_begindef('filter') return; end

opts = hlp_varargin2struct(varargin,'time',[],'freq',[]);

% time-domain selection
if ~isempty(opts.time)
    % do range selection, if specified
    range = [];
    if isnumeric(opts.time)
        range = opts.time; end
    if iscell(opts.time) && isnumeric(opts.time{end}) && length(opts.time{end}) == 2
        range = opt.time{end}; end
    if ~isempty(range) && ~isequal(range,[eeg.xmin eeg.xmax])
        eeg = pop_select(eeg,'time',range); end
    
    % apply window function, if specified    
    wnd = {};
    if ischar(opts.time)
        wnd = {opts.time,size(eeg.data,2)}; end
    if iscell(opts.time) && ischar(opts.time{1})
        if length(opts.time) > 1 && ((isnumeric(opts.time{2}) && length(opts.time{2}) == 2) || ischar(opts.time{2}))
            wnd = {opts.time{1},size(eeg.data,2),opts.time{2}};
        else
            wnd = {opts.time{1},size(eeg.data,2)};
        end
    end
    if ~isempty(wnd)
        windowmap = {'barthann',@barthannwin,'bartlett',@bartlett,'blackman',@blackman,'blackmanharris',@blackmanharris,'bohman',@bohmanwin,'cheb',@chebwin,...
            'flattop',@flattopwin,'gauss',@gausswin,'hamming',@hamming,'hann',@hann,'kaiser',@kaiserwin,'nuttall',@nuttallwin,'parzen',@parzenwin,'rect',@rectwin, ...
            'taylor',@taylorwin,'triang',@triangwin,'tukey',@tukeywin};
        W = repmat(window(windowmap.(wnd{1}),wnd{2:end}),size(eeg.data,1),1);
        for t=1:size(eeg.data,3)
            eeg.data(:,:,t) = eeg.data(:,:,t) * W; end
    end
end

% frequency-domain selection
if ~isempty(opts.freq)
    T = size(eeg.data,2);
    % almost correct frequency vector (fixed further below)
    frqs = (0:T-1)*eeg.srate/T;
    if isnumeric(opts.freq) && size(opts.freq,2) == 2
        % brickwall filters
        flt = zeros(1,T);
        for k=1:size(opts.freq,1)
            f = opts.freq(k,:);
            flt = max(flt,(frqs>=f(1) & frqs<=f(2))); end
    elseif isnumeric(opts.freq) && size(opts.freq,2) == 4
        % sloped filters
        flt = zeros(1,T);
        for k=1:size(opts.freq,1)
            f = opts.freq(k,:);
            flt = max(flt,min(max(0,min(1,(frqs-f(1)) / (f(2)-f(1)))), max(0,min(1, 1-(frqs-f(3)) / (f(4)-f(3)))))); 
        end
    else
        % function handle filters
        flt = opts.freq(frqs);
    end
    % fix the filter beyond the Nyquist freq
    flt = repmat(flt(:)' .* (frqs<(eeg.srate/2)),size(eeg.data,1),1);
    % filter the data
    for t=1:size(eeg.data,3)
        eeg.data(:,:,t) = 2*real(ifft(fft(eeg.data(:,:,t).*flt))); end
end

exp_enddef;