function fp = hlp_fingerprint(data)
% Fingerprint = hlp_fingerprint(Data)
% Make a fingerprint (hash) of the given data, reasonably efficiently; this includes all contents; however, large arrays (such as EEG.data) 
% are only spot-checked.
%
% In:
%   Data        : some data structure
%
% Out:
%   Fingerprint : an integer that identifies the data
%
% Notes:
%   The fingerprint is not unique and identifies the dataset only with a certain (albeit high) probability.
%
%                                   Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                                   2010-04-02

global bcilab;

try    
    if hlp_matlab_version < 707
        % save & override random state
        randstate = rand('state'); %#ok<*RAND>
        rand('state',5183);
    else
        % create a legacy-compatible RandStream for the fingerprinting
        bcilab.temp.randstream_fingerprint = RandStream('swb2712','Seed',5183);
    end
    
    % calculate the actual fingerprint
    fp = calchash(summarize(data));

catch
    e = lasterror; %#ok<LERR>
    env_handleerror(e);
end

if hlp_matlab_version < 707
    % restore random state
    rand('state',randstate);
end
    

% calculate a hash of a string
function hash = calchash(string)
po32 = [power(2,32),1/power(2,32)];
string = double(string);

hash = 0;
for i=1:length(string)
    hash = (33*hash + string(i))*po32(2);
    hash = (hash-floor(hash))*po32(1);
end
hash = uint32(hash);

    
% get a recursive string summary of arbitrary data
function s = summarize(x)
global bcilab

small_cutoff = 256;
large_fraction = 10000;

if isnumeric(x) || islogical(x)
    ne = numel(x);
    if ne <= 1
        s = sprintf('%d',x);        
    elseif ne <= small_cutoff
        % small matrix: obtain completely
        s = [sprintf('m(%d-%d:',size(x,1),size(x,2)) mat2str(x) ')'];
    else
        % large matrix: obtain a sparse random subset
        count = floor(small_cutoff + (ne-small_cutoff)/large_fraction);
        if hlp_matlab_version < 707
            indices = 1+floor((ne-1)*rand(1,count));
        else
            indices = 1+floor((ne-1)*rand(bcilab.temp.randstream_fingerprint,1,count));
        end            
        s = [sprintf('m(%d-%d:',size(x,1),size(x,2)) mat2str(x(indices)) ')'];
    end
elseif ischar(x)
    s = x(:)';
elseif iscell(x)
    s = cellfun(@summarize,x,'UniformOutput',false);
    s = [sprintf('c(%d-%d:',size(x,1),size(x,2)) s{:} ')'];
elseif isstruct(x)
	s = ['s(' summarize(fieldnames(x)) ':' summarize(struct2cell(x)) ')'];
elseif isobject(x)
    s = ['o(' char(class(x)) ':' summarize(struct(x)) ')'];
elseif isa(x,'function_handle')
    s = ['f(' char(x) ')'];
else
    warning('BCILAB:utl_fingerprint:unknown_type','utl_fingerprint: unknown type!');
    error; %#ok<LTARG>
end
