function res = io_loadset(filename, varargin)
% Set = io_loadset(Filename, Options...)
% Loads/imports a dataset from disk, across all formats supported by EEGLAB
% optionally, data can be loaded in reduced form to help conserve memory where applicable.
%
% In:
%   Filename      :   name of the file; platform-independent path preferred.
%
%   Options...    :   optional name-value pairs, with possible names:
%                     'channels'    : channel index subset to be loaded (memory-efficient for .vhdr, .eeg, .bdf (if BioSig works), .ctf, .ds)
%                     'samplerate'  : sampling rate for on-the-fly downsampling or if missing (necessary for .raw)
%                     'samplerange' : sample range to be read; [first_sample last_sample] (ONLY supported for .vhdr, .raw (EGI), .cnt (Neuroscan, EEProbe))
%                     'timerange'   : time range to be read, in seconds; [begin_time end_time] (ONLY supported for .cnt (Neuroscan, EEProbe), .bdf, .ctf, .ds)
%
%                     additional options supported for 
%                       .sna ('gain' -> pop_snapread)
%                       .cnt (-> pop_readcnt/readcnt)
%                       .eeg (-> pop_loadeeg)
%                       .bdf (-> pop_biosig/pop_readbdf)
%                       .ctf (-> 'trials' -> pop_readctf/ctf_read)
%                       .ds (-> 'trials' -> pop_ctf_read)
%
% Out:
%   Set :   dataset in EEGLAB format
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-03-28

if ~exp_begindef('filter') return; end

% read options
opts = hlp_varargin2struct(varargin,{'channels','chans'},[],{'samplerate','srate'},[],{'samplerange','srange'},[],{'timerange','trange'},[]);
filename = env_translatepath(filename);
[base,name,ext] = fileparts(filename);

% bitrate test function
wrong_bitrate = @(eeg) mad(mean(eeg.data(:,1:2:end),2)./mean(eeg.data(:,2:2:end),2)) > 0.5;

% dispatch to the appropriate loader / importer
disp(['loading ' filename '...']);
switch lower(ext)
    case '.set'
        % EEGLAB dataset
        args = hlp_struct2varargin(opts,'suppress',{'channels','samplerate','samplerange','timerange'});
        res = pop_loadset('filepath',[base filesep], 'filename', [name ext], args{:});
        if ~(isfield(opts,'nofixups') && opts.nofixups) && (~isfield(res,'bcilab') || ~isfield(res.bcilab,'online_expression'))
            disp('The loaded EEGLAB set is lacking an online expression; assuming it contains unfiltered data.')
            disp('If it contains filtered data, however, BCI models derived from it will likely not be online-capable.');
        end
    case '.vhdr'
        % BrainProducts Vision Recorder file
        res = pop_loadbv([base filesep],[name ext],opts.samplerange,opts.channels);
        % remove non-standard event fields
        res.event = rmfield(res.event,intersect({'code','bvtime','channel'},fieldnames(res.event)));
    case '.raw'
        try
            % EGI continuous raw data        
            res = pop_readegi(filename,opts.samplerange);
        catch
            disp('EGI .raw importer failed; falling back to ERPSS .raw importer.');
            % ... or ERPSS raw
            res = pop_read_erpss(filename,opts.samplerate);
        end
    case '.sna'
        % Snapmaster SNA
        args = hlp_struct2varargin(opts,'restrict',{'gain'});
        res = pop_snapread(filename,args{:});
    case '.cnt'
        % Neuroscan CNT
        try
            args = hlp_struct2varargin(opts,'suppress',{'channels','samplerate','samplerange','timerange'});
            if ~isempty(opts.samplerange)
                args = [args {'sample1',opts.samplerange(1),'ldnsamples',opts.samplerange(2)-opts.samplerange(1)+1}]; end
            if ~isempty(opts.timerange)
                args = [args {'t1',opts.timerange(1),'lddur',opts.timerange(2)-opts.timerange(1)}]; end
            res = pop_loadcnt(filename, args{:});
            % check if it should have been 32 bits...
            if wrong_bitrate(res) && ~isfield(opts,'dataformat')
                disp('The data is likely 32 bits; re-loading.');
                res = pop_loadcnt(filename, args{:}, 'dataformat','int32');
            end
        catch
            disp('Neuroscan importer failed; falling back to ANT EEProbe importer.');
            % ... or ANT EEProbe CNT
            args = hlp_struct2varargin(opts,'restrict',{'triggerfile'});
            if ~isempty(opts.samplerange)
                args = [args {'sample1',opts.samplerange(1),'sample2',opts.samplerange(2)}]; end
            if ~isempty(opts.timerange)
                args = [args {'time1',opts.timerange(1),'time2',opts.timerange(2)}]; end
            res = pop_loadeep(filename,args{:});
        end
    case '.avr'
        % ANT EEProbe average file
        res = pop_loadeep_avg(filename);
    case '.eeg'
        % Neuroscan EEG
        optseeg = hlp_varargin2struct(opts,'range_trials',[],'range_typeeeg',[],'range_response',[],'format','short');
        res = pop_loadeeg([name ext],[base filesep],opts.channels,optseeg.range_trials,optseeg.range_typeeg,optseeg.range_response,optseeg.format);
        if wrong_bitrate(res) && ~isfield(opts,'format')
            disp('The data is likely 32 bits; re-loading.');
            res = pop_loadeeg([name ext],[base filesep],opts.channels,optseeg.range_trials,optseeg.range_typeeg,optseeg.range_response,'int32');
        end
    case {'.bdf','.edf'}
        % BioSEMI BDF/EDF
        try
            args = hlp_struct2varargin(opts,'suppress',{'samplerate','samplerange'}, 'rewrite',{'timerange','blockrange'});
            res = pop_biosig(filename,args{:});
        catch
            disp('BioSig importer failed; falling back to EEGLAB''s importer.');
            % backup variant
            args = hlp_struct2varargin(opts,'suppress',{'channels','samplerate','samplerange'}, 'rewrite',{'timerange','range'});
            res = pop_readbdf(filename,args{:});
        end
    case '.mat'
        % BrainVision Analyzer Matlab file
        try
            res = pop_loadbva(filename);
        catch
            disp('BrainVision importer failed; falling back to MATLAB''s load().');
            % backup
            res = load(filename);
        end
    case '.ds'
        % CTF folder
        optsctf = hlp_varargin2struct(opts,'trials',[]);
        res = pop_ctf_read(filename,opts.channels,opts.timerange,optsctf.trials);
    case '.rdf'
        % ERPSS data
        res = pop_read_erpss(filename,opts.samplerate);
    case '.asc'
        % INStep ASC
        res = pop_loadascinstep(filename);
    case '.m4d'
        % 4D pdf file
        res = pop_read4d(filename);
    otherwise
        % Miscellaneous formats
        try
            % try File-IO
            args = hlp_struct2varargin(opts,'suppress',{'timerange','samplerate'},'rewrite',{'samplerange','samples'});
            res = pop_fileio(filename,args{:});
        catch
            disp('File-IO importer failed; falling back to BioSig importer.');
            fileio_error = lasterror; %#ok<LERR>
            try
                % try BioSig
                args = hlp_struct2varargin(opts,'suppress',{'samplerate','samplerange'}, 'rewrite',{'timerange','blockrange'});
                res = pop_biosig(filename,args{:});
            catch
                disp('BioSig importer failed; falling back to BCI2000 importer.');
                biosig_error = lasterror; %#ok<LERR>
                try
                    % try to load it as a BCI2000 ASCII file
                    res = pop_loadbci(filename,opts.samplerate);
                catch
                    env_handleerror(fileio_error);
                    env_handleerror(biosig_error);
                    error('Unknown data format; please check for plugins at www.sccn.ucsd.edu/eeglab/plugins/');                    
                end
            end
        end
end

% if no channel locations present, try to look them up
if ~(isfield(opts,'nofixups') && opts.nofixups) && (~all(isfield(res.chanlocs,{'X','Y','Z'})) || all(cellfun('isempty',{res.chanlocs.X})))
    % ... using the 10-20 system
    try
        res = pop_chanedit(res,'lookup','standard-10-5-cap385.elp');
        fraction_found = 1-mean(cellfun('isempty',{res.chanlocs.X}));
        % if only a small fraction of channel names were found
        if fraction_found < 0.25
            % check if it's the SCCN's custom cap
            locdb = readlocs(env_translatepath('resources:/sccn-cap-128.elp'));
            [found,idx_in_locdb,idx_in_res] = intersect({locdb.labels},{res.chanlocs.labels});
            if length(found)/length(res.chanlocs) > fraction_found
                % we found more channel names in that cap: copy the fields from there
                for fname = fieldnames(locdb)'
                    [res.chanlocs(idx_in_res).(fname{1})] = locdb(idx_in_locdb).(fname{1}); end
            end
        end
        % optimize the head center
        res = pop_chanedit(res,'eval','chans = pop_chancenter( chans, [],[]);');
    catch
    end
end

% describe it (note: the @rawdata expression indicates that this stage produces raw data...)
if isfield(res,'bcilab') && isfield(res.bcilab,'online_expression')
    exp_enddef('add_online',res.bcilab.online_expression)
else
    exp_enddef('add_online',@rawdata);
end

% and reduce the data if not already done so in the loader
if ~(isfield(opts,'nofixups') && opts.nofixups) && ~isempty(opts.channels)
    res = set_selchans(res,opts.channels); end

if ~(isfield(opts,'nofixups') && opts.nofixups) && ~isempty(opts.samplerate)
    res = flt_resample(res,[],opts.samplerate); end
