function fp = io_save(filename, varargin)
% function io_save(Filename, Arguments...)
% Like MATLAB's builtin save(), except that the filename can be a platform-independent path.
% Also, additional -options are supported, which offer additional fault-tolerance and automation, especially for long-running batch scripts.
%
% In:
%   Filename    :   Platform-Independent file name.
%
%   Arguments...:   optional arguments to save; in addition to what is supported by save(), the following options are allowed:
%                   -makedirs   : make directories (recursively), if possible
%                   -retryinput : ask the user whether he/she wants to retry creating the directory or the file, if unsuccessful (using input())
%                   -attributes attribs : specify file/directory attributes; these are the second or second and third argument to be passed to 
%                                         the fileattrib() function, separated by a comma, to set attributes for the current user or specific user groups, 
%                                         e.g., '+w' or '+w,a'.
%
% Example:
%   % save x and y to some mat file
%   io_save('data:/temp/results.mat', 'x','y');
%
% See also:
%   io_load()
%                                                     
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-06-21


% translate the file name
filename = env_translatepath(filename);

% parse arguments...
saveargs = {};      % args to save()
makedirs = false;
retryinput = false;
attributes = [];
i = 1;
while i <= length(varargin)
    arg = varargin{i};
    if ischar(arg) && ~isempty(arg) && arg(1) == '-'
        % we are dealing with an option argument
        switch(arg)
            case '-makedirs'
                makedirs = true;
            case '-retryinput'
                retryinput = true;
            case '-attributes'
                attributes = varargin{i+1};
                i = i+1; % skip next arg, too
            otherwise
                % regular argument; append it
                saveargs{end+1} = arg;
        end
    else
        % append the argument
        saveargs{end+1} = arg;
    end
    i = i+1; % go to next argument
end

% reformat the save-args into a string
tmp = [saveargs' repmat({' '},length(saveargs),1)]'; 
saveargs = [tmp{:}];

% reformat the attributes into a cell array...
attributes = eval(['{' attributes '}']);

while 1    
    if makedirs
        % first ensure that the directory exists
        while 1
            try
                io_mkdirs(filename,attributes);
                break;
            catch
                warning('BCILAB:io_save:cannot_create_directory',['Could not create target directory for ' filename ', probably due to permissions.']);
                if retryinput
                    disp('Please intervene or continue without saving.');
                    if strcmp(input('Retry? yes/no: ','s'),'no')
                        disp('skipped saving.');
                        break;
                    end
                else
                    break;
                end
            end
        end
    end
    
    % now save
    try
        evalin('caller',['save ''' filename ''' ' saveargs ';']);
        if ~isempty(attributes)
            % try to set file attributes...
            if ~isempty(attributes)
                warning off MATLAB:FILEATTRIB:SyntaxWarning
                fileattrib(filename,attributes{:}); 
            end            
        end
        break;
    catch
        warning('BCILAB:io_save:cannot_save_file',['Could not save ' filename ', probably due to permissions.']);
        if retryinput
            disp('Please intervene or continue without saving.');
            if strcmp(input('Retry? yes/no','s'),'no') 
                disp('Did not save the data.');
                break;
            end
        else
            break;
        end
    end
end
