function pred = ml_predictvote(trials, model)
% [Prediction] = ml_predictvote(Trials, Model)
% predict the value of a target variable from Trials, using a 1-vs-1 voting arrangement of binary classifier
%
% In:
%   Trials  : the data a matrix, as in ml_predict
%
%   Model   : predictive model as produced by ml_trainvote
%
% Out:
%   Prediction  : discrete probability distribution, formatted as
%                 {'disc' [NxC] [Cx1]}, with element #2 being the per-class probability and 
%                 element #3 the original target values per class
%                 thus, the expected target values are Prediction{2}*Prediction{3}
%
% See also:
%   ml_trainvote()
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-06-25

if isfield(model,'voted')
    % we do have a voted model
    predictor = model.opts{2};
    % we could have an NxF matrix of N F-dimensional feature vectors or a UxVxN matrix of N UxV-dimensional feature matrices...
    trialcount = size(trials,fastif(ndims(trials)==2,1,3));
    pred = {'disc' , zeros(trialcount,length(model.classes)), model.classes};
    % vote, adding up the probabilities from each vote
    for i=1:length(model.classes)
        for j=i+1:length(model.classes)
            outcome = predictor(trials,model.voted{i,j});
            if iscell(outcome)
                % the classifier produces a discrete probability distribution (preferred)
                pred{2}(:,[i j]) = pred{2}(:,[i j]) + outcome{2};
            elseif isnumeric(outcome)
                pred{2}(:,[i j]) = pred{2}(:,[i j]) + [outcome == model.classes(i), outcome == model.classes(j)];
            end
        end
    end
    % renormalize probabilities
    pred{2} = pred{2} ./ repmat(sum(pred{2},2),1,size(pred{2},2));
else
    % we don't have a voted model, nothing to do...
    pred = [];
end