function model = ml_train(data, args, varargin)
% Model = ml_train(Data, Arguments)
% learn a predictive model from the given data
%
% This function dispatches, depending on the first cell in Arguments, to one of the machine_learning/ml_train* functions (e.g., ml_trainlda, etc.),
% and may perform some additional data formatting depending on inputs and outputs. It is to be preferred over calling the ml_train* functions directly
% (especially since some callees may rely on the additional work done by ml_train).
%
% The function is the general tool to be applied when a predictive model shall be learned from a (usually unstructured, but labeled) set of trials [1,3].
% Each trial is typically fed to the function as a vector of previously extracted features (called a feature vector), as a generic format understood
% by a large variety of different methods [2]. Each trial typically has a label (or target value), which may be categorical (a number out of a small
% set of numbers), real-valued (or continuous: any number), multivariate (a vector of numbers), or structured (an arbitrary cell). If empty labels 
% are supplied, the goal is usually not to predict any labels, but to learn the probablity distribution of the supplied data.
%
% The resulting predictive model (usually a MATLAB struct()) is termed, depending on the label type, a 'classifier', 'regressor', or 'density estimator'.
% It captures in its parameters the necessary information to be able to predict the desired target value from a new instance of data (assumed to 
% originate from the same distribution as the training data). It is sometimes called the 'predictive function', and the term 'model' is also frequently
% used in the literature to refer to classes of such functions (with free parameters).
% 
% Chosing the right method for a given task is a difficult problem, and requires correct assumptions about the structure of the data, and often 
% further experimentation. The toolbox provides a relatively broad set of methods from different branches of machine learning, among others
% statistics (lda,qda,logreg), optimization (svm*,logreg,dal), support vector machines (svmlight,svmlinear,svmperf,rvm), (Bayesian) probability theory 
% (gmm,logreg,rvm) and meta-learning (vote). Prominent fields currently not covered are neural networks, fuzzy inference, and evolutionary methods.
%
% The ml_train* functions are named according to the fundamental approach, but typically several variants are provided by each, which themselves 
% may have multiple parameters and which may be implemented with different libraries. The overall approach can often be chosen based on assumptions
% about the data, whereas some parameters may have to be selected by parameter search (preferred for reasons explained in offline_analysis/bci_train), 
% or are otherwise adapted. In the context of the toolbox, it is usually not enough to select the right learning function, but instead, the entire
% data pipeline, usually starting with signal (pre-)processing, followed by feature extraction and concluded by machine learning needs to be determined.
% Extensive fully customizable default setups ("paradigms") are provided at higher levels (paradigms/para_*, offline_analysis/bci_train).
% 
% In:
%   Data    :   either a cell array of {Trials, Targets} or {Trials, Targets, Weights}, with
%                 - Trials a [NxF] numeric array (N... number of training instances, F... number of feature dimensions)
%                   or a [UxVxN] 3d array in special cases (N... number of training instances, U,V ... number of rows/columns of the feature matrices) 
%                 - Targets one of the following: 
%                    * a [NxT] array (N... number of training instances, T... number of target dimensions, T usually 1)
%                    * a {N} cell array (N... number of training instances, contents unspecified)  
%                      ... in this case, the method selected in Options must support structured prediction
%                    * an empty array, indicating that the probability density of the input data is to be estimated
%                 - Weights a [Nx1] numeric array of non-negative per-trial trial weights (N... number of training instances)
%               or, alternatively, a cell array of {{[N1xF],[N2xF], [N3xF], ...}}, with 
%                   Ni the number of instances for class i (possibly empty)
%                   F the number of feature dimensions                        
%
%   Arguments :   cell array of further arguments for learning the model; The first argument in Options is 
%                 considered the name of the learning method, e.g. 'lda', 'logreg', and specifies the ml_train<name> 
%                 learning function to be called all other options are passed unmodified into the learning function
% Out:
%   Model   : predictive model, can later be used in conjunction with ml_predict
%              Model.args is the arguments to ml_train
%              Model.model is the predictive model returned by the appropriate ml_train* function
%              where applicable,
%                 Model.model.b is the bias
%                 Model.model.w is the weights
%
% Examples:
%   size(trials) -> [160,10]; size(targets) -> [160,1]
%   Model = ml_train({trials,targets},{'lda'}): learn LDA predictor from some training data
%   Model = ml_train({trials,targets},{'gmm' 0.4 'xxx'}): learn GMM predictor with some additional parameters
%   Model = ml_train({{trials1,trials2,trials3}},{'svm'}): learn SVM predictor for three-class data in trials1/2/3
%   Model = ml_train({trials,[]}, {'gmm'}): learn GMM density estimator
%
% See also:
%   ml_predict()
%
% Notes:
%   Weighted learning is only supported by a fraction of the ml_train* functions.
%
% References:
%   [1] Bishop, C. M. "Pattern Recognition and Machine Learning." 
%       Information Science and Statistics. Springer, 2006.
%   [2] Hastie, T., Tibshirani, R., and Friedman, J. H. "The elements of statistical learning (2nd Ed.)." 
%	    Springer, 2009.
%   [3] MacKay, D. J. C. "Information theory, inference, and learning algorithms." 
%       Cambridge University Press, 2003.
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-03

if ~iscell(args)
    args = {args}; end

if ~isempty(varargin) 
    args = cat(2,args,varargin);  end

if isa(args{1},'function_handle') && isequal(1,strmatch('ml_train',char(args{1})))
    args{1} = strrep(char(args{1}),'ml_train',''); end

if ~ischar(args{1})
    error('The first element in Arguments must refer to a ml_train* function. either as a string (the * portion) or a function handle.'); end

if ~iscell(data)
    error('Data must be a cell array.'); end

if length(data) == 3
    % we have the {Trials,Targets,Weights} data format
    % this is passed on as ml_train*(trials,{targets,weights}, ...)
    trials = data{1};
    targets = data(2:3);
elseif length(data) == 2
    % we have the {Trials,Targets} data format
    % this is passed on as ml_train*(trials,targets, ...)
    trials = data{1};
    targets = data{2};
elseif length(data) == 1 && iscell(data{1})
    % we have a pack of per-class cell arrays: concatenate & construct labels
    trials = cat(1,data{1}{:});
    for k=1:length(data{1})
        targets{k} = k*ones(size(data{1}{k},1),1); end
    targets = cat(1,targets{:});
else
    % unknown data specification
    error('Data is specified in an unsupported format.');
end

if isempty(targets)
    model.model = feval(['ml_train' args{1}],trials,args{2:end});
else
    model.model = feval(['ml_train' args{1}],trials,targets,args{2:end});
end
model.args = args;

