function model = ml_traindal(trials,targets,lam,varargin)
% Model = ml_traindal(Trials, Targets, Lambda, Options...)
% Learn a linear probabilistic classification/regression model using the Dual-Augmented Lagrangian (DAL) method.
%
% The Dual-Augmented Lagrangian method [1] is an efficient and very robust approach to learning regularized linear classifiers or regressors,
% particularly for "noisy" biosignals. The regularization is very effective, so that a large number of features (e.g. every channel and time point)
% can be supplied for learning. Assumptions about how features are correlated or independent w.r.t. each other can (and should) be incorporated,
% by specifying the appropriate type of regularizer. If features are known to be all uncorrelated (e.g. derived from indepenent components), 'l1'
% is the appropriate regularizer. If features are correlated only within blocks, 'glr'/'glc' (group lasso by rows or columns) is the appropriate regularizer, 
% e.g. time points of concatenated independent components. If features are all correlated, but it is understood that there when arranged in a feature matrix, 
% the correlation structure would be low-rank, then 'ds' is the appropriate regularizer. The 'ds' mode is ideally suited when features are a matrix of 
% time-points by channels, where both time-points are mutually correlated and channels are, too, or when the features are covariance matrices, etc.
%
% To inform the classifier of the block size for 'glr'/'glc' or the matrix shape for 'ds', the trials should either be supplied as a 3d matrix (i.e. feature 
% matrices instead of feature vectors), or supplied in the regular fashion (2d matrix of feature vectors), but with the intended feature matrix shape
% specified in the 'shape' option. A few methods with different performance/accuracy/datasize tradeoffs are supplied. An important consideration when
% using DAL is that the data must be appropriately normalized for the method to be most effective, that is, it must be normalized across features and/or groups
% in 'l1' and 'glc'/'glr' modes, and normalized across both the horizontal and vertical axes of the feature matrix, and/or across blocks of a block-diagonal feature
% matrix, in the 'ds' mode.
%
% BCI paradigms which make extensive use of this classifier, according to [2], are provided in the paradigms/para_dal* functions. Among the methods provided
% in the toolbox, DAL is likely the best applicable method if the data is linearly separable (albeit not necessarily the easiest to use).
%
% In:
%   Trials       : training data, as in ml_train
%                  in addition, it may be specified as UxVxN 3d matrix,
%                  with UxV-formatted feature matrices per trial (N trials)
%
%   Targets      : target variable, as in ml_train
%
%   Lambda       : regularization parameter; a comprehensive search interval would be 2.^(-6:0.5:10)
%
%   Options  : optional name-value parameters to control the training details:
%              'loss': loss function to be used,
%                        'squared' for regression
%                        'logistic' for classification (default)
%
%              'regularizer': type of regularization to use:
%                             'l1': l1-norm on the features, gives sparse results
%                             'glr': grouped l1 norm ("group lasso"), gives blockwise sparse results
%                                   (groups the rows of the feature matrices)
%                             'glc': grouped l1 norm ("group lasso"), gives blockwise sparse results
%                                   (groups the columns of the feature matrices)
%                             'ds': dual spectral norm, gives low-rank results (default)
%
%              'shape': if trials is a NxF 2d matrix of vectorized matrices, 
%                           this is the dimensions of the matrices (default: Fx1)
%                       if trials is specified as an UxVxN 3d matrix, shape defaults to 
%                           [U,V] and trials are vectorized into the regular [N, U*V]
%                       if shape is specified with one of the values being NaN, 
%                           that value is set so that prod(shape)==F==U*V
%
%              misc parameters:
%              'scaling': pre-scaling of the data (see hlp_findscaling for options) (default: 'std') 
%
%              'bias': whether to include a bias term (default: 1)
%
%              'quiet': whether to suppress diagnostic outputs (default: 1)
%
%              'solver': solver to be used:
%                         'cg'   : Newton method with preconditioned conjugate gradient descent (default)
%                         'qn'   : Quasi-Newton method
%                         'nt'   : Newton method with Cholesky factorization
%                         'ntsv' : Newton method with memory saving (slightly slower)
%
% Out:
%   Model   : a predictive model
%
% See also:
%   ml_predictdal()
%
% References:
%  [1] Ryota Tomioka & Masashi Sugiyama, "Dual Augmented Lagrangian Method for Efficient Sparse Reconstruction", 
%      IEEE Signal Proccesing Letters, 16 (12) pp. 1067-1070, 2009.
%  [2] Ryota Tomioka and Klaus-Robert Mueller, "A regularized discriminative framework for EEG analysis with application to brain-computer interface",
%      Neuroimage, 49 (1) pp. 415-432, 2010. 
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-06-25

if ndims(trials) == 3
    shape = [size(trials,1) size(trials,2)];
    trials = double(reshape(trials,[],size(trials,3))');
else
    shape = [size(trials,2) 1];
end

% read parameters
opts = hlp_varargin2struct(varargin, 'shape',shape, 'loss','logistic', 'regularizer','ds', 'scaling','std', 'bias',1, 'quiet',1, 'solver','cg');
if ~exist('lam','var') || isempty(lam) lam = 1; end

% set NaN shape parameters appropriately
nf = size(trials,2);
ni = isnan(opts.shape);
if any(ni)
    opts.shape(ni) = nf / opts.shape(~ni); end
% check shape consistency
if nf ~= opts.shape(1)*opts.shape(2)
    error('shape parameter is inconsistent with feature space dimension.'); end

classes = unique(targets);
if length(classes) > 2
    % in this case we use the voter
    model = ml_trainvote(trials,targets, @ml_traindal, @ml_predictdal, lam, varargin{:});
else
    % not a bottleneck
    warning off PROPACK:NotUsingMex;
    
    % scale the data
    sc_info = hlp_findscaling(trials,opts.scaling);
    trials = hlp_applyscaling(trials,sc_info);
    
    % remap target labels to -1,+1
    targets(targets==classes(1)) = -1;
    targets(targets==classes(2)) = +1;

    % rewrite the bias to the format expected by dal
    opts.bias = fastif(opts.bias,0,[]);
    
    % learn
    if strcmp(opts.loss,'logistic')
        if strcmp(opts.regularizer,'ds')
            % dual-spectral logistic regression
            [model.w,model.b] = dallrds(zeros(opts.shape), opts.bias, double(trials), double(targets), lam, 'blks',[],'display',~opts.quiet,'solver',opts.solver);            
        elseif strcmp(opts.regularizer,'glc')
            % group-regularized logistic regression
            [model.w,model.b] = dallrgl(zeros(opts.shape), opts.bias, double(trials), double(targets), lam, 'blks',[],'display',~opts.quiet,'solver',opts.solver);            
        elseif strcmp(opts.regularizer,'glr')
            % group-regularized logistic regression, transposed data
            trials = double(reshape(trials',opts.shape(1),opts.shape(2),[]));
            ntrials = zeros(opts.shape(2),opts.shape(1),size(trials,3));
            for t=1:size(trials,3)
                ntrials(:,:,t) = trials(:,:,t)'; end
            trials = double(reshape(ntrials,[],size(ntrials,3))'); 
            [model.w,model.b] = dallrgl(zeros(opts.shape([2 1])), opts.bias, double(trials), double(targets), lam, 'blks',[],'display',~opts.quiet,'solver',opts.solver);            
        elseif strcmp(opts.regularizer,'l1')
            % sparse logistic regression
            [model.w,model.b] = dallrl1(zeros(nf,1), opts.bias, double(trials), double(targets), lam, 'display',~opts.quiet,'solver',opts.solver);
        else
            error('Unsupported regularizer.');
        end
    elseif strcmp(opts.loss,'squared')
        model.b = 0;
        if strcmp(opts.regularizer,'ds')
            % dual-spectral regularized regression
            model.w = dalsqds(zeros(opts.shape), [], double(trials), double(targets), lam, 'blks',[],'display',~opts.quiet,'solver',opts.solver);            
        elseif strcmp(opts.regularizer,'glc')
            % group LASSO
            model.w = dalsqgl(zeros(opts.shape), [], double(trials), double(targets), lam, 'blks',[],'display',~opts.quiet,'solver',opts.solver);            
        elseif strcmp(opts.regularizer,'glr')
            % group LASSO, transposed data
            trials = double(reshape(trials',opts.shape(1),opts.shape(2),[]));
            ntrials = zeros(opts.shape(2),opts.shape(1),size(trials,3));
            for t=1:size(trials,3)
                ntrials(:,:,t) = trials(:,:,t)'; end
            trials = double(reshape(ntrials,[],size(ntrials,3))');
            model.w = dalsqgl(zeros(opts.shape([2 1])), [], double(trials), double(targets), lam, 'blks',[],'display',~opts.quiet,'solver',opts.solver);            
        elseif strcmp(opts.regularizer,'l1')
            % LASSO
            model.w = dalsql1(zeros(nf,1), [], double(trials), double(targets), lam, 'display',~opts.quiet,'solver',opts.solver);            
        else
            error('Unsupported regularizer.');
        end
    else
        error('Unsupported loss function.');
    end
    
    model.w = model.w(:);    
    model.transpose = strcmp(opts.regularizer,'glr');
    model.classes = classes;
    model.sc_info = sc_info;
    model.opts = opts;
       
end
