function model = ml_traingauss(trials, targets, varargin)
% Model = ml_trainlda(Trials, Targets, Lambda, Options...)
% Learn a predictive model by means of a robust Gaussian Bayes classifier (with automatic feature selection).
%
% In:
%   Trials       : training data, as in ml_train
%
%   Targets      : target variable, as in ml_train; 
%                  can include a per-trial weighting (if a cell array {Targets,Weights})
%
%   Options...   : optional name-value pairs, with possible names:
%                  'dimensions': if given, reduce data dimensionality to this number of dimensions using a strong feature selector (default: [])
%                  'blend' : covariance blending parameter, blends covariance matrices (default: 1 = all covs identical)
%                            can be selected by a parameter search, e.g., by specifying search(0:0.1:1)
%                  'normprobs': produce normalized probabilities (0/1) (default: 1)
%                  'scaling': data pre-scaling (default: 'center')
%
% Out:
%   Model   : a predictive mdoel
%
% Examples:
%   model = ml_traingauss(trials,labels): learn a simple model...
%   
% See also:
%   ml_predictgauss()
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-03


opts = hlp_varargin2struct(varargin,'dimensions',[],'blend',1,'scaling','center','normprobs',1,'startcost',2);
model.opts = opts;

% scale the data
model.sc_info = hlp_findscaling(trials,opts.scaling);
trials = hlp_applyscaling(trials,model.sc_info);

% select features
if ~isempty(opts.dimensions)
    % start with some fairly high cost
    cost = opts.startcost;
    while 1
        model.selector = ml_trainlogreg(trials,targets,cost,'variant','l1','scaling','');
        w = sum(abs(model.selector.w(:,1:end-1)));
        % reduce cost until we have at last #dimensions non-zero weights
        if nnz(w) >= opts.dimensions
            break; end
        cost = cost/2;        
    end
    [dummy,I] = sort(w,'descend'); %#ok<ASGLU>
    model.feature_mask = full(sparse(1,I(1:opts.dimensions),true,1,length(I)));
    model.feature_weight = w;
    trials = trials(:,model.feature_mask);
else
    model.feature_mask = true(1,size(trials,2));
    model.feature_weight = ones(1,size(trials,2));
end

% obtain weights
if iscell(targets)
    [targets,weights] = deal(targets{:});
else
    weights = ones(size(targets,1),1);
end

% estimate the distribution of each class, using a weighted estimator (with shrinkage)...
model.classes = unique(targets);
for c=1:length(model.classes)
    w = weights(targets == model.classes(c));
    X = trials(targets == model.classes(c),:);
    model.mu{c} = mean_w(X,w);
    model.sigma{c} = cov_shrink(X,w);
end

% blend covariance matrices
if opts.blend ~= 0
    meansigma = mean(cat(3,model.sigma{:}),3);
    for c=1:length(model.classes)
        model.sigma{c} = (1-opts.blend)*model.sigma{c} + opts.blend*meansigma; end
end
