function model = ml_trainlogreg(trials,targets,lam,varargin)
% Model = ml_trainlogreg(Trials, Targets, Lambda, Options...)
% Learn a linear probabilistic predictive model by logistic regression.
%
% Logistic regression is the simplest fully probabilistic classifier in the toolbox, and can be understood as a linear projection followed by a
% mapping through a sigmoid (0-1) function, which allows to translate the input space smoothly into probabilities. Logreg is available in several
% variants, first the three variational Bayes (vb) ones, which are not dependent on a regularization parameter [1]. 'vb' and 'vb-iter' are about as
% versatile as LDA, except that they are fully probabilistic and not as outlier-prone; they differ mostly in numerical robustness. 'vb-ard' uses
% automatic relevance determination [2] to obtain a mapping that is sparse in the features, i.e. only a subset of the features is used for prediction.
% The Bayesian implementations compute not only a point estimate of the weights, but a full posterior distribution (gaussian), which enables more
% advanced post-processing of the model and outputs.
%
% The other two variants are computed via convex optimization. They might generally be somewhat more robust in extreme corner cases than the variational
% variants. Both require that a regularization parameter is specified, which can in simple cases be left at the default, but which is ideally optimized
% using parameter search, especially in the presence of many features and/or hard-to-handle distributions (more important for l1 than l2). This is
% computationally expensive, however. The 'l2' variant is the general-purpose one and in practice very similar to the vb/vb-iter variants.
% The 'l1' variant uses an l1-norm on the weights as regularizer and therefore gives a model that is sparse in the features (like vb-ard).
%
% Logistic regression is a very powerful, robust and versatile classifier for biosignals, and also very fast (when not regularized) - it is worth
% trying it on almost every problem that is hoped to be (more or less) linearly separable. Sparse logistic regression is a special case which allows
% to select features (out of exponentially many irrelevant ones) [3,4] on top of doing classification. EEG-derived feature are however in most cases
% not sparse, so that l1 or vb-ard may give worse results. Exceptions are features derived from independent components and other sparsity-inducing
% feature extractors. Some signal bases (e.g. wavelets) may be half-way in between sparse and non-sparse.
%
% The Bayesian variants are implemented using the bayes_logit toolbox by Jan Drugowitsch, and the regularized variants are implemented using
% the LIBLINEAR package [5] or CVX [6] as a fall-back.
%
% In:
%   Trials       : training data, as in ml_train
%
%   Targets      : target variable, as in ml_train
%                  may also contain weights, in the form {Targets,Weights}; supported by l1 & l2 variants
%
%   Lambda       : regularization parameter, for l1/l2 logistic regression variant (default: 1)
%                  a comprehensive search interval would be 2.^(-6:0.5:10)
%
%   Options  : optional name-value parameters to control the training details:
%              'variant': one of several logreg variants:
%                        'vb': variational Bayes (Jakkola&Jordan, 2000), using a joint uninformative
%                              prior on the weights, but possibly numerically unstable in corner cases (default)
%                        'vb-iter': variational Bayes (Jakkola&Jordan, 2000), iterative, using a per-weight prior,
%                                   more or less uninformative, but numerically more robust than the 'vb' variant
%                        'vb-ard': variational Bayes (Jakkola&Jordan, 2000), implements sparse logistic regression
%				       (using automated relevance determination)
%                        'l1' : sparse logistic regression using l1 regularization, using either LIBLINEAR
%                               or CVX
%                        'l2' : logistic regression using l2 regularization using LIBLINEAR
%
%              'eps'    : desired accuracy (default: [] - the default of the respective library, currently only supported for l1/l2)
%
%              'scaling': pre-scaling of the data (see hlp_findscaling for options) (default: 'std')
%
% Out:
%   Model   : a linear model;
%             w is the linear weights, b is the bias;
%             V is the covariance matrix of a posterior distribution around the weights (jointly over w and b)
%               (only available for the three 'vb' modes)
%             classes indicates the class labels which the model predicts
%             additional parameters determine a posterior distribution over the weights
%
% Examples:
%   model = ml_trainlogreg(trials,labels): learn a Bayesian logistic regression model using variational lower-bounding
%   model = ml_trainlogreg(trials,labels,1,'l2'): learn a l2-regularized logistic regression model using convex optimization
%   model = ml_trainlogreg(trials,labels,1,'l1'): learn a (sparse) l1-regularized logistic regression model using convex optimization
%
% See also:
%   ml_predictlogreg()
%
% References:
%   [1] Jaakkola, T. S., and Jordan, M. I. "A variational approach to bayesian logistic regression models and their extensions."
%       In Proceedings of the Sixth International Workshop on Artificial Intelligence and Statistics (1997).
%   [2] Wipf, D., Nagarajan, S., Platt, J., Koller, D., Singer, Y., and Roweis, S. "A New View of Automatic Relevance Determination."
%       MIT Press, 2008, pp. 1632, 1625.
%   [3] P. Zhao and B. Yu. "On model selection consistency of Lasso."
%       JMLR, 7:2541–2563, 2006.
%   [4] M. J. Wainwright. "Sharp thresholds for noisy and high-dimensional recovery of sparsity using l1-constrained quadratic programming."
%      Technical Report 709, Dpt. of Statistics, UC Berkeley, 2006.
%   [5] R.-E. Fan, K.-W. Chang, C.-J. Hsieh, X.-R. Wang, and C.-J. Lin. "LIBLINEAR: A library for large linear classification"
%       Journal of Machine Learning Research 9(2008), 1871-1874.
%   [6] M. Grant and S. Boyd. "Graph implementations for nonsmooth convex programs"
%       Recent Advances in Learning and Control (a tribute to M. Vidyasagar), V. Blondel, S. Boyd, and H. Kimura, editors, pages 95-110,
%       Lecture Notes in Control and Information Sciences, Springer, 2008.
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-04


% read parameters
if ~exist('lam','var') || isempty(lam) lam = 1; end
opts = hlp_varargin2struct(varargin, 'variant','vb', 'eps',[], 'scaling','std');

% obtain weights
if iscell(targets)
    [targets,weights] = deal(targets{:});
    weights = weights/sum(weights);
else
    weights = [];
end

% we have to use CVX for logreg, if the appropriate LIBLINEAR version is missing
need_fallback = ~isempty(strfind(which('train'),'pre7.3')) || (~isempty(weights) && isempty(which('llwtrain')));
use_cvx = any(strcmp(opts.variant,{'l1','l2'})) && need_fallback;

% identify and remap the classes
classes = unique(targets);
if ~isempty(strfind(opts.variant,'vb')) || use_cvx
    % these are two-class classifiers
    if length(classes) > 2
        % so we need to vote in this case
        model = ml_trainvote(trials,targets,@ml_trainlogreg,@ml_predictlogreg,lam,varargin{:});
        return;
    end
    % remap target labels to -1,+1
    targets(targets==classes(1)) = -1;
    targets(targets==classes(2)) = +1;
else
    % remap target labels to 0..k 
    targ = targets;
    for c=1:length(classes)
        targets(targ==classes(c)) = c-1; end
end

% scale the data
sc_info = hlp_findscaling(trials,opts.scaling);
trials = hlp_applyscaling(trials,sc_info);

% add bias variable to training data for the VB's
if ~isempty(strfind(opts.variant,'vb'))
    warning off Bayes:maxIter
    trials = [trials ones(size(trials,1),1)];
end

switch opts.variant
    case 'vb'
        [model.w, model.V, model.invV, model.logdetV] = bayes_logit_fit(trials,targets);
    case 'vb-iter'
        [model.w, model.V, model.invV, model.logdetV] = bayes_logit_fit_iter(trials,targets);
    case 'vb-ard'
        [model.w, model.V, model.invV, model.logdetV] = bayes_logit_fit_ard(trials,targets);
    case 'l1'
        lam = 1/lam; % lambda is parameterized differently in LIBLINEAR, we adapt to that
        if ~isempty(weights)
            try
                % LIBLINEAR-weights 1.6+
                model = llwtrain(double(weights),double(targets),sparse(double(trials)),[...
                    '-s 6 -c ' num2str(lam) ' -B 1' fastif(~need_fallback,' -q ','') ...
                    fastif(~isempty(opts.eps),[' -e ' num2str(opts.eps)],'')]);
            catch
                % CVX fallback
                cvx_begin
                    variables W(size(trials,2)) b
                    minimize(sum(weights.*log(1+exp(-targets.*(trials*W+b))))*lam + norm(W,1))
                cvx_end                
                model = struct('w',W,'b',b);
            end
        else
            try
                % LIBLINEAR 1.51+, MATLAB 7.3+
                model = train(targets,sparse(double(trials)),[...
                    '-s 6 -c ' num2str(lam) ' -B 1' fastif(~need_fallback,' -q ','') ...
                    fastif(~isempty(opts.eps),[' -e ' num2str(opts.eps)],'')]);
            catch
                % CVX fallback
                cvx_begin
                    variables W(size(trials,2)) b
                    minimize(sum(log(1+exp(-targets .* (trials*W+b))))*lam + norm(w,1))
                cvx_end                
                model = struct('w',W,'b',b);
            end
        end
        
    case 'l2'
        lam = 1/lam;
        if ~isempty(weights)
            try
                % LIBLINEAR-weights 1.6+
                model = llwtrain(double(weights),double(targets),sparse(double(trials)),[...
                    '-s 0 -c ' num2str(lam) ' -B 1' fastif(~need_fallback,' -q ','') ...
                    fastif(~isempty(opts.eps),[' -e ' num2str(opts.eps)],'')]);
            catch
                % CVX fallback
                cvx_begin
                    variables W(size(trials,2)) b
                    minimize(sum(weights.*log(1+exp(-targets.*(trials*W+b))))*lam + norm(W,2))
                cvx_end                
                model = struct('w',W,'b',b);
            end
        else
            try
                % LIBLINEAR 1.33+
                model = train(targets,sparse(double(trials)),[...
                    '-s 0 -c ' num2str(lam) ' -B 1' fastif(~need_fallback,' -q ','') ...
                    fastif(~isempty(opts.eps),[' -e ' num2str(opts.eps)],'')]);
            catch
                % CVX fallback
                cvx_begin
                    variables W(size(trials,2)) b
                    minimize(sum(log(1+exp(-targets .* (trials*W+b))))*lam + norm(w,2))
                cvx_end                
                model = struct('w',W,'b',b);
            end
        end
end

model.classes = classes;
model.sc_info = sc_info;
model.variant = opts.variant;
