function model = ml_trainsvmlight(trials, targets, cost, varargin)
% Model = ml_trainsvmlight(Trials, Targets, Cost, Options...)
% Learn a linear or non-linear predictive model by Support Vector Machines, as implemented in the SVMlight package.
%
% SVMlight [1] is a comprehensive and fast implementation of Support Vector Machines [2] that supports, in addition to classification,
% also regression and ranking; it further offers a variety of kernels for state-of-the-art non-linear prediction. An alternative implementation
% which offers a greater variety of loss measures for classification is SVMperf (see ml_trainsvmperf). For further details, see ml_trainsvmperf.
%
% For non-linear ranking problems, this is the online applicable method in the toolbox, for non-linear regression problems, the only alternative
% is the Relevance Vector Machine, and for (arbitrary) non-linear classification problems, the alternatives is the Relevance Vector Machine and SVMperf.
%
% In:
%   Trials   : training data, as in ml_train
%
%   Targets  : target variable, as in ml_train
%
%   Cost     : regularization parameter, reasonable range: 2.^(-5:2:15), greater is stronger
%              default: [avg. x*x]^-1
%   Options  : optional name-value parameters to control the training details, see below in code for full help
%               'type': classification (c), regression (r), and ranking (p), default: c
%               'tube': epsilon width of tube for regression (default 0.1)
%               'balance': positive weight / negative weight
%              kernel parameters:
%               'kernel': type of kernel function (lin,poly,rbf,user)
%               'd': parameter d in polynomial kernel (int)
%               'g': parameter gamma in rbf kernel (float)
%               's': parameter s in sigmoid/poly kernel (float)
%               'r': parameter c in sigmoid/poly kernel (float)
%               'u': parameter of user-defined kernel (string)
%              misc options:
%               'eps': tolerance (e.g., 0.1)
%               'bias': bias present? (0,1)
%               'clean': clean inconsistent data before start? (0,1)
%               'quiet': quiet mode (0,1)
%               'scaling': pre-scaling, see hlp_findscaling (default: 'std')
%
% Out:
%   Model   : a linear model; 
%             classes indicates the class labels which the model predicts
%             sc_info is the scaling info
%
% Examples:
%   model = ml_trainsvmlight(trials,labels,1): learn a SVM model
%   
% See also:
%   ml_predictsvmlight()
%
% References:
%   [1] Thorsten Joachims, "Learning to Classify Text Using Support Vector Machines"
%       Dissertation, Kluwer, 2002.
%   [2] Schlkopf, B., and Smola, A. "Learning with Kernels: Support Vector Machines, Regularization, Optimization, and Beyond"
%       (Adaptive Computation and Machine Learning). The MIT Press, Dec. 2001.
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-04

% read parameters
opts = hlp_varargin2struct(varargin, ...
    ... % --- key options ---
    'type','c',...              % classification (c), regression (r), and ranking (p), default: c
    'tube',0.1, ...             % epsilon width of tube for regression (default 0.1)
    'balance', 1, ...           % Cost: cost-factor, by which training errors on positive examples outweight errors on negative examples (default 1)
    'bias', [], ...              % bias enabled? (0/1, default:1, only for lin kernel)
    ... % --- kernel options ---
    'kernel',0, ...              % type of kernel function:
    ...                          %   * 0/'lin': linear (default)
    ...                          %   * 1/'poly': polynomial (s a*b+c)^d
    ...                          %   * 2/'rbf': radial basis function exp(-gamma ||a-b||^2)
    ...                          %   * 3/'sig': sigmoid tanh(s a*b + c)
    ...                          %   * 4/'user': user defined kernel from kernel.h        
    'd',[], ...                  % parameter d in polynomial kernel (int)
    'g',[], ...                  % parameter gamma in rbf kernel (float)
    's',[], ...                  % parameter s in sigmoid/poly kernel (float)
    'r',[], ...                  % parameter c in sigmoid/poly kernel (float)
    'u',[], ...                  % parameter of user-defined kernel (string)
    ... % --- misc options ---
    'scaling', 'std', ...        % scaling mode (see hlp_findscaling)
    'clean',0, ...               % remove inconsistent training examples and retrain (default 0)
    'quiet', 1, ...              % verbosity level
    'eps',0.1...                 % epsilon: allow that tolerance for termination criterion (default 0.1)
     );

if ~exist('cost','var') cost = []; end
                            
% find the class labels
classes = unique(targets);
if length(classes) > 2
	% in this case we use the voter
    model = ml_trainvote(trials,targets,@ml_trainsvmlight,@ml_predictsvmlight,cost,varargin{:});
else    
    % scale the data
    sc_info = hlp_findscaling(trials,opts.scaling);
    trials = hlp_applyscaling(trials,sc_info);
    
    % remap target labels to -1,+1
    targets(targets==classes(1)) = -1;
    targets(targets==classes(2)) = +1;
    
    % build the command
    command = ['-z ' num2str(opts.type)];
    if ~isempty(cost) command = [command ' -c ' num2str(cost)]; end
    command = [command fastif(opts.quiet,' -v 0',' -v 1')];
    command = [command ' -w ' num2str(opts.tube)];
    command = [command ' -j ' num2str(opts.balance)];
    if ~isempty(opts.bias) command = [command ' -b ' num2str(opts.bias)]; end
    command = [command ' -i ' num2str(opts.clean)];
    command = [command ' -e ' num2str(opts.eps)];
    if ischar(opts.kernel)
        mapping = struct('lin',0,'poly',1,'rbf',2,'sig',3,'user',4);
        command = [command ' -t ' mapping.(opts.kernel)];
    else
        command = [command ' -t ' num2str(opts.kernel)];
    end
    if ~isempty(opts.d) command = [command ' -d ' num2str(opts.d)]; end
    if ~isempty(opts.g) command = [command ' -g ' num2str(opts.g)]; end
    if ~isempty(opts.s) command = [command ' -s ' num2str(opts.s)]; end
    if ~isempty(opts.r) command = [command ' -r ' num2str(opts.r)]; end
    if ~isempty(opts.u) command = [command ' -u ' opts.u]; end
    
    % run the command
    model = svmlearn(trials,targets,command);
    model.sc_info = sc_info;
    model.classes = classes;
end
