function model = ml_trainvote(trials, targets, learner, predictor, varargin)
% Model = ml_trainvote(Trials, Targets, Binary-Learner, Binary-Predictor, Arguments...)
% Learn a multinomial classification model using a binary learning function and optional arguments, implemented as a 1-vs-1 voting arrangement.
%
% Voting is a simple meta-algorithm that allows to use binary learners to learn a predictive model over multiple classes; Most ml_train* functions
% make implicit use of this function when the input has more than two classes. In the implemented 1-vs-1 voting schema, a binary predictive model is 
% computed for every pair of classes, and during prediction, the probabilities that the predictors assign for class i are summed, for every i, and finally
% re-normalized.
%
% In:
%   Trials       : training data, as in ml_train
%
%   Targets      : target variable; should be a Nx1 array
%
%   Binary-Learner : binary learning function, e.g. one of the ml_train* functions
%                    accepts Trials & Targets as its first two arguments and optionally further
%                    arguments which can be supplied as Learner-Arguments
%
%   Binary-Predictor : binary prediction function, e.g. the corresponding ml_predict* function
%                      accepts Trials as its first argument and a model as second argument.
%           
%   Learner-Arguments : list of further arguments to be passed through to the learner
%                       (the first 2 are a subset of Trials and Targets)
%
% Out:
%   Model   : a voted model
%             .classes is the target values per class
%             .voted is a cell array of binary models
%             .opts is the options
%
% Examples:
%   model = ml_trainvote(trials,labels,@ml_trainlda): learn a voted arragement of LDA classifiers, supporting multiple classes
%   
% See also:
%   ml_predictvote()
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-06-25

% first get the classes for training
model.classes = unique(targets);

for i=1:length(model.classes)
    for j=i+1:length(model.classes)
        % learn an i-vs-j model
        subset = targets == model.classes(i) | targets == model.classes(j);
        if ndims(trials) == 2
            % feature vectors
            model.voted{i,j} = learner(trials(subset,:), targets(subset), varargin{:});
        else
            % feature matrices
            model.voted{i,j} = learner(trials(:,:,subset), targets(subset), varargin{:}); 
        end
    end
end

% and remember the options
model.opts = [{learner, predictor} varargin];