function onl_loaddetector(id, det)
% onl_loaddetector(Identifier, Model-or-Filename)
% Load the given predictive model (here called detector) and associate it with some identifier (by which it can be referenced in onl_predict).
%
% The function onl_loaddetector is one of four functions in BCILAB's online processing Application Programming Interface (online API). The four functions are 
% onl_setup, onl_loaddetector, onl_feed, and onl_predict. For a complete example of using the online interface, see onl_setup. onl_loaddetector is used to
% link a copy of a predictive model (as computed by bci_train) to an online stream, so that it can subsequently be used to make predictions online.
% 
% Each online stream (currently there is only the global one) maintains a list of its linked detectors, which are identified by a user-specified name (id).
% onl_loaddetector is used to add, remove or reset detectors from this list. Once a detector is loaded or reset, it will initially have empty internal data buffers,
% and the detector may only produce predictions after a sufficient amount of data has been fed into the online stream after loading the detector (the online 
% stream maintains no arbitrary 'backlog' of data independently of any loaded detectors).
%
% The offline/online processing system of BCILAB is designed to ensure that the outputs of a given model w.r.t. some data are identical, whether they
% are obtained offline (e.g. via bci_predict, or implicitly via bci_train) or online (including via onl_stream, which is "pseudo-online"). The code of all
% components (signal processing, feature extraction, machine learning, meta algorithms, paradigms) is identical across offline and online; the only difference
% is the granularity (block size) at which data is being fed into the components, and the execution environment in which this code runs (e.g. bci_train/bci_predict
% vs. onl_predict). Nevertheless, as it is each component's responsibility to give identical results under any block size, it is possible that some models
% produce outputs that differ between offline and online use (due to bugs); for this reason, it is recommended to test a paradigm configuration using bci_predict 
% and onl_stream, and to compare results (taking care of offsets) before a critical online operation (or publication). Once the code for a given paradigm has 
% been shown to work, it is usually not necessary to test its consistency before every online use as long as no code changes have been made.
%
% Aside from these considerations, it is possible to design components, paradigms and models which are fundamentally non-causal (and thus can not be used online), 
% but which nevertheless may be of scientific value (e.g. for offline predicitions of some latent variable(s) given an EEG dataset). One example is the 
% filter flt_envelope, which needs to look into the future to not delay the signal or, likewise, flt_fir in the 'zero-phase' mode.
%
% In:
%   Identifier        : MATLAB-conformant identifier of the used detector, for semantic distinction, e.g. 'lr' or 'lr_spec' or 'emg'
%
%   Model-or-Filename : detector model structure as returned by bci_train, or alternatively the name of a file that contains a model.
%                       (assuming that the file contains a single variable - otherwise an error will be thrown), or a {Filename, Structname} cell array. 
%                       If specified as [] and if a detector model was previously loaded under the same Identifier, it is deleted by this call.
%
% Examples:
%   onl_loaddetector('mybci', mydetectorstruct); 
%   onl_loaddetector('mybci', 'classifier-subject10.mat');
%   onl_loaddetector('mybci', {'somedetectors.mat', 'specdet'});
%   onl_loaddetector('mybci', []);
% 
% See also:
%   onl_setup, onl_feed, onl_predict
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-06-21

global bcilab;

if ~exist('det','var') 
    error('No detector specified.'); end
if ~isvarname(id)
    error('The first argument must be MATLAB-conformant identifier.'); end

if ~isempty(det)

    if ischar(det)
        % load from file
        contents = io_load(det);
        fn = fieldnames(contents);
        if length(fn) ~= 1
            error('BCILAB:onl_loaddetector:ambiguous_file',['The file ' det ' must contain exactly one variable (namely the detector).']); end
        % obtain the variable
        det = contents.(fn{1});
    elseif iscell(det) && length(det)==2
        % load from file, with variable name specified
        try
            det = getfield(io_load(det{1},det{2}),det{2}); %#ok<GFLD>
        catch
            error('BCILAB:onl_loaddetector:load_error',['The file ' det{1} ' and/or its detector ' det{2} ' could not be loaded.']);
        end
    end
    
    % load into online system
    try
        bcilab.onl_detectors.(id) = struct('det',{det},'pipeline',{det.bcilab.online_expression});
    catch
        env_handleerror(lasterror); %#ok<LERR>
        error('BCILAB:onl_loaddetector:unexpected',['Detector ''' id ''' has an unexpected structure.']);
    end
    
else
    % erase from online system
    try
        bcilab.onl_detectors = rmfield(bcilab.onl_detectors, id);
    catch
        disp(['No detector named ''' id ''' is loaded.']);
    end;
end
