function y = onl_predict(id)
% Result = onl_predict(Identifier)
% Obtain predictions of a previously loaded detector (predictive model) given the online stream's data up to and including the most recently fed sample.
%
% The function onl_predict is one of four functions in BCILAB's online processing Application Programming Interface (online API). The four functions are 
% onl_setup, onl_loaddetector, onl_feed, and onl_predict. For a complete example of using the online interface, see onl_setup. onl_predict is used to compute
% predictions w.r.t. to the most recent sample in the buffer. 
%
% If the respective detector was loaded before data was fed into the buffer, onl_predict implements a strong consistency model: the detector's output 
% for any given sample i (counting from the first sample that was fed) is identical to what bci_predict, bci_train or onl_feed would produce for sample 
% #i on a recorded dataset that begins with the first supplied sample after the onl_setup call (except if there is an error in one of the filter 
% components used by the model). 
%
% If a new detector is loaded into a stream which has already been fed samples and on which predictions have already been made, onl_predict implements by default a 
% best-effort model: as much previously supplied data as is available in the stream's buffer will be used for models that include infinite 'memory' 
% (e.g., IIR filters). Models that require a fixed epoch of data from the past will produce empty outputs up to the point where that data is 
% available (whereas the user can produce dummy control signals in this case if necessary). If strong consistency is absolutely required, a dummy detector can be 
% loaded after the setup call, which is never queried. Since the online system generally buffers data until it has been "consumed" by all loaded models, 
% all data from the beginning of the stream will be available to detectors that are loaded late (at the expense of an indefinitely growing buffer of raw data).
%
% Data processing in the online system happens only when it is needed, and is therefore implemented in a pull model: a call to onl_predict triggers a data 
% processing requests, which may lead to further requests into lower pipeline stages. For models with infinite memory which are only sporadically 
% queried, this may involve on-demand processing of large amounts of data and incur significant delay. A strategy to avoid uncontrolled delays for such models
% is to let onl_predict pre-compute intermediate results by periodically querying it, even if the results are not needed. 
%
% The function onl_stream can be used to test the behavior of the online system offline on pre-recorded data.
%
% In:
%   Identifier : name of the detector in to be used; if empty, a struct with results
%                for all loaded detectors is returned. (default: empty)
%
% Out:
%   Result : Predictions of the selected detector(s) w.r.t. to the most recent data.
%
% Example:
%   See onl_stream() for a simulated online use case.
%
%                                        Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                                        2010-04-03

global bcilab captured_output;
if ~exist('id','var')
    id = []; end

% do some consistency checks
if ~isfield(bcilab,'onl_buffer')
    error('ext_setup() must be called prior to onl_predict(), to set up the stream format.'); end
if ~isfield(bcilab,'onl_detectors')
    error('ext_loaddetector() must be called prior to onl_predict(), to load a detector.'); end

% if the expression mechanism is not disabled...
if ~isequal(1,exp_eval(@disable_expressions))
    % ... then do it now, for the duration of onl_predict (we don't want to run this code for every sample)
    y = utl_block({exp_rule(@disable_expressions,1)}, exp_lambda(@()onl_predict(id)));
else
    try
        % update & predict, either from the specified detector or from all loaded detectors (with captured console output)
        if ~isempty(id)
            [captured_output,y] = evalc('do_predict(id)');
        else
            for d = fieldnames(bcilab.onl_detectors)'
                [captured_output,y.(d{1})] = evalc('do_predict(d{1})'); end
        end
        
        % discard now-obsolete contents of the raw data buffer: first find all pipeline nodes representing raw data...
        raw_nodes = structfun(@(x)find_nodes(x.pipeline,@rawdata),bcilab.onl_detectors);
        if cellfun(@(node)isfield(node,'buffer'), struct2cell(raw_nodes))
            % ... but proceed only if all have a buffer (which is only the case when all have been invoked at least once)
            % the smallest smax of all nodes is the point up to which we can discard past data from the global buffer
            min_smax = min(cellfun(@(node)node.buffer.smax, raw_nodes));
            
            % this requires a write lock to be "timer-safe" (-> timers not yet allowed)
            bcilab.onl_buffer.data(:,1:(min_smax-bcilab.onl_buffer.smin+1)) = [];
            bcilab.onl_buffer.pnts = size(bcilab.onl_buffer.data,2);
            bcilab.onl_buffer.smin = (bcilab.onl_buffer.smax+1) - bcilab.onl_buffer.pnts;
            bcilab.onl_buffer.xmin = (bcilab.onl_buffer.smin-1)/bcilab.onl_buffer.srate;
        end
    catch
        disp('An exception occured inside onl_predict():');
        env_handleerror(lasterror); %#ok<LERR>
        y = NaN;
    end
end


% compute the prediction for a given detector
function y = do_predict(id) %#ok<DEFNU>
global bcilab; 
det = bcilab.onl_detectors.(id).det;
% update the detector's filter pipeline
bcilab.onl_detectors.(id).pipeline = update_pipeline(bcilab.onl_detectors.(id).pipeline, det.bcilab.window_length, bcilab.onl_buffer);
% check that we have enough data
if det.bcilab.window_length == size(bcilab.onl_detectors.(id).pipeline.buffer.data,2)
    % invoke the detector's prediction function
    if is_stateful(det.bcilab.prediction_function,[],[])
        [y,bcilab.onl_detectors.(id).det] = det.bcilab.prediction_function(bcilab.onl_detectors.(id).pipeline.buffer,'prediction','model',det);
    else
        y = det.bcilab.prediction_function(bcilab.onl_detectors.(id).pipeline.buffer,'prediction','model',det);
    end
    if isnumeric(y) && isa(y,'single')
        y = double(y); end
else
    y = NaN;
end


% update the given filter pipeline from (possibly new) raw data, requesting a specified number of previous samples (0 if only new samples are requested) 
function p = update_pipeline(p,requested,raw)
% add potentially missing fields
if isequal(p,@rawdata)
    p = struct('head',{@rawdata}, 'parts',{{}}); end
if ~isfield(p,'buffer')
    p.buffer = struct('data',{[]}, 'smax',{raw.smin-1}); end
if ~isfield(p,'stateful')
    p.stateful = ~isequal(p.head,@rawdata) && is_stateful(p.head); end
if p.stateful && ~isfield(p,'state')
    p.state = p.parts{2}; end % we take the value of stage's state argument as the initial state

% update the input pipelines, and collect the arguments that we need to invoke the pipeline stage
args = p.parts;
for k = find(cellfun(@is_pipeline,p.parts))
    % if this is a stateful pipeline, only new samples are requested (requested=0), otherwise the number of requested samples is passed down
    p.parts{k} = update_pipeline(p.parts{k},requested * ~p.stateful,raw);
    % also, for those args that are pipelines, we take the buffers of the pipelines, rather than the structs themselves)
    args{k} = p.parts{k}.buffer;
end

% invoke the processing function of the pipeline to get an updated buffer
if isequal(p.head,@rawdata)
    % raw data pipeline: get the data directly from the raw buffer
    buffer = setfield(raw,'data', raw.data(:,end-(raw.smax-(p.buffer.smax+1)):end)); %#ok<SFLD>
else
    if p.stateful
        % stateful function: the second argument is, by contract, the state
        [buffer,p.state] = p.head(args{1}, p.state, args{3:end});
        if strcmp(p.state,'<noncausal>')
            error('A noncausal filter stage cannot be used in online predictions.'); end
    else        
        buffer = p.head(args{:});
    end
end

% if some other amount of data than just what's new has been requested...
if requested ~= 0    
    if size(buffer.data,2) >= requested
        % cut excess data from the buffer
        buffer.data = buffer.data(:,end-requested+1:end);
    else
        % stateful filters must concatenate the newly-produced outputs with the
        % data that they already have in the buffer from previous invocations
        if p.stateful
            if size(p.buffer.data,2) == requested
                % we can do an in-place update without reallocations
                buffer.data = [p.buffer.data(:,size(buffer.data,2)+1:end) buffer.data];
            else
                % append new samples & cut excess data
                buffer.data = [p.buffer.data buffer.data];
                if size(buffer.data,2) > requested
                    buffer.data = buffer.data(:,end-requested+1:end); end
            end
        end
    end
    % update metadata
    buffer.pnts = size(buffer.data,2);
    buffer.smax = raw.smax*(buffer.srate/raw.srate);
    buffer.smin = (raw.smax - (buffer.pnts-1))*(buffer.srate/raw.srate);
    buffer.xmin = buffer.xmax - (buffer.pnts-1)/buffer.srate;
end
% write back
p.buffer = buffer;


% find all nodes of a given pipeline/expression that have the given head
function cases = find_nodes(x,head)
cases = cellfun(@(p)find_nodes(p,head),x.parts(cellfun(@is_pipeline,x.parts)));
if isequal(x.head,head)
    cases = [x cases]; end


% check whether x is a pipeline
function r = is_pipeline(x)
r = all(isfield(x,{'head','parts'})) || (isa(x,'function_handle') && is_undefined_function(x));


% check whether the given function is stateful (i.e. has a second output), optionally with some additional arguments
function r = is_stateful(f,varargin)
try 
    [a,b] = f(varargin{:}); %#ok<NASGU,ASGLU>
catch
    e = lasterror; %#ok<LERR>
    r = ~any(strcmp(e.identifier,{'MATLAB:TooManyOutputs','MATLAB:maxlhs'})); 
end
