function onl_setup(srate, channames, varargin)
% onl_setup(Sampling-Rate, Channel-Names, Optional-Fields...)
% (Re-)initialize the online data stream and dependent processing pipelines.
%
% The function onl_setup is one of four functions in BCILAB's online processing Application Programming Interface (online API). The four functions are 
% onl_setup, onl_loaddetector, onl_feed, and onl_predict. The online API is what ultimately performs the online processing in BCILAB, whether it is 
% invoked from a BCILAB plugin for BCI2000, OpenViBE, MatRiver/DataRiver or from a custom standalone application, via the MATLAB Engine interface or as a 
% compiled library, or whether it is called directly from a MATLAB function which manages the data streaming directly.
%
% Online processing begins by opening an "online stream", which is a formatted multichannel signal with meta-data such as sampling rate, channel names (and optionally 
% further data as allowed by EEGLAB), and the actual data time series (which is initially empty). The online stream is the equivalent to a raw data file / data set
% in offline processing (and therefore contains largely similar meta-data). Currently, there is only one global data stream per MATLAB instance (which contains all 
% merged data, such as EEG, MoCap, etc.), and onl_stream is the function to open (or re-initialize) it. Sampling rate and channel names are mandatory arguments,
% and further other fields (in the EEGLAB naming scheme) can be specified as name-value pairs.
%
% Once a stream has been established, raw data can be fed into it (via the function onl_feed), detectors/predictive models that shall operate on it can be 
% loaded (via the function onl_loadddetector), and predictions from previously loaded models can be obtained, always w.r.t. the most recent data sample in the buffer
% (via the function onl_predict). Multiple models can operate on the same stream, and the rates at which data is fed and predictions are requested is arbitrary.
%
% Example:
%   % calibrate (or load) a model
%   calib = io_loadset('data:/projects/attention/mason.bdf');
%   [loss,model] = bci_train({'data',calib,'paradigm',@para_windowmeans},'epoch',[0 0.75],'events',{{'11','12'},'13'}, ...
%       'filter',[0.1 15],'wnds',[0.25 0.3;0.3 0.35;0.35 0.4; 0.4 0.45;0.45 0.5;0.5 0.55;0.55 0.6]);
%   
%   % open a new online stream
%   onl_setup(200,{'FC1','FCz','FC2','C5','C3','C1','Cz','C2','C4','C6','CP5','CP3','CP1','CPz','CP2','CP4','CP6','P3','Pz','P4','POz','O1','O2'});
%   % link the model to the online stream
%   onl_loaddetector('attention',model);
%
%   while 1
%       % obtain a new block of data from some acquisition system
%       datachunk = get_eeg_samples(h);
%       % feed the block into the online stream
%       onl_feed(datachunk);
%       % obtain predictions
%       outputs = onl_predict('attention');
%       % handle them
%       bar(outputs); drawnow; pause(0.1);
%   end
%
%
% In:
%   Sampling-Rate      : sampling rate of the new online stream
%
%   Channel-Names      : cell array of channel names or chanlocs data structure of the new online stream
%
%   Optional-Fields... : additional fields (e.g. icaweights, dipfit, etc.) to be added to the online stream
%                        (which is effectively an EEGLAB dataset)
%
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-03-28

global bcilab;

% (re-)initialize the raw data buffer
bcilab.onl_buffer = exp_eval(set_new(varargin{:}, 'data',[], 'srate',srate, 'chanlocs',channames, 'smin',1, 'smax',0, 'event',[], 'urevent',[], 'epoch',[], 'icaact',[]));

% bcilab.onl_detectors is the set of currently loaded detectors (together with their entire processing pipelines)
%  .det contains the detector
%  .pipeline the preprocessing pipeline (incl. buffer state)
if isfield(bcilab,'onl_detectors') && ~isempty(bcilab.onl_detectors)
    % reset the pipelines (to the pipeline descriptions of the respective detectors)
    for fn=fieldnames(bcilab.onl_detectors)'
        bcilab.onl_detectors.(fn{1}).pipeline = bcilab.onl_detectors.(fn{1}).det.bcilab.online_expression; end
end