function [predictions,predict_at] = onl_stream(eeg, det, varargin)
% [Predictions,Latencies] = onl_stream(In-Set, Model, Options...)
% Obtain predictions of cognitive state at specified time points in a given raw dataset, using a previously learned model.
% 
% The function onl_stream is the standard way in BCILAB to apply predictive models to datasets, i.e. to receive predictions of cognitive state
% at various specified time points in the given dataset. The dataset must be unfiltered, but it may have been structurally edited (e.g. chanlocs, dipfit, etc.),
% The model is usually one that has been previously computed using bci_train (see bci_train) on some calibration dataset (note that predictions on the 
% calibration dataset itself result in overly optimistic estimates, and are usually scientifically invalid results).
% 
% The time points at which the model shall be invoked can be specified in flexible ways: among others, a model can be invoked at regular intervals ('interval'),
% at certain specified latencies ('points'), at a given rate ('sampling_rate'), or relative to events of certain types ('markers'). Generally, a constant offset
% can be added to the specified time points, which is important when classifying relative to certain stimulus events: usually the data necessary for the prediction 
% is only fully available at some time after the stimulus event (as a rule of thumb: when a model was calibrated relative to events, using an epoch clause which
% ended x seconds past the event, then this is the amount of time that needs to be specified as the offset (converted in samples).
% 
% The predictions produced by onl_stream are of the same format as those produced by bci_predict and ml_predict (and may depend on how the model was calibrated), 
% see bci_predict for an explanation of the most common case(s).
%
% onl_predict uses the online infrastructure of the toolbox to produce its results (and can currently not be used while online processing is in progress in the current
% MATLAB instance). Therefore, the outputs are exactly identical to what would be computed by the model if it was applied in real time. The analysis done by 
% onl_stream is typically called "pseudo-online analysis", since the data is actually streamed (causally) through the processing chain in a simulated online fashion.
% In contrast, the related bci_predict does what is often called "offline analysis", in which the data is processed not necessarily causally in successive stages, 
% and predictions are obtained in a trial-by-trial fashion, not in a time-point-by-time-point fashion; offline analysis typically involves known desired outputs, 
% against which the predictions of the model are compared, to get loss estimates. The framework of bci_predict is designed to give identical results to what would be
% computed by onl_stream in the respective situations (e.g. at the time points of certain events), but the data processing is generally more efficient (for example,
% intermediate results can be cached in memory or on disk, whereas onl_stream usually processes the data in small chunks). For these reasons, onl_stream is the 
% preferred method when details of the time course of a model's output are required (e.g. how it behaves in the proximity of events or in areas of the data where
% unknown cognitive processes take place), and/or when it must be 100% certain that the model can be applied online as-is. bci_predict is the preferred method when
% high-throughput offline analyses on certain constrained portions of the data (e.g. relative to events) need to be executed as efficiently as possible (e.g. in batch
% analyses).
%
%
% Example:
%  % load calibration data and compute a model (see bci_train)
%  calib = io_loadset('datasets/mary/stresslevels_calib.eeg');
%  [loss,model] = bci_train({'data',calib,'paradigm',@para_speccsp},'events',{'low','medium','high'});
%
%  % apply pseudo-online, at at rate of 5x per second
%  testdata = io_loadset('datasets/mary/stresslevels_realworld.eeg');
%  [predictions,latencies] = onl_stream(testdata, model, 'sampling_rate',5);
%
%  % plot time course of the model's output
%  plot(latencies, predictions{2}*predictions{3}); title('expected stress level');
%
%
% In:
%   In-Set    : raw EEGLAB dataset that shall be streamed through the online pipeline
%
%   Model     : the predictive model (as produced by bci_train) that should be used to make predictions
%               at certain latencies in the data (specified in the Options)
%
%   Options...: name-value pairs controlling data points at which predictions are requested:
%                'markers':  predict when an event appears whose type matches any of the markers
%                'points':   predict at the given data sample latencies
%                'interval': predict every 'interval' samples
%                'sampling_rate': predict at the given sampling rate (in Hz)
%                'offset': add this offset (in samples) to the times at which the predictive model is invoked
%               misc name-value pairs:
%                'show_progress': whether to display a progress update (default: 0)
%                'feed_percent': feed only this percentage of the data (default: 100)
%                'merge_results': merge the results into an aggregated data structure (as in ml_predict) (default: 1)
%
% Out:
%   Predictions : the prediction results for every point at which the detector should be invoked
%
%   Latencies   : dataset points for the corresponding predictions (in samples)
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-03-28

opts = hlp_varargin2struct(varargin, 'markers',[], 'points',[], 'interval',[], 'sampling_rate',[], 'offset',0, ...
	'show_progress',0, 'feed_percent',100, 'merge_results',1);

% make sure that EEG is evaluated
eeg = exp_eval(eeg);

% aggregate the points at which the model should be invoked
predict_at = [];
if ~isempty(opts.points)
    predict_at = [predict_at opts.points]; end
if ~isempty(opts.interval)
    predict_at = [predict_at 1:opts.interval:size(eeg.data,2)]; end
if ~isempty(opts.sampling_rate)
    predict_at = [predict_at 1:(eeg.srate/opts.sampling_rate):size(eeg.data,2)]; end
if ~isempty(opts.markers)
    for e=1:length(eeg.event)
        if any(strcmp(eeg.event(e).type, opts.markers))
            predict_at = [predict_at eeg.event(e).latency]; end
    end
end
predict_at = unique(min(size(eeg.data,2),max(1,round(sort(predict_at)))));
predict_at = predict_at+opts.offset;
predictions = cell(1,length(predict_at));


% initialize the online stream
onl_setup(eeg.srate,{eeg.chanlocs.labels});
onl_loaddetector('onl_stream_detector', det);

cursor = 0; % data up to an including this sample latency has been streamed in
% for each latency of interest...
for i=1:length(predict_at)
    % update the online stream up to the cursor
    onl_feed(eeg.data(:,(cursor+1):predict_at(i)));
    cursor = predict_at(i);
    % run the detector
    predictions{i} = onl_predict('onl_stream_detector');
    % progress monitoring
    if opts.show_progress && mod(i,floor(length(predict_at)/1000))==0
        fprintf([num2str(ceil(1000*i/length(predict_at))/10) '%% ']); end        
    if opts.feed_percent < 100 && (100*i/length(predict_at)) > opts.feed_percent
        break; end    
end

if opts.show_progress
    fprintf('\n'); end

if opts.merge_results
    if all(cellfun(@isnumeric,predictions))
        % numeric results (e.g. regression)
        predictions = [predictions{:}];
    else
        cellinds = cellfun(@(x)~(isnumeric(x)&&isnan(x)),predictions);
        if any(cellinds)
            % cell results... try aggregating all that's not a NaN
            if any(cellfun(@(x)strcmp(x{1},'disc'),predictions(cellinds)))
                % we have discrete probability distributions, aggregate them!
                probs = cellfun(@(x)x{2},predictions(cellinds),'UniformOutput',false);
                probs = vertcat(probs{:});
                probs(cellinds,:) = probs;
                probs(~cellinds,:) = NaN;
                firstresult = predictions(find(cellinds,1));
                predictions = {'disc', probs, firstresult{1}{3}};
            end
        end        
    end
end