function out_signal_dim = bci_Preflight(in_signal_dim)
% Out-Signal-Dim = bci_Preflight(In-Signal-Dim)
% Check the parameters (put into bci_Parameters by the BCI2000 runtime) for consistency prior to running...
%
% In:
%  In-Signal-Dim  : a two-element vector of the form [#Channels, #Samples] which describes the input signal's dimensions
%
% Out:
%  Out-Signal-Dim : a two-element vector of the form [#Channels, #Samples] which describes the input signal's dimensions
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-09-08

global bci_Parameters;

% these variables are shared with Initialize and Process
global predictor meta_data outrate maxcpu outdims outform;


% --- toolbox check ---

global bcilab;
if isempty(bcilab)
    disp('Loading BCILAB execution engine...');
    wd = pwd;
    cd ../../../;
    startup;
    cd(wd);
end

% --- biopredictor checks ---

% check if the biopredictor is present
if ~exist(bci_Parameters.Biopredictor,'file')
    error('BCILAB:BCI2000_plugin:predictor_not_found',['Could not find the biopredictor file specified as ' bci_Parameters.Biopredictor '.']); end

% check if it is uniquely determined in the file's contents
predictor = bci_Parameters.Biopredictor;
variables = load(predictor);
fn = fieldnames(variables);
if length(fn) > 1
    error('BCILAB:BCI2000_plugin:predictor_not_unique',['The file specified as ' bci_Parameters.Biopredictor ' must not contain more than one variable (the biopredictor object).']); end

% check if the structure seems roughly okay...
if ~isfield(variables.(fn{1}),'bcilab')
    error('BCILAB:BCI2000_plugin:predictor_unknown',['The variable in ' bci_Parameters.Biopredictor ' is not a BCILAB structure; Please review the bci_train() help to get started.']); end


% --- stream meta-data checks ---

% try to get the sampling rate
try
    srate = str2num(bci_Parameters.SamplingRate{1});
catch
    srate = [];
end

% try to get the channel names
try
    channels = bci_Parameters.ChannelNames;
catch
    channels = [];
end

% check if the meta-data file exists
if ~exists(bci_Parameters.MetaFile,'file') && (isempty(channels) || isempty(srate))
    error('BCILAB:BCI2000_plugin:metafile_not_found',['Could not find the stream''s meta-data file specified as ' bci_Parameters.MetaFile '.']); end

% check if the meta-data file can be loaded
try
    meta_data = exp_eval(io_loadset(bci_Parameters.MetaFile));
catch
    meta_data = struct();
    if isempty(channels) || isempty(srate)
        error('BCILAB:BCI2000_plugin:metafile_unsupported',['Could not load the stream''s meta-data file specified as ' bci_Parameters.MetaFile '; unknown format; Please check the HOWTO.txt.']); end
end

% override the contents of the metafile with those of the BCI2000 parameters
if ~isempty(srate)
    meta_data.srate = srate; end

% check if the meta-data file is consistent with the channels
if ~isempty(channels)
    if isfield(meta_data.chalocs) && ~isequal(channels,{meta_data.chanlocs.labels})
        error('BCILAB:BCI2000_plugin:metafile_inconsistent',['The channel location names specified in the stream meta-data file ' bci_Parameters.MetaFile ' do not match the ChannelNames parameter in the BCI2000 setup.']); end
    if isfield(meta_data.channels) && ~isequal(channels,meta_data.channels)
        error('BCILAB:BCI2000_plugin:metafile_inconsistent',['The channel names specified in the stream meta-data file ' bci_Parameters.MetaFile ' do not match the ChannelNames parameter in the BCI2000 setup.']); end
    % and create the respective meta-data field
    if ~isfield(meta_data,{'channels','chanlocs'})
        meta_data.chanlocs = struct('labels',channels); end
end

% check that the input signal dimension matches the number of channel names
if length(meta_data.chanlocs) ~= in_signal_dim(1)
    error('BCILAB:BCI2000_plugin:metadata_inconsistent','The number of channel names (specified in the ChannelNames parameter or the MetaFile) does not match the signal''s dimension.'); end


% --- check the outputform ---
if ~any(strcmp(bci_Parameters.OutputForm,{'expectation','distribution','mode'}))
    error('The OutputForm parameter must be either ''expectation'', ''distribution'', or ''mode'''); end

% --- get remaining parameters ---
outform = bci_Parameters.OutputForm{1};
outrate = str2num(bci_Parameters.OutputRate{1});
maxcpu = str2num(bci_Parameters.MaxCPU{1});
outdims = str2num(bci_Parameters.OutputDims{1});

out_signal_dim = [outdims, size(in_signal_dim,2)];
