function run_DataRiver(varargin)
% run_DataRiver(Arguments...)
% Run BCILAB's real-time engine as a module in the DataRiver [1,2] real-time system; BCILAB reads raw data from some input stream and writes predictions to some output stream.
%
% In:
%   Arguments... : name-value pairs; some mandatory, some optional:
%
%                  mandatory arguments:
%                  * 'biopredictor' : a file or struct that contains a predictive model as previously computed by bci_train; mandatory
%
%                  * 'in_metadata' : meta-data specification for the input stream; mandatory
%                      can be in a variety of formats:
%                      - as a cell array of name-value pairs: e.g., {'name1', value1, 'name2', value2, ...}
%                      - as a struct with field names indicating names: e.g., struct('name1',{value1}, 'name2', {value2}, ...)
%                      - as a name of a .mat file that contains variables; these variables are taken as the names/values of meta-data parameters
%                      - as a name of an EEGLAB-supported dataset; meta-data will be sourced from this dataset
%                      necessary fields are 'srate' (the sampling rate), and either 'chanlocs' (EEGLAB channel locations structure) or 'channels' (cell array of channel names)
%                      optional fields are any fields that are appropriate for EEGLAB datasets (e.g., 'dipfit', 'icaweights', ...)
%
%                  optional arguments:
%                  * 'in_stream'  : name of the input stream to be used (e.g., from the acquisition system) (default: 'C:/tmp/DataRiver')
%
%                  * 'out_stream' : name of the output stream (carrying the control signal); must be a valid file name (default: 'C:/tmp/BCILAB')
%
%                  * 'out_form'   : form of the output stream, can be one of the following:
%                                   'expectation' : the expected value (= posterior mean) of the outputs; can be multi-dimensional but is usually 1d (default)
%                                       this mode is appropriate for simple applications that expect a smooth control signal, or for applications that expect a regression output
%                                   'distribution' : parameters of the output distribution; for discrete distributions, this is one probability value for each target (adding up to 1)
%                                       this mode is appropriate for more advanced applications that use the full output distribution (e.g., for decision-theoretical processing)
%                                   'mode' : the most likely output value (currently only supported for discrete distributions)
%                                       this mode is appropriate for simple applications that take a non-probabilistic classifier decision (e.g., as from a Support Vector Machine)
% 
%                  * 'out_srate'   : rate, in Hz, at which the output stream shall be sampled (default: 10)
%                                    note: if the sampling rate is higher than the rate at which BCILAB can process data, 
%                                          samples will be replicated (i.e., the same output comes multiple times)
%
%                  * 'max_cpu'     : the maximum fraction of CPU time allocated for this predictor (default: 0.6)
%                                    note: this should leave some room for other important processes (e.g., recording, data transfer, visualization, 
%                                          or other BCILAB predictors), but setting it too low (to sustain the given sampling rate) will lead to output samples being replicated
%
%                  * 'startup_args': cell array of arguments to the toolbox startup() function, if the toolbox needs to be loaded (default: {})
%
% Example:
%   run_DataRiver('biopredictor','mypredictor.mat', 'in_stream','C:/tmp/DataRiver', 'in_metadata',{'srate',512,'channels',{'C3','C4','Cz','FP1', ...}}, ...
%       'out_form','distribution', 'out_srate',25, 'max_cpu',0.25);
%
% Notes:
%   The DataRiver real-time system is in active development, and therefore, only a beta version can be obtained at this point.
%
% References:
%   [1] Vankov A., Bigdely-Shamlo N.,  Makeig S. "DataRiver ? A software platform for real time management of multiple data streams"
%       Fourth International BCI Meeting, Carmel, CA, June 2010  
%   [2] http://sccn.ucsd.edu/wiki/DataSuite
% 
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-09-08

% read options
addpath ../../code/helpers
opts = hlp_varargin2struct(varargin, {'biopredictor','predictor','detector','classifier'}, '<mandatory>', 'in_stream','C:/tmp/DataRiver', 'in_metadata','<mandatory>',  ...
    'out_stream','C:/tmp/BCILAB', 'out_form','expectation', 'out_srate',10, 'max_cpu',0.6, 'startup_args',{});

% load the toolbox, if necessary
global bcilab;
if isempty(bcilab)    
    disp('Toolbox not present; loading it now using default startup arguments...');
    wd = pwd;
    cd ../../../;    
    startup(opts.startup_args{:});
    cd(wd);
end

% load DataSuite, if necessary
global ds_lib;
if isempty(ds_lib)
    disp('Loading DataSuite...');
    startup_ds; 
end

% read in_metadata, if necessary
if ischar(opts.in_metadata)
    opts.in_metadata = exp_eval(io_loadset(opts.in_metadata)); end

% parse in_metadata (e.g. rewrite 'channels')
opts.in_metadata = hlp_varargin2struct(opts.in_metadata,'srate','<mandatory>');
if isfield(opts.in_metadata,'channels')
    if isfield(opts.in_metadata,'chanlocs')
        [opts.in_metadata.chanlocs.labels] = deal(opts.in_metadata.channels{:}); 
    else
        opts.in_metadata.chanlocs = struct('labels',opts.in_metadata.channels);
    end    
    opts.in_metadata = rmfield(opts.in_metadata,'channels');
end

% load the predictor
onl_loaddetector('BCILAB',opts.biopredictor);

% set up a BCILAB online stream
onl_setup(opts.in_metadata.srate, opts.in_metadata.chanlocs, opts.in_metadata);

% open a DataSuite input stream
in_stream = ds_OpenRead(opts.in_stream);
in_cleanup = onCleanup(@()ds_CloseRead(in_stream));

% open a DataSuite output stream
out_stream = ds_OpenWrite(opts.out_stream,opts.out_srate);
out_cleanup = onCleanup(@()ds_CloseWrite(out_stream));

% our temporary sample data structure
tmpsample = ds_array;
tmpsample.nItems = 1;
tmpsample.Data = 0;

% run
in_samples = 0;  % number of samples received so far
out_samples = 0; % number of samples produced so far
while 1
    t0 = tic;
    % while we have samples to read until the next output sample can be produced...
    while round(in_samples*(opts.out_srate/opts.in_stream.srate)) <= out_samples
        % read in at least one more sample
        received = false;
        while ~received
            % try to get a new sample
            [received,sample] = ds_Read(in_stream);
            if ~received
                pause(1/opts.in_metadata.srate); end
        end        
        % feed sample to BCI
        onl_feed(double(sample.Data(2:sample.nItems)'));
        in_samples = in_samples+1;
    end
    
    % make a prediction to produce an output sample
    t1 = tic;
    prediction = onl_predict('BCILAB');
    prediction = utl_formatprediction(prediction,opts.out_form);
    if ~(isnumeric(prediction) && isnan(prediction))
        % valid data: set up the sample
        tmpsample.nItems = length(prediction);
        tmpsample.Data = prediction;
    else
        % keep the dimensionality of the last output (or dummy output), but set data to NaN's
        tmpsample.Data = tmpsample.Data.*NaN;
    end
    % write out the data
    ds_Write(out_stream,tmpsample);
    out_samples = out_samples+1;
    
    % make sure that we are not exceeding our CPU budget
    computation_time = toc(t1);
    if opts.max_cpu < 1
        pause(computation_time * (1 - opts.max_cpu) / opts.max_cpu); end
    
    % check how much processing time is left for this output sample
    time_spent = toc(t0);
    time_left = 1/opts.out_srate - time_spent;
    if time_left < 0
        % we exceeded our time budget: replicate samples
        for k=1:ceil(-time_left*opts.out_srate)
            ds_Write(out_stream,tmpsample);
            out_samples = out_samples+1;
        end
    end
end
