function result = para_dal(data, mode, varargin)
% Result = para_dal(Input-Data, Operation-Mode, Options...)
% Paradigm for slow cortical potentials and/or oscillatory processes, using the Dual-Augmented Lagrange method for regularized (logistic) regression.
%
% The DAL paradigm implements the framework described in [1], and can be customized to give state-of-the-art results in almost any current BCI situation. 
% Note that DAL is the name of the optimization method [2], and not an accepted or recognized name for BCI paradigms using it (but is used here for the 
% lack of a better name). Since this method offers a lot of flexibility, simplified versions are also available, restricting it to either slow cortical potentials
% (para_dal_lf) or to oscillatory processes (para_dal_hf).
%
% The DAL paradigm can use optional signal (pre-)processing stages, but blurs the boundary between the subsequent feature extraction and machine learning stages.
% Features are defined using regions, in time and frequency, of the data epoch (which must be chosen depending on the task at hand), and an optimization 
% criterion which includes a 'loss' term and a 'regularization' term, each defined w.r.t. these features. The loss term can be chosen depending on whether the 
% task is to predict categories (for classifaction) or real values (for regression). The regularization term allows to impose assumptions about the structure 
% of the data (in the epoch), and, especially, in what respect(s) it is assumed to be "simple" (to control for overfitting). Overfitting control counter-balances 
% the extreme flexibility of the model, which can essentially assign a different weight for every sample and channel of the data. Regularization is very 
% time-consuming (since a parameter search is necessary), so that it may be practical to temporarily disable regularization to get quick (yet suboptimal) results.
%
% The measure of simplicity depends on the type of data being supplied. For raw sensor-space data, the assumption is that a) there are only relatively few 
% persistent informative source constellations over time, which create the patterns measured at sensor sites, and b) there are relatively few informative 
% temporal activity patterns across activated sources. This assumption is implemented in the dual-spectral, 'ds', regularizer. For independent component data, 
% the assumption is usually that only few components are informative (which is the group sparsity with row groups, 'glr', regularizer); the same regularizer can 
% also be used to select the most informative sensors. Another possible assumption is that only few channels/components, and only few time-points (or, e.g. 
% spectral components when operating on spectrally-transformed data) are relevant; this is the 'l1' regularizer. Despite these powerful constraints, the method
% can be made to overfit, especially when many different (possibly irrelevant) regions are configured. Due to its ability to select the few most relevant structures
% in the data, the method can not just be used to obtain predictive models, but also to investigate neuroscientific questions about the underlying processes - 
% for example, which areas of the brain (or collections of areas) are information w.r.t. some known latent (condition) variable.
%
% The region selection includes some advanced parameters (order and norm) which can be customized and/or parameter-searched if desired (in [1], a parameter
% search is recommended), but reasonable defaults are assigned which cover a large fraction of use cases.
%
% Example 1: Consider the well-known BCI task of classifying motor imagery; the goal is to predict whether the user is imagining a movement of his left or right hand.
% A calibration dataset would typically include events with types such as 'left-imag' and 'right-imag', to indicate the time and type of instruction stimuli
% presented to the subject (see, e.g. [3]). The subject would then imagine either one or the other movement over the course of several seconds. Several EEG features
% can be used to infer the type of movement, including the readiness potential ([4], which is a slow cortical potential) and event-related 
% synchronization/desynchronization ([5], which are oscillatory processes). Depending on the choice of frequency bands (here, separate regions are used
% for the readiness potential (0.5-10Hz) and alpha (7-15Hz) and beta (15-30Hz) rhythms), appropriate order and normalization coefficients are automatically
% chosen by para_dal, and a combined feature matrix as in [1] is constructed, from which the relevant portions are automatically selected. 
% The split in alpha and beta rhythm is here for demonstration purposes; it is not necessarily a superior choice in practice.
%
%   calib = io_loadset('datasets/john/gestures.eeg')
%   [loss,model,stats] = bci_train({'data',calib, 'paradigm',@para_dal}, 'events',{'left-imag','right-imag'}, 'learner',{'dal',search(2.^(-5:0.75:10))}), ...
%       'regions', {{'freq',[0.5 10]}, {'freq',[7 15]}, {'freq',[15 30]}})
%
%
% Example 2: Consider the working-memory load example from para_multiband_csp. Here, the assumption is that at least the theta (4-6Hz) and alpha bands (7-15Hz) 
% are relevant for predicting the number of items held in working memory by the subject. In this case, we will not pose the problem as one of classification,
% but rather as one of regression (where the target variable takes on values in {1,2,3}).
%
%   dataset = io_loadset('datasets/mary/nback.eeg')
%   [loss,model,stats] = bci_train({'data',dataset, 'paradigm',@para_dal}, 'events',{'n1','n2','n3'}, ...
%       'epoch',[-0.5 2.5], 'regions',{{'freq' [4 6]},{'freq' [7 15]}}, 'learner',{'dal',search(2.^(-5:0.75:10)), 'loss','squared'})
%
%
% Example 3: Another (hypothetical) scenario would be one in which complex temporal structure in oscillatory and slow-changing processes is expected, 
% for example during the visual processing of faces versus houses (see, e.g., [6]). We assume a dataset annotated with events of either the 'face', 
% the 'house', or the 'noise' type, indicating the presentation of a stimulus from the respective category. It has been shown that there are significant
% differences in relevant ERP components, as well as modulations in the 5-15Hz band and further effects, over the course to 400ms. We first obtain 
% a performance estimate, to find out to what degree these features can be used for category prediction.
%
%   calib = io_loadset('datasets/john/facesvshouses.eeg')
%   [loss,model,stats] = bci_train({'data',calib, 'paradigm',@para_dal}, 'events',{'face','house','noise'}, 'learner',{'dal',search(2.^(-5:0.75:10))}), ...
%       'regions', {{'freq',[0.5 10]}, {'freq',[5 15],'time',[0.05 0.15]}, {'freq',[5 15],'time',[0.15 0.25]}, {'freq',[5 15],'time',[0.25 0.4]}})
%
% The resulting model can not be visualized using the techniques detailed in [1] to find the relevant source projections and their temporal activations.
%
%
% Example 4: An interesting question from the previous example is whether and how phase resetting and amplitude spectrum in source activations are relevant 
% for the face vs. house processing. To this end, we exchange the default preprocessing of the paradigm to an independent component basis, by enabling ICA,
% and switch from the time domain into the amplitude/phase spectrum, using the fourier filter. Since the data is now in the spectral domain, the region
% selection of DAL in terms of frequencies is meaningless, so we pragmatically select 0-1000Hz. Our goal is not just to find the relative importance of 
% components and frequency bands, but also the minimal set of relevant components supporting an optimal prediction from these features, by setting the 
% 'regularizer' parameter of the DAL component to 'glr', which stands for group lasso (group sparsity) along rows. In other words, a small (sparse) set 
% of rows is selected as being the most relevant w.r.t. the condition.
% 
%   calib = io_loadset('datasets/john/facesvshouses.eeg')
%   [loss,model,stats] = bci_train({'data',calib, 'paradigm',@para_dal}, 'events',{'face','house','noise'}, 'ica','on', 'fourier','polar',
%       'learner',{'dal',search(2.^(-5:0.75:10))}, 'regularizer', 'glr'), 'regions', {{'freq',[0 1000]}})
%
%
% In:
%   Input-Set       : some data set
%
%   Operation-Mode  : 'preprocessing', to pre-process the Input-Set according to the paradigm (and parameters)
%                     'learning', to learn a predictive model from a pre-processed Input-Set (and parameters)
%                     'prediction', to make predictions given a pre-processed Input-Set and a Model
%
%   User Options... : optional parameters of the paradigm; name-value pairs 
%                     --- for preprocessing (see flt_default)
%                     * 'srate':   sampling rate, in Hz (default: 50)
%                     * 'epoch':  epoch window, in seconds relative to event (default: [-1.5 1.5])
%                     * 'events':  event types to determine epochs and the target variable (user-specified)
%                     * others, by default turned off: see flt_default
%                     --- for learning
%                     * 'regions':  cell array of time/frequency regions, each a cell array of name-value pairs with the following names:
%                                   'freq': free-form frequency selection (as in flt_select), must be specified
%                                   'time': time-window selection (as in flt_select), default: entire epoch
%                                   'order': order of the data obtained from the region, either 1 for a first-order (spatio-temporal) model
%                                            or 2 for a second-order (spatio-spectral) model; default: 1 if any freq<6Hz, otherwise 2.
%                                   'norm': two-element cell array of powers for the lhs and rhs normalization matrices that are applied to the data
%                                           default: {-0.25,-0.25} if a first-order region, {-0.5,-0.5} if a second-order region
%                                   (default: {{'freq', [7 30]}})
%                     * 'chan_prior': prior per-channel relevance, [#Chans x 1] vector (default: [] = all ones)
%                     * 'time_prior': prior per-timepoint relevance, [#Samples x 1] vector (default: [] = all ones)
%                     * 'learner': learning function to be used on the features; (default: 'dal')
%                                  note: use {'dal',search(2.^(-6:0.5:10)} here to run a parameter search
%                     --- for prediction
%                     * 'model': the detector model (as produced during the learning)
% Out:
%   Result : depends on the Operation-Mode; 
%            * for 'preprocessing', this is the preprocessed dataset
%            * for 'learning', this is the learned model
%            * for 'prediction', this is the predictions produced by the model, given the data
% 
% References:
%  [1] Ryota Tomioka and Klaus-Robert Mueller, "A regularized discriminative framework for EEG analysis with application to brain-computer interface",
%      Neuroimage, 49 (1) pp. 415-432, 2010. 
%  [2] Ryota Tomioka & Masashi Sugiyama, "Dual Augmented Lagrangian Method for Efficient Sparse Reconstruction", 
%      IEEE Signal Proccesing Letters, 16 (12) pp. 1067-1070, 2009.
%  [3] Blankertz, B., Dornhege, G., Krauledat, M., Mller, K., and Curio,G. 
%      "The non-invasive Berlin Brain-Computer interface: Fast acquisition of effective performance in untrained subjects."
%      NeuroImage 37, 2 (Aug. 2007), 539?550.
%  [4] Deecke, L.; Groezinger, B.; Kornhuber H.H. "Voluntary finger movement in man: Cerebral potentials and theory." 
%      Biol Cybern 23: 99?119, 1976
%  [5] Pfurtscheller, G., and da Silva, L. "Event-related EEG/MEG synchronizaion and desynchronization: basic principles." 
%      Clin Neurophysiol 110 (1999), 1842-1857.
%  [6] Guillaume A. Rousselet, Jesse S. Husk, Patrick J. Bennett and Allison B. Sekuler, "Single-trial EEG dynamics of object and face visual processing"
%      NeuroImage 36 (3), 843-862, 2007
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-06-25


result = para_skeleton(data, mode, hlp_varargin2struct(varargin, ...
	... % preprocessing defaults:
    'epoch',[0.5 3.5], 'srate',100, ...
    ... % feature-extraction defaults:
    'regions',{{'freq', [7 30]}}, ...
    'chan_prior',[], 'time_prior',[], ...
    ... % learning defaults:
    'learner','dal', ...
    'featureadapt',@dal_adaptfeatures,'featureextract',@dal_extractfeatures));


% adapt the feature space (e.g., normalizers)
function opts = dal_adaptfeatures(data,opts)
% find out whether we are using a dual-spectral regularizer or not
dual_spectral = ~(any(strcmp(opts.learner,'regularizer')) && ~strcmp('ds',opts.learner(find(strcmp(opts.learner,'regularizer'),1,'last')+1)));

% unify region format
if ~iscell(opts.regions{1})
    opts.regions = {opts.regions}; end

% for each region...
for r = 1:length(opts.regions)
    % convert to struct
    reg = hlp_varargin2struct(opts.regions{r},'freq','<mandatory>','time',[],'norm',[],'order',[]);
    
    % set the order, if necessary: we assume the region to be first-order if any retained frequency is below 6, otherwise second-order
    if isempty(reg.order)        
        reg.order = fastif(any(reg.freq<6),1,2); end
    
    % set the norms, in necessary (squared data needs a different power than raw data)
    if isempty(reg.norm)
        reg.norm = fastif(reg.order==1,{-0.25,-0.25},{-0.5,-0.5}); end    
    
    % compute the left- and right-multiplied preprocessing matrices for the filtered data
    flt = exp_eval(flt_select(data,reg));
    if dual_spectral
        X = num2cell(flt.data,[1 2]);
       % in this case we use matrix powers of the inverse pooled covariance matrices for scaling
        if reg.order == 1
            % first-order: spatio-temporal scaling
            reg.preproc = {cov_shrink(cat(2,X{:})')^reg.norm{1}, cov_shrink(cat(1,X{:}))^reg.norm{2}};
            % apply spatio-temporal prior
            if ~isempty(opts.chan_prior)
                reg.preproc{1} = reg.preproc{1} .* diag(opts.chan_prior); end
            if ~isempty(opts.time_prior)
                reg.preproc{2} = reg.preproc{2} .* diag(opts.time_prior); end            
        else
            % second-order: just spatial scaling
            tmp = cov_shrink(cat(2,X{:})');            
            reg.preproc = {tmp^reg.norm{1}, tmp^reg.norm{2}};
            % apply spatial prior (temporal one does not apply)
            if ~isempty(opts.chan_prior)
                reg.preproc{1} = reg.preproc{1} .* diag(opts.chan_prior.^2); 
                reg.preproc{2} = reg.preproc{2} .* diag(opts.chan_prior.^2); 
            end
            if ~isempty(opts.time_prior)
                error('temporal prior does not apply in the case of a second-order detector'); end
        end
    else
        X = flt.data;
        % in this case we use per-channel / per-timepoint scaling matrices        
        if reg.order == 1
            % first-order: spatio-temporal scaling
            reg.preproc = {diag(mean(squeeze(var(X,[],2)),2))^-0.5, diag(mean(squeeze(var(X,[],1)),2))^-0.5};
            % apply spatio-temporal prior
            if ~isempty(opts.chan_prior)
                reg.preproc{1} = reg.preproc{1} .* diag(opts.chan_prior); end
            if ~isempty(opts.time_prior)
                reg.preproc{2} = reg.preproc{2} .* diag(opts.time_prior); end
        else
            % second-order: just spatial scaling
            tmp = diag(mean(squeeze(var(X,[],2)),2))^-0.5;
            reg.preproc = {tmp, tmp};
            % apply spatial prior (temporal one does not apply)
            if ~isempty(opts.chan_prior)
                reg.preproc{1} = reg.preproc{1} .* diag(opts.chan_prior.^2); 
                reg.preproc{2} = reg.preproc{2} .* diag(opts.chan_prior.^2); 
            end
            if ~isempty(opts.time_prior)
                error('temporal prior does not apply in the case of a second-order detector'); end            
        end
    end
    
    % compute & append the block scale of the region
    X = flt.data; [lhs,rhs] = reg.preproc{:};
    if reg.order == 1
        Y = zeros(size(X));
        for t=1:size(X,3)
            Y(:,:,t) = lhs*X(:,:,t)*rhs; end
    else
        Y = zeros(size(X,1),size(X,1),size(X,3));
        for t=1:size(X,3)
            Y(:,:,t) = lhs*cov(X(:,:,t)')*rhs; end
    end
    reg.preproc = [reg.preproc 1/sqrt(sum(sum(var(Y,[],3))))];
    opts.regions{r} = reg;
end
opts.vectorize = ~any(strcmp(opts.learner,'dal'));

 
% normalize, block-diagonalize and optionally vectorize the data
function X = dal_extractfeatures(data,opts)
block = cell(length(opts.regions));
for r = 1:length(opts.regions)
    reg = opts.regions{r};
    % select a region
    flt = exp_eval(flt_select(data,reg));
    % scale the features
    X = flt.data; [lhs,rhs,bs] = reg.preproc{:};
    if reg.order == 1
        Y = zeros(size(X));
        for t=1:size(X,3)
            Y(:,:,t) = bs*(lhs*X(:,:,t)*rhs); end
    else
        Y = zeros(size(X,1),size(X,1),size(X,3));
        for t=1:size(X,3)
            Y(:,:,t) = bs*(lhs*cov(X(:,:,t)')*rhs); end
    end
    % store, for later block-diagonalization
    block{r} = Y;
end
% combine all blocks into a 3d matrix of block-diagonal matrices
X = zeros(sum(cellfun('size',block,1)),sum(cellfun('size',block,2)),size(block{1},3));
for t=1:size(X,3)
    tmp = cellfun(@(m)m(:,:,t),block,'UniformOutput',false);
    X(:,:,t) = blkdiag(tmp{:});
end
% and optionally vectorize the result
if opts.vectorize
    X = double(reshape(X,[],size(X,3))'); end
