function result = para_dal_hf(data, mode, varargin)
% Result = para_dal_hf(Input-Data, Operation-Mode, Options...)
% Paradigm for oscillatory processes using the Dual-Augmented Lagrange method for trace-norm regularized logistic regression.
%
% The DAL-HF paradigm is an easy-to-use and generic method for operating on oscillatory processes. It is a special case of the general 
% framework described in [1] (using only the "second-order" detector); the general variant, which can use multiple bands and time windows,
% as well as slow cortical potentials under a joint optimality criterion, is further explained in para_dal. DAL is the name of the 
% optimization method [2], and not an accepted or recognized name for BCI paradigms using it (but is used here for the lack of a better name).
%
% Like CSP (para_csp), DAL-HF can be used for a wide range of oscillatory processes and should theoretically out-perform it in many cases, at the
% cost of much increased computation times (due to the need for regularization). Like for CSP, the frequency band and time window on which DAL-LF shall 
% operate needs to be determined a priori. The frequency band is usually chosen to be the alpha and/or beta band for the majority of tasks, but under 
% some circumstances, it can be chosen heuristically, as in [3].  The learner parameter is either chosen to be 'dal', to run without regularization for
% quick results, or {'dal',search(2.^(-6:0.5:10)} to run with full regularization, which takes much longer but is the proper way to apply the method.
%
% Example: Consider a user of an eye tracking software to control a mouse cursor. The user wants to execute a control thought (e.g., sensorimotor) in 
% order to issue selections (i.e., a 'click'). We assume that the control signal is reflected in oscillatory processes. either in the alpha or 
% the beta band, and that it takes approx. 2 seconds to think it. For a possible calibration design, see, e.g., [4], in which the dataset contains 
% events in which the user is either in the 'search' condition or in the 'select' condition. A predictive model can then be obtained with DAL-HF as follows:
%  
%   calib = io_loadset('datasets/john/selection.eeg')
%   [loss,model,stats] = bci_train({'data',calib,'paradigm',@para_dal_hf}, 'epoch',[0 2], 'events',{'search','select'}, ...
%   'learner', {'dal',search(2.^(-5:0.75:10))});
%
%
% In:
%   Input-Set       : some data set
%
%   Operation-Mode  : 'preprocessing', to pre-process the Input-Set according to the paradigm (and parameters)
%                     'learning', to learn a predictive model from a pre-processed Input-Set (and parameters)
%                     'prediction', to make predictions given a pre-processed Input-Set and a Model
%
%   User Options... : optional parameters of the paradigm; name-value pairs 
%                     --- for preprocessing (see flt_default)
%                     * 'spectrum':  spectral filter (default: [7 30])
%                     * 'srate':   sampling rate, in Hz (default: 50)
%                     * 'epoch':  epoch window, in seconds relative to event (default: [-1.5 1.5])
%                     * 'events':  event types to determine epochs and the target variable (user-specified)
%                     * others, by default turned off: see flt_default
%                     --- for learning
%                     * 'normalizers': normalization parameters (default: [-0.5 -0.5])
%                     * 'learner': learning function to be used on the features; (default: 'dal')
%                                  note: use {'dal',search(2.^(-6:0.5:10)} here to run a parameter search
%                     --- for prediction
%                     * 'model': the detector model (as produced during the learning)
% Out:
%   Result : depends on the Operation-Mode; 
%            * for 'preprocessing', this is the preprocessed dataset
%            * for 'learning', this is the learned model
%            * for 'prediction', this is the predictions produced by the model, given the data
%
% References:
%  [1] Ryota Tomioka and Klaus-Robert Mueller, "A regularized discriminative framework for EEG analysis with application to brain-computer interface",
%      Neuroimage, 49 (1) pp. 415-432, 2010. 
%  [2] Ryota Tomioka & Masashi Sugiyama, "Dual Augmented Lagrangian Method for Efficient Sparse Reconstruction", 
%      IEEE Signal Proccesing Letters, 16 (12) pp. 1067-1070, 2009.
%  [3] Benjamin Blankertz, Ryota Tomioka, Steven Lemm, Motoaki Kawanabe, and Klaus-Robert Mueller. 
%      "Optimizing spatial filters for robust EEG single-trial analysis." 
%      IEEE Signal Process Mag, 25(1):41-56, January 2008
%  [4] Zander T.O., Gaertner M., Kothe C., Vilimek R. "Combining Eye Gaze Input with a Brain-Computer Interface for Touchless Human-Computer Interaction"
%      International Journal of Human-Computer Interaction, in press.
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-06-25

result = para_skeleton(data, mode, hlp_varargin2struct(varargin, ...
	... % preprocessing defaults:
    'spectrum',[7 30], 'epoch',[0.5 3.5], 'srate',200, ...
    ... % learning defaults:
    'learner','dal', 'normalizers',{-0.5 -0.5}, ...
    'featureadapt',@scp_getscale,'featureextract',@scp_doscale));

% compute temporal & spatial normalizers
function mdl = scp_getscale(data,opts)
X = num2cell(data.data,[1 2]);
for t=1:length(X)
    X{t} = cov(X{t}'); end
mdl.P = {cov_shrink(cat(2,X{:})')^opts.normalizers{1},cov_shrink(cat(1,X{:}))^opts.normalizers{2}};
mdl.vectorize = ~any(strcmp(opts.learner,'dal'));

% normalize and optionally vectorize the data
function X = scp_doscale(data,mdl)
X = zeros(size(data.data,1),size(data.data,1),size(data.data,3));
for t=1:size(data.data,3)
	X(:,:,t) = mdl.P{1}*cov(data.data(:,:,t)')*mdl.P{2}; end
if mdl.vectorize
    X = double(reshape(X,[],size(X,3))'); end
