function result = para_multimodel(data, mode, varargin)
% Result = para_dal(Input-Data, Operation-Mode, Options...)
% Paradigm for oscillatory processes using multi-model AMICA and derived spectral features modeled by a mixture of weighted classifiers.
% Not ready for production use yet.
%
% In:
%   Input-Set       : some data set
%
%   Operation-Mode  : 'preprocessing', to pre-process the Input-Set according to the paradigm (and parameters)
%                     'learning', to learn a predictive model from a pre-processed Input-Set (and parameters)
%                     'prediction', to make predictions given a pre-processed Input-Set and a Model
%
%   User Options... : optional parameters of the paradigm; name-value pairs 
%                     --- for preprocessing (see flt_default)
%                     * 'srate':   sampling rate, in Hz (default: 50)
%                     * 'epoch':  epoch window, in seconds relative to event (default: [-1.5 1.5])
%                     * 'events':  event types to determine epochs and the target variable (user-specified)
%                     * others, by default turned off: see flt_default
%                     --- for learning
%                     * 'models': AMICA multiple-model structure (as produced by loadmodoout), or a path to a .mat file containing such a structure, named 'm'
%                     * 'modelchans': channel subset on which the model runs (default: [])
%                     * 'spatial_prior': prior per-component relevance, cell array of one [#Chans x 1] vector, per model (default: [] = all ones)
%                     * 'spectral_prior': prior per-frequncy relevance, [#Samples x 1] vector (default: [] = all ones)
%                     * 'featuresel': feature-selection strength (l1 regularization parameter) (default: 300)
%                     * 'spec_decomp': spectral decomposition, as produced by Nima's IM code (or name of a .mat file containing such a structure, named
%                                      'icSpectralDecomposition' (default: [] = use the fourier amplitude spectrum)
%                     * 'compidx': cell array of spectral component indices to use for each model (default: [])
%                     * 'prune': prune all-zero features from the data (reducing the strain on classifiers) (default: 1)
%                     * 'model_prior': impose a prior on models when predicting (default: [])
%                     * 'unweighted': disable the log-likelihood weighting (default: 0)
%                     * 'learner': machine learning component to use (default: {'gauss','select',300, 'blend',0.25, 'quiet',1, 'pairwisedims',30, 'totaldims', 90})
%                     --- for prediction
%                     * 'model': the detector model (as produced during the learning)
% Out:
%   Result : depends on the Operation-Mode; 
%            * for 'preprocessing', this is the preprocessed dataset
%            * for 'learning', this is the learned model
%            * for 'prediction', this is the predictions produced by the model, given the data
% 
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-06-25

        
% process input arguments & set defaults
opts = hlp_varargin2struct(varargin, ...
	... % preprocessing defaults:
    'epoch',[-1.5 1.5], 'srate',180, 'fir',{[0.5 2],'hp','minimum-phase'},  ...
    ... % feature-extraction defaults:
    'modelchans',[], 'models',[], 'spatial_prior',[],'spectral_prior',[], 'spec_decomp', [],  ...
    'compidx',[], 'prune',1, 'model_prior',[], 'unweighted', 0, 'smoothspec',0, 'fouriertype','amplitude', ...
    ... % learning defaults:
    'learner', {'gauss','select',300, 'blend',0.25, 'quiet',1, 'pairwisedims',30, 'totaldims', 90} ...
    );

switch mode
    case 'preprocessing'
        % apply default signal processing chain
        result = flt_default(data, opts);
        
    case 'learning'
        result = struct();
        if ischar(opts.models)
            models = getfield(hlp_microcache('amica',@io_load,opts.models),'m');  %#ok<GFLD>
        else
            models = opts.models;
        end
        N = models.num_models;
        [C,S,T] = size(data.data); %#ok<ASGLU>
        % get per-model average probabilities
        if ~opts.unweighted
            probchans = cellfun(@(v)isequal(v,1), strfind({data.chanlocs.labels},'prob'));
            if ~any(probchans)
            	error('the dataset must contain model probabilities in channels ''prob1''..''probN'', or set ''unweighted'' to 1'); end
            probs = squeeze(nan_mean(data.data(probchans,:,:),2));
        else
            probs = ones(N,T);
        end
        % compute probability-weighted mixture of classifiers
        for m = 1:N
            fprintf('learning model %d...\n',m);
            % get normalized trial weights
            w = probs(m,:) / sum(probs(m,:));
            % get the spatio-spectral decomposition for this model
            if ~isempty(opts.compidx) && ~isempty(opts.spec_decomp)
                eeg = exp_eval(flt_specdecomp(flt_project(flt_selchans(data,opts.modelchans),models.W(:,:,m)*models.S),opts.spec_decomp,opts.compidx{m}));
            else
                eeg = exp_eval(flt_fourier(flt_project(flt_selchans(data,opts.modelchans),models.W(:,:,m)*models.S),opts.fouriertype));
            end
            disp('preprocessing');
            Xr = eeg.data; X = num2cell(eeg.data,[1 2]);
            % get weighted left/right-multiplied preprocessing/normalization matrices lhs/rhs
            lhs = sum(bsxfun(@times,w,reshape(var(Xr,[],2),size(Xr,1),[])),2).^-0.5; 
            lhs(~isfinite(lhs)) = 0; medlhs = median(lhs); lhs(lhs>3*medlhs) = medlhs; lhs = diag(lhs);
            try
                if ~isempty(opts.spatial_prior)
                    lhs = lhs .* diag(opts.spatial_prior{m}); end
            catch, end
            if opts.smoothspec
                for t=1:length(X)
                    % add some tiny amount of noise to keep the cov from blowing up 
                    X{t} = X{t} + randn(size(X{t}))*std(X{t}(:))*0.05;
                end
                % use a smoothing spectral normalizer
                rhs = cov_shrink(cat(1,X{:}))^-0.5;
            else
                % use a per-bin spectral normalizer
                rhs = sum(bsxfun(@times,w,reshape(var(Xr,[],1),[],size(Xr,3))),2).^-0.5; 
                rhs(~isfinite(rhs(:))) = 0; medrhs = median(rhs(:)); rhs(rhs(:)>3*medrhs(:)) = medrhs(:); if any(size(rhs)==1) rhs = diag(rhs); end
                try
                    if ~isempty(opts.spectral_prior)
                        rhs = rhs .* diag(opts.spectral_prior); end
                catch, end                
            end 
            disp('extracting features');
            % extract per-trial feature matrices
            Xp = zeros(size(Xr));
            for t=1:T
                Xp(:,:,t) = lhs*Xr(:,:,t)*rhs; end
            % vectorize
            X = double(reshape(Xp,[],T)');
            % prune
            if opts.prune
                disp('pruning');
                result.pruning{m} = sum(abs(X))>eps; 
                if ~any(result.pruning{m})
                    result.pruning{m} = ones(size(result.pruning{m})); end
                X = X(:,result.pruning{m});
            end
            disp('extracting targets');
            % get targets
            y = set_gettarget(data);
            % learn a weighted multinomial model
            disp('computing classifier');
            classifiers{m} = ml_train({X,y,w'},opts.learner);
            % and incorporate feature-extraction parameters
            classifiers{m}.lhs = lhs;
            classifiers{m}.rhs = rhs;
        end
        % aggregate result
        result.opts = opts;
        result.models = models;
        result.classifiers = classifiers;
        result.classes = unique(set_gettarget(data));
        result.modelprior = mean(probs,2);
        
    case 'prediction'
        models = opts.model.models;
        classifiers = opts.model.classifiers;
        trialcount = exp_eval(set_partition(data,[]));
        result = {'disc' , zeros(trialcount,length(opts.model.classes)), opts.model.classes};
        if isempty(opts.model.opts.model_prior)
            opts.model.opts.model_prior = ones(1,length(classifiers)); end
        for m=1:length(classifiers)
            % get the spatio-spectral decomposition for this model
            if ~isempty(opts.model.opts.compidx) && ~isempty(opts.model.opts.spec_decomp)
                eeg = exp_eval(flt_specdecomp(flt_project(flt_selchans(data,opts.model.opts.modelchans),models.W(:,:,m)*models.S),opts.model.opts.spec_decomp,opts.model.opts.compidx{m}));
            else
                eeg = exp_eval(flt_fourier(flt_project(flt_selchans(data,opts.model.opts.modelchans),models.W(:,:,m)*models.S),opts.fouriertype));
            end
            % apply normalization matrices
            X = zeros(size(eeg.data));
            for t=1:size(X,3)
                X(:,:,t) = classifiers{m}.lhs*eeg.data(:,:,t)*classifiers{m}.rhs; end
            % vectorize
            X = double(reshape(X,[],size(X,3))');
            % prune
            if opts.model.opts.prune
                X = X(:,opts.model.pruning{m}); end
            % predict & sum probabilities
            probs{m} = ml_predict(X, classifiers{m});
            result{2} = result{2} + probs{m}{2} * opts.model.opts.model_prior(m);
        end
        % renormalize probabilities
        result{2} = result{2} ./ repmat(sum(result{2},2),1,size(result{2},2));
end
