function ival= utl_select_timeival(erd, mrk, varargin)
% ival = utl_select_timeival(eeg, markers, <arguments>)
% Select a time interval with good descrimination
%
% In:
%   EEG : continuous eeg set, optionally resampled, laplaced and filtered
%         does also work for epoched data sets, but the window must be large enough to encompass
%         max_ival.

%SELECT_TIMEIVAL - Select a time interval with good descrimination
%
%Synopsis:
% IVAL= utl_select_timeival(CNT, MRK, <OPT>)
%
%Description:
%This function is called with band-pass filtered (!!) signals. They are
%Laplace filtered [opt.do_laplace] and then the envelope is calculated
%using the Hilbert transform.
%Then a score is calculated as signed r-square value [opt.score_proc]
%of the two classes for each channel and time point of the maximal time
%interval [opt.max_ival]. These scores are splitted into positive and
%negative part, rectified (i.e. the negative part is multiplied by -1)
%and then the maximum of positive and -negative part is calculated and
%and smoothed in time by a (centered) moving average of 250ms. The
%average of this score across selected channels is calcuated. Since an
%ERD in one part and an ERS in another part of the time interval within
%one channel is undesirable, but an ERD in one channel and an ERS in
%other channel at the same time is fine, the sign of the score is
%corrected such that it is positive in each channel at the time point
%of dest descrimination.
%Next the two time points having the two highest scores are
%selected. These define the initial interval that is iteratively
%enlarged in the subsequent steps. In each step the interval is
%enlarged either one sample to the lower or to the upper side,
%depending on whether the sum of the score below the lower limit or
%above the upper limit is larger. This process is repeated until the
%score within the selected time interval encompasses 80%
%[opt.threshold] of the total score.
%
%Arguments:
% CNT - Struct of continuous data
% MRK - Struct of markers
% OPT - Struct or property/value list of optinal parameters, see
%    description above
%
%Returns:
% IVAL - Selected time interval

% Author(s): Benjamin Blankertz
%            adapted to the PhyPA toolbox by Christian Kothe

% a set of interesting channel areas (each of them a set of channels)
motor_areas = {{'FC5' 'FC3' 'CFC5' 'CFC3' 'C5' 'C3' 'CCP5' 'CCP3' 'CP5' 'CP3'},
               {'FC1' 'FC2' 'CFC1' 'CFC2' 'C1' 'C2' 'CCP1' 'CCP2' 'CP1' 'CP2'}, 
               {'FC4' 'FC6' 'CFC4' 'CFC6' 'C4' 'C6' 'CCP4' 'CCP6' 'CP4' 'CP6'}};
          
% don't do a laplacian twice
try
    done_laplace = ~isempty(regexp(erd.bcilab.cache_term,'flt_laplace'));
catch done_laplace = 1; end
% don't resample twice
try
    if isempty(regexp(erd.bcilab.cache_term,'flt_resample')) new_srate = 200; else new_srate = []; end
catch new_srate = []; end

% accept inputs
opt = hlp_varargin2struct(varargin, ...
                  'srate', new_srate, ...
                  'filter', [], ...
                  'max_ival', [0.25 5], ...
                  'start_ival', [0.75 3.5], ...
                  'score_proc', 'r_squared_signed', ...
                  'areas', motor_areas, ...
                  'do_laplace', ~done_laplace, ...
                  'threshold', 0.80, ...
                  'expand', 1);
              
% --- preprocessing ---
     
% it is allowed to specify regular bci_preproc filter configurations
if iscell(opt.filter)
    % in case of an FFT window function (except nownd), implicitly expand the result window
    if any(strfind(opt.filter{1},'fft-')) && ~any(strfind(opt.filter{1},'-nownd'))
        opt.expand = opt.expand*2;
    end
end

% resample to something bearable, if necessary
if opt.srate
    erd = flt_resample(erd,opt.srate);
end
              
% do a laplacian transform, if necessary
if opt.do_laplace
    erd = flt_laplace(erd);
end

% apply a bandpass filter, if requested (originally IIR, as it seems (interesting!))
if ~isempty(opt.filter)
    if iscell(opt.filter)
        if ischar(opt.filter{1}) && strcmp(opt.filter{1},'fft')
            erd = flt_fast(erd, opt.filter{2}(1),opt.filter{2}(2));
        elseif ischar(opt.filter{1}) && strcmp(opt.filter{1},'iir')
            erd = flt_bandpass(erd, opt.filter{2});
        end
    else        
        erd = flt_fast(erd, opt.filter(1),opt.filter(2));        
    end
end

% calc the signal's envelope
erd = flt_envelope(erd);

% remove entire baseline? (TODO: or how can this be done trialwise here, with no markers?)
%erd = flt_rmbase(erd, []);

% generate epochs (with class-idx)
erd = prc_unpack(set_selchans(erd, cat(2,opt.areas{:}),1));
if ndims(erd.data) == 2
    erd = prc_unpack(set_makepos(erd, opt.max_ival, mrk));
end    

% calculate signed r^2 values between classes

[x,score] = utl_claco(erd,[],[],'r_squared_signed'); score.x = x';

if isempty(opt.areas)
    % empty means all channels are in one big area, if requested -> TODO: or is it one area per channel?
    opt.areas = {{erd.chanlocs.labels}};
end

% --- interval detection ---

%% for each channel calc a positive (class 1 > class 2) and a
%% negative score (class 2 > class 1)
[T,nChans] = size(score.x);
scp= max(0, score.x);
scn= max(0, -score.x);
for cc= 1:nChans,
  disc= find(scp(:,cc)<0.1*max(scp(:,cc)));
  scp(disc,cc)= 0;
  disc= find(scn(:,cc)<0.1*max(scn(:,cc)));
  scn(disc,cc)= 0;
end

%% calculate score for each channel and choose good channels
%% (one from each area)
chanscp= sqrt(sum(scp.^2, 1));
chanscn= sqrt(sum(scn.^2, 1));
chanscore= max(chanscp, chanscn);
for aa= 1:length(opt.areas),
  ci= set_chanid(score, opt.areas{aa},1); % CKO
  [mm,mi]= max(chanscore(ci));
  chansel(aa)= ci(mi);
end

%% determine favorable short time interval
smooth_msec= 250;
smooth_sa= smooth_msec/1000*erd.srate;
timescore= max(scp, scn);
timescore = utl_picktimes(timescore', [(1:length(timescore))'-smooth_sa/2 (1:length(timescore))'+smooth_sa/2])'; % CKO
timescore= mean(timescore(:,chansel),2);
tempscore= zeros(size(timescore));
idx = find((erd.times >= opt.start_ival(1)*1000) & (erd.times <= opt.start_ival(2)*1000)); % CKO
tempscore(idx)= timescore(idx);
[topscore,mm]= max(tempscore);
dt= 100/1000*erd.srate;
bidx= [max(1,mm-dt):min(mm+dt,length(timescore))];

%% determine for each channel the favorable sign of class difference
%% and adjust score accordingly
timescore= score.x(:,chansel);
for aa= 1:length(chansel),
  if mean(scp(bidx,chansel(aa))) < mean(scn(bidx,chansel(aa))),
    timescore(:,aa)= -score.x(:,chansel(aa));
  end
end
timescore= mean(timescore, 2);

%% choose initial ival as limited by the two top score time points
[topscore,mm]= max(timescore);
[ss,si]= sort(timescore);
ivalsel= sort(si(end-1:end));

%% iteratively enlarge ival
absscore= max(timescore, 0);
thresh= opt.threshold*sum(absscore);
while sum(absscore(ivalsel(1):ivalsel(2))) < thresh,
  b= [ivalsel(1)-1 ivalsel(2)+1];
  if ivalsel(1)==1,
    extend_dir= 2;
  elseif ivalsel(2)==T,
    extend_dir= 1;
  else
    d1= timescore(b);
    d2= [max(timescore(1:b(1))); max(timescore(b(2):end))];
    dirscore= d1+d2;
    [mm,extend_dir]= max(dirscore);
  end
  ivalsel(extend_dir)= b(extend_dir);
end

ival = erd.times(ivalsel)/1000;

% expand the interval, according to the expand parameter
ival = mean(ival) + (ival-mean(ival))*opt.expand;