function fig = vis_erp(sets, channels, dispmode, ptitle, varargin)
% vis_erp(Input-Sets, Channels, Display-Mode, Nocolors, additional parameters...)
% visualize the event-related potential of multiple epoched input sets
%
% In:
%       Input-Sets      :   the input eeg sets that should be compared
%                           can be a single set or a cell array of sets
%       Channels        :   the channels that should be displayed
%                           one plot per channel is generated
%       Display-Mode    :   if this string contains 
%                             - 'sqrt', the probability densities get compressed
%                             - 'mmh', a miss-minus-hit plot is generated
%                             - 'means', a means plot is generated
%                             - 'rsq', an r-square value plot is generated
%                             - 'nodist' no distribution plot is generated
%                             - 'negscale', the data is negated and scaled to mV
%      Additional Parameters: 'ylim' -> set the y data limits
%                             'xlabel' -> x axis label
%                             'ylabel' -> y axis label
%                             'xtick' -> x axis ticks...
%                             'ytick' -> y axis ticks...
%                             'title' -> plot title
%                             'screensize' -> [x y]
%                             'plotsize' -> plot size within window (one value)
%                             'papertype' -> set the paper type
%                             'papersize' -> set the paper size 
%                             'fontsize' -> font size for axis labels 
%                             'linesize' -> line size plot lines
%                             'filename' -> pdf filename where the figure should be saved
%                             'nocolors' -> 0 for colored plots, 1 for no colors, default: 0
%
% Examples:
%       vis_erp(myset, [7 12]) 
%           -- display only channels 7,12 of myset
%       vis_erp({left right}, [7 12]) 
%           -- compare sets left/right in the same diagram (per channel)
%              left will be displayed in red, right in green
%              overlapping regions are yellow (red+green additive)
%
%                               Christian Kothe, FG MMS, TU-Berlin
%                               30.03.2006

% operation happens in stages:
% - data gathering (from the inputs)
%   * various methods, e.g. cassed set, cell-array of sets
% - data preprocessing for visualization
%   * various methods, e.g. scaling, negating, miss-minus-hit
% - plot content generation
%   * various methods, e.g. mean curves, statistical envelopes, error bars, gaussian red-green plots, gaussian mixture red-green plots, smooth-kernel red-green plots
% - axis & title descriptions
%   * pretty standard, e.g. time axis, frequency axis, class<->color mapping, units, channel, plot title

if exist('dispmode') && ~isempty(strfind(lower(dispmode),'sqrt')) use_sqrt = 1; else use_sqrt = 0; end
if exist('dispmode') && ~isempty(strfind(lower(dispmode),'mmh')) miss_minus_hit = 1; else miss_minus_hit = 0; end
if exist('dispmode') && ~isempty(strfind(lower(dispmode),'means')) show_means = 1; else show_means = 0; end
if exist('dispmode') && ~isempty(strfind(lower(dispmode),'rsq')) r_square = 1; else r_square = 0; end
if exist('dispmode') && ~isempty(strfind(lower(dispmode),'nodist')) no_dist = 1; else no_dist = 0; end
if exist('dispmode') && ~isempty(strfind(lower(dispmode),'negscale')) negscale = 1; else negscale = 0; end

% optionally override the defaults
params = struct();
for i=1:2:length(varargin) params.(varargin{i}) = varargin{i+1}; end
if ~isfield(params,'fontsize') params.fontsize = 10; end
if ~isfield(params,'linesize') params.linesize = 0.5; end
if ~isfield(params,'nocolors') params.nocolors = 0; end
% we need to swizzle the dimensions of data so we can vectorize our
% expressions
if iscell(sets)
    % handle vis_erp({l r}, ...);
    tmp = {};
    for s=1:length(sets)
        if isstruct(sets{s})
            tmp{s} = permute(sets{s}.data, [3 2 1]);
        else
            tmp{s} = permute(sets{s}, [3 2 1]);
        end
    end
else        
    % handle vis_erp(set, ...);
    l = exp_eval(set_picktrials(sets,'rank',1));
    r = exp_eval(set_picktrials(sets,'rank',2));
    
    tmp{1} = permute(l.data, [3 2 1]);
    if ~isempty(r.data)
        tmp{2} = permute(r.data, [3 2 1]);
    end

    
    % transcript the channels, if necessary
    if iscell(channels)
        channels = utl_transchan(sets,channels);        
    end
    
    if ~exist('xmin') || ~exist('xmax')
        xmin = sets.xmin;
        xmax = sets.xmax;
    end
%    xmin = xmin*sets.srate-sets.xmin;
%    xmax = xmax*sets.srate-sets.xmin;    
end

if negscale
    for d=1:length(tmp)
        tmp{d} = tmp{d}*-0.1;
    end
end

if ~no_dist 
    figure;
    % for each channel (one channel per row)
    for i=1:numel(channels)
        % select the plot
        subplot(size(channels,1),size(channels,2),i);
        % --- render the image ---
        % get min/max value of all sets for this channel
        % so we can plot them into the same diagram
        vals=[];
        for s=1:length(tmp)
            if ~isreal(tmp{s})
                disp('warning: replacing complex values by their absolute value');
                tmp{s} = abs(tmp{s});
            end
            t{s} = tmp{s}(:,:,channels(i)); t{s} = t{s}(1:size(t{s},1)*size(t{s},2));
            vals = [vals t{s}];
        end
        mean_data = mean(vals);
        std_data = std(vals);
        min_data = mean_data-std_data*3;
        max_data = mean_data+std_data*3;
        % get one plane (r g b) for each set
        img = {};
        for s=1:length(tmp)
            [img{s} mu{s}] = plane(tmp{s}(:,:,channels(i)), min_data, max_data);
            if use_sqrt
                img{s} = sqrt(img{s});
            end
        end
        % we might need to fill up the 2nd or 3rd plane with zeros
        if length(tmp) < 2
            img{2} = zeros(size(img{1},1), size(img{1},2));
        end
        if length(tmp) < 3
            img{3} = zeros(size(img{1},1), size(img{1},2));
        end
        % we need to rescale the colors manually
        max_img = max([max(max(img{1})) max(max(img{2})) max(max(img{3}))]);
        img = cat(3, scale(img{1},0,max_img), scale(img{2},0,max_img), scale(img{3},0,max_img));
        if exist('xmin') && exist('xmax')
            image(img, 'XData',[xmin xmax],'YData',[min_data max_data]);
        else
            image(img);
        end
        try
            xlabel(['time (' sets.chanlocs(channels(i)).labels ')']); ylabel('voltage');
            if exist('ptitle') title(ptitle); end
        catch,end
    end
end
    
% set plot size
try fig = figure('Position',[0 0 params.screensize(1), params.screensize(2)]); catch fig = figure; end
set(gcf,'PaperUnits','centimeters');
% set paper type
try set(gcf,'PaperType',params.papertype); catch,end
% set paper size
try set(gcf,'PaperSize',params.papersize); catch,end

% for each channel (one channel per row)
for i=1:numel(channels)
    if channels(i) == 0 continue;end
    % select the plot
    subplot(size(channels,1),size(channels,2),i);
    % --- render the image ---
    % get min/max value of all sets for this channel
    % so we can plot them into the same diagram
    vals=[];
    for s=1:length(tmp)
        if ~isreal(tmp{s}) 
            disp('warning: replacing complex values by their absolute value');
            tmp{s} = abs(tmp{s});
        end
        t{s} = tmp{s}(:,:,channels(i)); t{s} = t{s}(1:size(t{s},1)*size(t{s},2));
        vals = [vals t{s}];
    end
    mean_data = mean(vals);
    std_data = std(vals);
    min_data = mean_data-std_data*3;
    max_data = mean_data+std_data*3;    
    % get one plane (r g b) for each set
    img = {};
    for s=1:length(tmp)
        [img{s} mu{s}] = plane(tmp{s}(:,:,channels(i)), min_data, max_data);
        if use_sqrt
            img{s} = sqrt(img{s});
        end
    end

    hold on;
    if params.nocolors
        colors = {'-','--','*'};
    else
        colors = {'r','g','b'};
    end
    
    if miss_minus_hit
        plot(xmin:((xmax-xmin)/(length(mu{2})-1)):xmax, mu{1}-mu{2},'black','LineWidth',params.linesize);
        xlabel(['time (' sets.chanlocs(channels(i)).labels ')'],'FontSize',params.fontsize); ylabel('voltage','FontSize',params.fontsize);            
    elseif r_square
        tr = [tmp{1}(:,:,channels(i)); tmp{2}(:,:,channels(i))];
        cls = [zeros(size(tmp{1},1),1); ones(size(tmp{2},1),1)];        
        plot(xmin:((xmax-xmin)/(length(mu{s})-1)):xmax, corr(tr,cls).^2,'black','LineWidth',params.linesize);
        xlabel(['time (' sets.chanlocs(channels(i)).labels ')'],'FontSize',params.fontsize); ylabel('r² value','FontSize',params.fontsize);
    else        
        for s=1:length(mu)
            h= plot(xmin:((xmax-xmin)/(length(mu{s})-1)):xmax, mu{s},colors{s},'LineWidth',params.linesize);
            set(h,'LineWidth',2);
        end
        %xlabel(['time (' sets.chanlocs(channels(i)).labels ')'],'FontSize',params.fontsize); ylabel('voltage','FontSize',params.fontsize);            
    end
    title(sets.chanlocs(channels(i)).labels,'FontSize',params.fontsize);
    try ylim(params.ylimit); catch,end;
    try ylabel(params.ylabel,'FontSize',params.fontsize); catch,end;
    try xlabel(params.xlabel,'FontSize',params.fontsize); catch,end;
    try title(params.title,'FontSize',params.fontsize); catch,end;
    try set(gca,'ytick',params.ytick); catch,end;
    try set(gca,'xtick',params.xtick); catch,end;
    try set(gca,'CameraViewAngle',params.plotsize); catch,end; 
    grid on
    line([sets.xmin sets.xmax],[0 0],'color','black');
    line([0 0],[-10 10],'color','black');    
%    ylim([-1.1 1.1])
    set(gca,'FontSize',params.fontsize);
    hold off;
    try
       if exist('ptitle') && ~isempty(ptitle) title(ptitle,'FontSize',params.fontsize); end
    catch,end        
end

 
try
   % save as pdf/png etc...
   save2pdf(params.filename,fig,600);
   %... and save as figure
   params.filename(end-3:end) = '.fig';
   saveas(fig,params.filename);
catch,end

% calculate an image plane from a (trials x samples)-matrix
% display only the values between min_data and max_data
function [img mu sigma] = plane(mat, min_data, max_data)
    % get mu/sigma per time sample
    mu = mean(mat(:,:));
    sigma = std(mat(:,:));
    % generate the image
    img = zeros(round(max_data-min_data), size(mat,2));
    for x=1:size(img,1)
        img(size(img,1)-x+1,:) = normpdf(((x/size(img,1))*(max_data-min_data)+min_data), mu, sigma);
    end

% normalize an array to 0..1 (min_mat and max_mat are optional)
function res = scale(mat, min_mat, max_mat)
    if (~exist('min_mat')) min_mat = min(min(mat)); end
    if (~exist('max_mat')) max_mat = max(max(mat)); end
    res = (mat - min_mat) * 1/(max_mat - min_mat);

% sample a normal probability distribution specified by mu/sigma at the time points given by x
function y = normpdf(x,mu,sigma)
    y = exp(-0.5 * ((x - mu)./sigma).^2) ./ (sqrt(2*pi) .* sigma);

    
    