function res = utl_block(assignments, expr, context)
% Result = utl_block(Assignments, Expression)
% define a block with dynamically scoped values for symbols used in the contained expression.
%
% In:
%   Assignments : cell array of exp_set/exp_setdelayed expressions; the left-hand side (first argument) of the assignment expressions is 
%                 replaced by the respective right-hand side (second argument) in the scope of the block expression; this implements dynamically scoped variables.
%
%   Expression  : an expression to be evaluated with assignments as listed in the block
%
% Out:
%   Result : evaluation of the expression, given the block definitions
%
%                                        Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                                        2010-05-03

global bcilab;
persistent frametable; % a table of recently created stackframe functions (these are expensive to compute!)

if ~iscell(assignments)
    error('assignments should be given as a cell array of the form {Set(...),Set(...),SetDelayed(...),Set(...), ...}'); end

% evaluate the assignments and put them into a new stack frame
new_frame = struct();
for assignment=assignments
    a = assignment{1};
    if any(strcmp(char(a.head),{'Set','exp_set'}))
        rhs = exp_eval(a.parts{2});
    else
        rhs = a.parts{2};
    end    
    new_frame.(char(get_function_symbol(a.parts{1}))) = rhs;    
end

% store the stack frame, tagged with a fresh frameid
new_frameid = sprintf('f%d',max([0 cellfun(@(id)sscanf(id(2:end),'%d',1),fieldnames(bcilab.stackframes.ov)')])+1);
bcilab.stackframes.ov.(new_frameid) = new_frame;

% evaluate the expression (with the frameid introduced into MATLAB's stack)
if isfield(frametable,new_frameid)
    res = frametable.(new_frameid)(expr);
else
    f = eval(['@(x,active_stackframe__' new_frameid ')exp_eval(x)']);
    frametable.(new_frameid) = f;
    res = f(expr);
end

% garbage-collect the stack frame(s)...
if ~exist('context','var')
    try
        throw; %#ok<LTARG>
    catch
        context = lasterror;  %#ok<LERR>
    end
end
% get all stackframe entries in the context
entries = {context.stack.name};
framepos = strfind(entries,'active_stackframe__');
% and collect all active frameids
active_ids = {};
for i=find(~cellfun('isempty',framepos))
    active_ids{end+1} = entries{i}((framepos{i}+19):end-12); end
% remove all non-active stack frames
allframes = fieldnames(bcilab.stackframes.ov);
remove = true(length(allframes),1);
for i=1:length(active_ids)
    remove = remove & ~strcmp(allframes,active_ids{i}); end
bcilab.stackframes.ov = rmfield(bcilab.stackframes.ov,allframes(remove));