function [measure,stats] = utl_crossval(data, varargin)
% [Measure, Stats] = utl_crossval(Data, Arguments...)
% A generic cross-validation function over indexable data.
%
% In:
%   Data        :   some data that can be partitioned using index sets, such as, for example,
%                   {X,y} with X being the [NxF] training data and y being the [Nx1] labels (N=#trials, F=#features)
%
%   Arguments   :   optional name-value pairs specifying the arguments:
%                   'trainer': training function; receives a partition of the data (as produced by the specified partitioner),
%                              possibly some further arguments as specified in args, and returns a model (of any kind)
%                              (default: @ml_train; natively supports the {X,y} data format)
%
%                   'tester' : testing function; receives a partition of the data (as produced by the partitioner) and a model 
%                              (as produced by the trainer), and returns a prediction for every index of the input data,
%                              in one of the formats that can be produced by ml_predict (default: @ml_predict)
%
%                   'scheme': cross-validation scheme, can be one of the following formats (default: 10)
%                     * 0: skip CV, return NaN and empty statistics
%                     * k: k-fold randomized CV (or, if 0<k<1, k-holdhout CV)
%                     * [r k]: r times repartitioned k-fold randomized CV (or, if 0<k<1, k-holdout CV (with k a fraction))
%                     * [r k m]: r times repartitioned k-fold randomized CV (or, if 0<k<1, k-holdout CV (with k a fraction)) with m indices margin width
%                                (between training and test set)
%                     * 'loo': leave-one-out CV
%                     * {'chron', k} or {'block', k}: k-fold chronological/blockwise CV
%                     * {'chron', k, m} or {'block', k, m}: k-fold chronological/blockwise CV with m indices margin width (between training and test set)
%
%                   'partitioner': partitioning function for the data, receives three parameters: (data, index vector, packed trainer args OR model)
%                                   * if the index vector is empty, should return the highest index in the data
%                                   * otherwise, it should return data subindexed by the index vector
%                                  default: provides support for cell arrays, numeric arrays, struct arrays and {Data,Target} cell arrays
%                                  note: the third parameter is for convenience and may optionally be taken into account; trainer args are passed packed into 
%                                  a cell array for both index set generation and computation of the training partition(s), and the model is passed for the
%                                  computation of testing partitions
%
%                   'target': a function to derive the target variable from a partition of the data (as produced by the partitioner),
%                             for evaluation; the allowed format is anything that may be output by ml_predict
%                             default: provides support for {Data,Target} cell arrays
%
%                   'metric': metric to be employed, applied both to results of each fold and results aggregated over all folds
%                              * function handle: a custom, user-supplied loss function; receives target data in the first 
%                                argument and prediction data in the second argument; each can be in any format that can be 
%                                produced by ml_predict (but can be expected to be mutually consistent).
%                                shall return a real number indicating the summary metric over all data, and optionally 
%                                additional statistics in a struct
%                              * string: use ml_calcloss, with 'metric' determining the loss type
%                              * default/empty: use 'mcr','mse','nll','kld' depending on supplied target & prediction data formats, via ml_calcloss
%
%                   'args': optional arguments to the training function, packed in a cell array (default: empty)
%                           note: if using the default trainer/tester combination, args must at least specify the learning function to be used,
%                           e.g. {'lda'} for linear discriminant analysis (see ml_train* functions for options)
%
%                   'repeatable': whether the randomization procedure shall give repeatable results (default: 1); different numbers (aside from 0)
%                                 give different repeatable runs, i.e. the value determines the randseed
%
%                   'engine': parallelization engine to use (default: 'global'); see hlp_beginsschedule
%                   
%                   'pool'  : worker pool to use (default: 'global'); see hlp_beginsschedule
%
%                   'policy': scheduling policy to use (default: 'global'); see hlp_beginschedule
%
% Out:
%   Measure : a measure of the overall performance of the trainer/tester combination, w.r.t. to the target variable returned by the target function
%             computed by the metric
%
%   Stats   : additional statistics, as produced by the metric
%
% Example:
%   trials: [NxF] array of training instances
%   targets: [Nx1] array of labels
%   [loss,stats] = utl_nested_crossval({trials,targets}, 'args',{'lda'})  % cross-validate using (shrinkage) linear discriminant analysis
%   [loss,stats] = utl_nested_crossval({trials,targets}, 'args',{'hkl' 'kernel' 'hermite'})  % cross-validate using hierarchical kernel learning with a specific kernel
%
% Configuration Examples:
%   A simple training function would be:
%     @ml_train, with args being {'lda'} -- for this case, the data X (size NxF) and labels y (size Nx1) should be supplied as {X,y} to utl_crossval
%
%   A simple prediction function would be:
%     @ml_predict
%
%   A simple partitioner for epoched EEG datasets would be:
%     function result = my_partitioner(data,indices,misc)
%     if isempty(indices)
%         result = data.trials;
%     else
%         result = pop_select(data,'trial',indices);
%     end
%
%   A simple mean-square error loss metric would be:
%     my_metric = @(target,predicted) mean((target-predicted).^2);
%
%   A simple target extraction function for epoched EEG datasets would be (assuming that there is an epoch-associated target value):
%     my_target = @(data) [data.epoch.target];
%
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-04-07

% read arguments
opts = hlp_varargin2struct(varargin, 'scheme',10, 'partitioner',@utl_default_partitioner, 'target',@utl_default_target, ...
    'metric',@(T,P) ml_calcloss([],T,P), 'trainer',@ml_train, 'tester',@ml_predict, 'args',{}, 'repeatable',1, 'engine','global', 'pool','global', 'policy','global');

% string arguments are considered to be variants of the default metric
if isempty(opts.metric) || ischar(opts.metric) || (iscell(opts.metric) && all(cellfun(@ischar,opts.metric)))
    opts.metric = @(T,P)ml_calcloss(opts.metric,T,P); end
if ~has_stats(opts.metric)
    opts.metric = @(T,P)add_stats(opts.metric(T,P)); end

% derive indices from CV scheme & N
inds = make_indices(opts.scheme,opts.partitioner(data,[],opts.args),opts.repeatable);

if isempty(inds)
    measure = NaN;
    stats = struct();
else
    
    time0 = tic;
    
    
%     % evaluate each fold
%     for p = 1:length(inds)
%         % learn a model
%         model = opts.trainer(opts.partitioner(data,inds{p}{1},opts.args),opts.args{:});
%         % and record both the real test targets and the prediction on the test data
%         testset = opts.partitioner(data,inds{p}{2},model);
%         targets{p} = opts.target(testset);
%         predictions{p} = opts.tester(testset,model);
%     end
    
    % generate tasks for each fold
    for p = 1:length(inds)
        tasks{p} = {@utl_evaluate_fold,opts,data,inds{p}}; end
    
    % schedule the tasks
    results = hlp_schedule(tasks, 'engine',opts.engine, 'pool',opts.pool, 'policy',opts.policy);
    
    % collect results
    for p=1:length(inds)
        % collect results
        targets{p} = results{p}{1};
        predictions{p} = results{p}{2};
    end
        
    % compute aggregate metric
    [measure,stats] = opts.metric(aggregate_results(targets{:}),aggregate_results(predictions{:}));

    % add per-fold metric    
    for p=1:length(targets)
        stats.per_fold(p) = hlp_getresult(2,opts.metric,targets{p},predictions{p}); end
    
    % also add the original targets & predictions
    for p=1:length(targets)
        stats.per_fold(p).targ = targets{p};
        stats.per_fold(p).pred = predictions{p};
    end

    % add original index sets
    if all(cellfun(@(inds) length(inds{1}) < 10000 && length(inds{2}) < 10000,inds))
        for p=1:length(targets)
            stats.per_fold(p).indices = inds{p}; end
    end                
    
    % add additional stats
    stats.expression = struct('head',{@utl_crossval}, 'parts',{varargin});
    stats.time = toc(time0);
    
end
end


function inds = make_indices(S,N,repeatable)
% Inds = make_indices(Scheme,Index-Cardinality)
% make cross-validation indices for each fold, from the scheme and the index set cardinality

% set parameter defaults
k = 10;                     % foldness or fraction of holdout data
repeats = 1;                % # of monte carlo repartitions
randomized = 1;             % randomized indices or not
margin = 0;                 % width of the index margin between training and evaluation sets

% parse scheme grammar
if isnumeric(S) && length(S) == 3
    % "[repeats, folds, margin]" format
    repeats = S(1);
    k = S(2);
    margin = S(3);
elseif isnumeric(S) && length(S) == 2
    % "[repeats, folds]" format
    repeats = S(1);
    k = S(2);
elseif iscell(S) && ~isempty(S) && ischar(S{1}) && ...
        (~isempty(strmatch('chron',S{1})) || ~isempty(strmatch('block',S{1})))
    % "{'chron', k, m}" format
    randomized = 0;
    if length(S) > 1 && isscalar(S{2})
        k = S{2}; end
    if length(S) > 2 && isscalar(S{3})
        margin = S{3}; end
elseif isscalar(S)
    % "k" format
    k = S;
elseif ischar(S) && strcmp(S,'loo')
    % "'loo'" format
    k = N;
else
    error('unknown cross-validation scheme format');
end

% check for skipped CV
inds = {};
if k <= 0
    return; end

if randomized && repeatable     
    if hlp_matlab_version < 707
        % save & override RNG state
        randstate = rand('state'); %#ok<*RAND>
        rand('state',5182+repeatable); %#ok<RAND>
    else
        % create a legacy-compatible RandStream
        randstream = RandStream('swb2712','Seed',5182+repeatable);
    end
end

% generate evaluation index sets from the parameters
try
    perm = 1:N;
    for r=1:repeats
        if randomized
            if hlp_matlab_version < 707
                perm = randperm(N);
            else
                perm = randstream.randperm(N);
            end
        end
        if k < 1
            % p-holdout
            inds{end+1} = {sort(perm(1+(0:(round(N*k)-1))))};
        else
            % k-fold
            for i=0:k-1
                inds{end+1} = sort(perm(1+floor(i*N/k) : min(N,floor((i+1)*N/k)))); end
        end
    end
catch    
    indexgen_error = lasterror; %#ok<LERR> % this error is carried over across the subsequent delicate RNG state restoration
end

if randomized && repeatable && hlp_matlab_version < 707
    % restore saved RNG state
    rand('state',randstate); %#ok<RAND>
end

if exist('indexgen_error','var')
    rethrow(indexgen_error); end % throw the error that happened during previous index set creation 

% add complementary training index sets
for i=1:length(inds)
    tmpinds = true(1,N);
    tmpinds(inds{i}) = 0;
    for j=1:margin
        tmpinds(max(1,inds{i}-j)) = 0;
        tmpinds(min(N,inds{i}+j)) = 0;
    end
    inds{i} = {find(tmpinds),inds{i}};
end
end



% aggregate the given results (in any format allowed for ml_predict) into a single array
function res = aggregate_results(varargin)
    if isnumeric(varargin{1})
        res = vertcat(varargin{:});
    elseif iscell(varargin{1}) && length(varargin{1}) > 1
        % make sure that the result formats are all equal
        v1 = varargin{1}([1 3:end]);
        for i=2:length(varargin)
            if ~isequal(v1,varargin{i}([1 3:end]))
                error('result formats are not identical.'); end
        end
        % grab the actual parameters
        params = cellfun(@(x)x{2},varargin,'UniformOutput',false);
        if isnumeric(varargin{1}{2})
            % numeric parameters are vertically concatenated
            res = {varargin{1}{1} vertcat(params{:}) varargin{1}{3:end}};
        elseif iscell(varargin{1}{2})
            % cell parameters are horizontally concatenated
            res = {varargin{1}{1} horzcat(params{:}) varargin{1}{3:end}};
        else
            error('unsupported result format.');
        end
    else
        error('unrecognized result format.');
    end
end



% test whether the given metric supplies stats or not
function result = has_stats(metric)
try 
    [x,y] = metric([],[]);  %#ok<NASGU,ASGLU>
    result = true;
catch
    e = lasterror; %#ok<LERR>
    result = isempty(strmatch(e.identifier,{'MATLAB:TooManyOutputs','MATLAB:maxlhs','MATLAB:unassignedOutputs'},'exact'));
end
end

% add stats to the result of some metric
function [result,stats] = add_stats(result)
stats.value = result;
end