function res = utl_dataset_partitioner(paradigm,data,inds,args)
% Result = dataset_partitioner(Trainer, Raw-Data, Index-Set, Arguments)
% Internal to bci_train.
% compute a partitioned version of the specified raw dataset, given the trainer (which knows where in the preprocessing chain to partition, and what the other preprocessing shall be) 
% and its arguments; the hard part is to handle nested partitioning, caching of intermediate results and processing stages before/after the partitioning correctly

exp_declare_symbols partitioner prepartitioner
exp_declare_patterns x y

% Step 0: extract prepartitioner() annotations from the data expression:
% retain them in a variable - and their index set parameter in particular - for subsequent insertion of a set_partition(x,indexset) processing step
% at an appropriate place of the data processing expression
prepartitioners = utl_cases(data,prepartitioner(exp_blank,exp_blank),Inf);
if ~isempty(prepartitioners)
    % ... and remove the prepartitioner()'s from the data expression
    data = utl_replacerepeated(data, {exp_rule(prepartitioner(x,exp_blank),x)}); end


% Step 1: obtain symbolically processed data according to the paradigm and its arguments (data & args)
% this creates an expression which may contain partitioner() hints interpersed with actual preprocessing steps
if isfield(args,'bcilab') && isfield(args.bcilab,'online_expression')
    % Arguments is a model: it contains the definitive online expression for processing the data online (thus we use it to derive the test set)
    % for this, we substitute any set_partition(x,'all') in it by a partitioner(x) node, and its online @rawdata parameter by the raw dataset
    processed = {{ utl_replacerepeated(args.bcilab.online_expression, {exp_rule(set_partition(x,'all'),partitioner(x)),exp_rule(@rawdata,data)}) }};
else
    % Arguments are paradigm arguments: these determine the preprocessing steps to be applied, but there may be multiple argument combos specified by the user, 
    % each of which can determine a different preprocessing, enumerated by search() clauses. Thus, we obtain all expressions by grid-searching through them 
    % (invoking the paradigm in each instance).
    processed = hlp_getresult(3,@utl_gridsearch,'clauses',@(data,varargin) paradigm(data,'preprocessing',varargin{:}), data, args{:});
end

% Step 2: Complete / refine partitioner() hint placement
for i=1:length(processed)
    % if the data is not yet outfitted with a partitioner() node, add one as the last step
    % if it isn't, it means that the description of the dataset does not allot a special place where to do the partitioning (e.g. before an ICA)
    if ~utl_count(processed{i}{1},exp_blank(partitioner),Inf)
        processed{i}{1} = partitioner(processed{i}{1}); end
    % remove all except for the innermost partitioners (the point is to partition the dataset once, and do it as early in the processing chain as necessary)
    processed{i}{1} = utl_replacerepeated(processed{i}{1},{exp_rule(exp_condition(partitioner(x),exp_count(x,partitioner(exp_blank),Inf)),x)});
    % if applicable, insert the prepartitioning directly before the partitioner(s) that are now found in the data expression, using set_partition()
    % the prepartitioning is essentially an annotation which says that the data "should actually have been partitioned into XYZ, it was just not clear at what stage to do it"
    if ~isempty(prepartitioners)
        processed{i}{1} = utl_replaceall(processed{i}{1}, {exp_rule(partitioner(x),partitioner(set_partition(x,prepartitioners{1}.parts{2})))}); end
end

% Step 3: transform into a concrete evaluatable expression (partitioner and prepartitioner, etc., are only hints / symbols)
% and optimize disk and memory caching behavior (certain stages shall be disk-cached, others memory-cached)
if isempty(inds) || length(processed) == 1    
    if ~isempty(inds)
        % a concrete processed data expression is to be partitioned; contains exactly one partitioner() node -- rewrite that into a set_partition() node
        processed = utl_replacerepeated(processed{1}{1},{exp_rule(partitioner(x),set_partition(x,inds))});
        
        % find the expression right below the lowest set_partition()
        lowest_set_partition = utl_cases(processed, ...                     % All cases of...
            	exp_condition( ....                                         % patterns matching ...
                    set_partition(y,exp_blank), ...                         % set_partition(y,_) ...
                    exp_not(exp_count(y,exp_blank(@set_partition),Inf)) ... % where y contains no set_partition() node.
                    ), ...
                Inf);
                
        % set caching/checks in the evaluation options
        eval_options = { ...
            ... % turn off safety guards
            exp_rule(@fingerprint_check,0), exp_rule(@fingerprint_create,0), exp_rule(@memoize, ...
            ... % mem-cache the expression that sits directly below the first set_partition() node
            {'memory', exp_match(@expression,lowest_set_partition{1}.parts{1}),...
            ... % disk-cache everything that does not contain a set_partition() node
            'disk',exp_not(exp_count(@expression,set_partition(exp_blank,exp_blank),Inf))})};
        
        % evaluate it
        res = utl_block(eval_options, processed);
    else
        % the index set cardinality is requested for a given dataset / collection of datasets
        % find all partitioner(...) expressions to evaluate the cardinality of their subexpressions' index sets
        partitioner_subexpressions = cellfun(@(x) utl_cases(x{1},exp_blank(partitioner),Inf),processed,'UniformOutput',false);

        % set caching/checks in the evaluation options
        eval_options = {...
            ... % turn off safety guards
            exp_rule(@fingerprint_check,0), exp_rule(@fingerprint_create,0), exp_rule(@memoize, ...
            ... % mem-cache the expression that sits directly below the first partitioner() node 
            ... % (todo: ... that sit below any of the partitioner nodes!)
            {'memory', exp_match(@expression,partitioner_subexpressions{1}{1}.parts{1}),... 
            ... % disk-cache anything that does not contain a set_partition() expression
             'disk', exp_not(exp_count(@expression,set_partition(exp_blank,exp_blank),Inf))})};
         
        % evaluate the cardinality, using set_partition(x,[]), for all partitioner subexpressions
        % and by directly rewriting partitioner(set_partition(...,indices)) into length(indices) as an optimization
        cardinality = cellfun(@(d) utl_block(eval_options, ...
            utl_replacerepeated(d{1},...                
                {exp_rule(partitioner(set_partition(exp_blank,y)), exp_lambda(@length,y)), ...
                 exp_rule(partitioner(x),set_partition(x,[]))} )), ...
            partitioner_subexpressions);
        
        % ... and make sure that all cardinalities are equal
        res = unique(cardinality);
        if length(res) > 1
            error('the amount of samples/trials for crossvalidation is not consistent across search parameters; partitioning is not possible.'); end
        return;
    end
else
    % a partitioning for multiple argument combos is requested, thus we are in the outer partitioning of a nested cross-validation: 
    % remember to partition the data when we later step into that nested partition
    res = prepartitioner(data,inds);
end