function [hash,inargs,outargs] = utl_fileinfo(filename,identifier)
% [Hash,In-Args,Out-Args] = utl_fileinfo(Filename,Identifier)
% returns an md5 hash for a given m-file and also the list of input/output argument names, efficiently
%
% In:
%   Filename   : full pathname of an m-file
%   Identifier : MATLAB identifier of the file
%
% Out:
%   Hash     : md5 hash of the m-file
%   In-Args  : names of input arguments
%   Out-Args : names of output arguments
%
% Notes:
%   Relies on file time stamps, can be slow if the file system does not cache these.
%
%						Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                       2010-03-28

persistent filetable;
refresh_period = 1;  % refreshing our view of the file system at most every refresh_period seconds
renew = false; % indicates whether the database entry for this file is valid or has to be (re)created

% look up the identifier from the DB
try
    namerecord = filetable.(identifier);
catch
    % or create a blank identifier entry
    namerecord = struct('filenames',{{}},'filerecords',{{}});
    filetable.(identifier) = namerecord;
    renew = true;
end

% look for matching file record for this identifier...
idx = find(strcmp(namerecord.filenames,filename),1);
if isempty(idx)
    % or create a blank file entry
    idx = length(namerecord.filenames)+1;
    namerecord.filenames{idx} = filename;
    namerecord.filerecords{idx} = [];
    filerecord = [];
    renew = true;
else
    filerecord = namerecord.filerecords{idx};
end

% check if we need to recreate or refresh with info from the OS
if renew || (toc(filerecord.info.last_check) > refresh_period)
    fileinfo = dir(filename);
    % check if the file was actually changed
    if renew || ~strcmp(filerecord.info.date,fileinfo.date)
        % the record has to be renewed
        fileinfo.md5 = md5(filename);
        [fileinfo.inargs,fileinfo.outargs] = hlp_inoutargs(filename);
        fileinfo.last_check = tic;
        filerecord.info = fileinfo;
        namerecord.filerecords{idx} = filerecord;
        filetable.(identifier) = namerecord;
    else
        % remember when we last checked the OS record
        filetable.(identifier).filerecords{idx}.info.last_check = tic;
    end
end

inargs = filerecord.info.inargs;
outargs = filerecord.info.outargs;
hash = filerecord.info.md5;
