function prediction = utl_formatprediction(prediction,out_format)
% Formatted-Prediction = utl_formatprediction(Raw-Prediction,Format)
% Bring a prediction into a certain form; utility for post-processing the outputs of onl_predict.
%
% In:
%   Prediction : a prediction as made by ml_predict / onl_predict / bci_predict (for one or more target values)
%
%   Format     : format of the output prediction (in the descriptions, N is the number of predictions); can be one of:
%                 * 'expectation': the output is the expected value (i.e., posterior mean) of the quantity to be predicted; can be multi-dimensional [NxD]
%                 * 'distribution': the output is the probability distribution (discrete or continuous) of the quantity to be predicted 
%                                   usually, this is a discrete distribution - one probability value for every possible target outcome [NxV]
%                                   it can also be the parameters of a parametric distribution (e.g., mean, variance) - yielding one value for each parameter [NxP]
%                 * 'mode': the mode [Nx1], or most likely output value (only supported for discrete probability distributions)
% 
% Out:
%   Formatted-Prediction : The formatted output, as described in Format
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-09-08


% convert into the desired output format
if ~(isnumeric(prediction) && isnan(prediction))
    % we have valid outputs...
    if iscell(prediction)
        % we have a formatted distribution
        if strcmp(out_format,'expectation')
            % expected value is requested
            if strcmp(prediction{1},'disc')
                % discrete distribution
                prediction = prediction{2}*prediction{3};
            elseif ischar(prediction{1}) && ~strcmp(prediction{1},'struct')
                % standard distribution (e.g. Normal, Binomial, ...)
                args = cell(size(prediction{2},2);
                for k = 1:size(prediction{2},2)
                    args{k} = prediction{2}(:,k); end
                prediction = feval([prediction{1} 'stat'],args{:});
            else
                % either a user-specified distribution or a structured prediction
                error('Unsupported predicted distribution');
            end
        elseif strcmp(out_format,'mode')
            % the mode of the distribution is requested
            if strcmp(prediction{1},'disc')
                % discrete distribution: most likely class
                prediction = prediction{3}(argmax(prediction{2}'));
            else
                % anything else: we have no way to compute this
                error('Unsupported predicted distribution');
            end
        elseif strcmp(out_format,'distribution')
            % we just return the distribution's parameters; the type of the distribution is meta-data
            prediction = [prediction{2:end}];
        else
            error('Unknown output format');
        end
    elseif isnumeric(prediction)
        % we have a point estimate; there is nothing to be transformed anyway...
        if strcmp(out_format,'distribution')
            error('A distribution is not available from this predictor'); end
    else
        error('Unknown prediction output');
    end
else
    % we have NaN outputs (probably due to an error); leave them unmodified
end
