function [x,changed,rules] = utl_replaceall(x,rules)
% [Out-Expression,Changed,Remaining-Rules] = utl_replaceall(In-Expression, Rule/Rule-List)
% apply substitution rules to In-Expression and return the result in Out-Expression; 
% each part of the expression receives at most a single replacement, and each rule is used at most once
% rules may contain blanks, named sub-expressions and conditional expressions
%
% In:
%   In-Expression : some pure expression in which substitution shall take place
%   Rule-List     : cell array of substitution rules
%
% Out:
%   Out-Expression     : In-Expression with substitutions performed
%   Changed            : whether a replacement has taken place
%   Remaining-Rules    : the list of rules that have not been applied
%
% Notes:
%   Impure expressions degrade into pure expressions, when substitutions are performed on them (i.e. they lose their attached value).
%   Named patterns inside the right-hand-sides of substitution rules are treated as if just their symbols were given instead.
%
%                                       Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                                       2010-04-19



if isfield(x,'head') && strcmp(char(x.head),'Pattern')
	% strip off spurious pattern expressions from the expression (they are not allowed there)
    x = x.parts{1}; end

% try all rules on the full expression (which may be canonical, impure, or atomic); impurity is removed iff there is a match...
r = 1;
changed = false;
while r <= length(rules)
    rule = rules{r};
    [matched,dict] = utl_match(x,rule.parts{1},struct());
    if matched
        % found a match; replace named sub-patterns in the substitution pattern, and return that
        if ~isempty(dict)
            x = utl_replacerepeated(rule.parts{2},dict);
        else
            x = rule.parts{2};
        end
        changed = true;
        % remove the rule from the rule list
        rules = rules([1:r-1 r+1:end]);
        return;
    else
        % step to the next rule
        r = r+1;
    end
end

if isfield(x,{'head','parts'})
    % the expression has substructure: try to match every subpart
    [x.head,headchanged,rules] = utl_replaceall(x.head,rules);
    partschanged = false(1,length(x.parts));
    for p=1:length(partschanged)
        [x.parts{p},partschanged(p),rules] = utl_replaceall(x.parts{p},rules); end
    changed = changed || headchanged || any(partschanged);
end
