  function [eeg,state] = flt_ica(eeg, state, variant, varargin)
% [Result-Set,State] = flt_ica(Input-Set, State, Variant, Options...)
% Annotate the Input-Set with a spatial decomposition into independent components (using Independent Component Analysis [1], or ICA)
%
% The IC-decomposed signal can be considered to have "better" channels than the raw sensor signal, for several reasons:
%  * The components are less correlated than the raw channels, which allows for easier-to-handle statistical assumptions in higher levels
%    (e.g. using diagonal covariance matrices).
%  * The relevant information can be assumed to be localized in only a small subset of the components, where the majority of components will carry less relevant data;
%    this allows for the assumption of (channel-wise) sparsity in the derived features (e.g. using l1-regularized classifiers and/or feature extractors).
%  * The components have a higher degree of semantic meaning than channels (which needs to be computed, though), such as the presence of eye artifacts, muscle artifacts,
%    brain activity, etc.
%  * A fraction of components can be localized in the brain using dipole fitting, beamforming, sparse bayesian learning, and other methods, which gives access to
%    semantics that can be derived/computed from the component's location in the brain (e.g., postcentral gyrus -> high chance of touch-related brain activity, etc.).
%
% In:
%   Input-Set  : a continuous dataset
%
%   State      : state, for online updates (default: [])
%
%   Variant    : type of ICA model to run, possible values are:
%                'amica' : AMICA [2] is currently the best-of-breed ICA method and generally preferred. (default)
%                          It uses a flexible model of source signal densities (generalized gaussian scale mixtures) which
%                          allow it to obtain better solutions for EEG data (in terms of statistical independence) than the other available methods;
%                          furthermore, multiple models can be learned, which allow it to capture non-stationarities in a principled manner.
%                          If available, a binary implementation is used. Options: help runamica
%                'infomax' : (extended) Infomax [3] is the second-best approach to ICA for EEG data, and may be faster than the MATLAB implementation
%                            of AMICA, or possibly easier to handle. If available, a binary implementation is used. Options: help runica
%                'fastica' : FastICA [4] is a relatively simple ICA implementation, which is converges relatively quickly, though at the expense of
%                            solution quality. In many cases it can not attain results as good as extended Infomax or AMICA, but for repeated computations
%                            (e.g. inside a cross-validation) it can be a reasonable choice due to its speed. Options: help fastica
%                'kernelica' : KernelICA [5] has a similarly flexible model of source densities as AMICA, using a kernel approach,
%                              but requires massive computation time, so KernelICA can realistically at best be used on small datasets. Options: help kernel_ica_options
%
%                others, if in path: 'jader','jadeop','jade_td_p','MatlabshibbsR','tica','erica','simbec','unica','amuse','fobi','evd','evd24',
%                                    'sons','sobi','ng_ol','acsobiro','acrsobibpf','pearson_ica','egld_ica','eeA','tfbss','icaML','icaMS'.
%
%   Options... : optional name-value pairs controlling the preprocessing, decomposition and postprocessing (specific to the chosen variant)
%
%                'cleaning_level': degree of data cleaning, can be 'default', 'strong', 'hardcore', 'superhardcore' (default: []='default')
%                                  sets the default values for the below clean_*** options (one can then deviate from the defaults set by the 
%                                  cleaning_level, by specifying the parameter of interest with a custom value)
%
%                All of the following rejection options can be disabled by setting them to [].
%
%                'clean_drifts'  : reject drifts from the model-source data using a non-causal FIR filter specified as {Freqs, Mode} as in flt_fir. 
%                                  (default: {[0.5 2],'hp'})
%                'clean_windows' : reject windows with abnormal (extreme) activity from the model-source data using a ranked window power statistic,
%                                  specified as in flt_clean_windows (default: {0.15}); multiple passes can be issued by specifying a cell array of options cell arrays.
%                'clean_channels': reject broken channels from the model-source data using a correlation measure, specified as in flt_clean_channels
%                                  (default: {0.6})
%                'clean_peaks'   : reject local peaks from the model-source data using a subspace projection method, specified as in flt_clean_peaks
%                                  (default: [])
%                'clean_shaping' : dampen strong high-frequency activity by applying a low-pass FIR filter, specified as in flt_fir
%                                  (default: [])
%
%                'transform'   : 0 if the decomposition should be added to the data as annotation only (default)
%                                1 if the data itself should also be directly transformed
%
%                special AMICA options / defaults (for more, see runamica.m):
%
%                'qsub'          : schedule AMICA to use multiple cluster nodes (via the Sun Grid Engine); 'on'/'off' (default: 'off')
%
%                'poll_interval' : if running on the cluster, polling interval used to monitor AMICA's progress, in seconds (default: 10)
%
%                'max_wait_time' : if running on the cluster, maximum waiting time until (if no progress) it is assumed that the AMICA 
%                                  job could not be launched, in seconds (default: 60)
%
%                'numprocs'      : if running on the cluster, number of processors to use (default: 20)
%
%                'reduce_factor' : if running on the cluster, factor by which the number of processors is reduced when the max_wait_time is exceeded
%                                  (default: 0.75)
%
%                Other parameters are available depending on the chosen variant.
%
%
% Out:
%   Result-Set : continuous dataset annotated with an ICA decomposition, and optionally with data transformed into IC activations
%
% References:
%   [1] Makeig S., Bell A.J., Jung T-P. and Sejnowski T.J. 1995. "Independent Component Analysis of Electroencephalographic Data"
%       in Mozer M. et al (eds) Advances in Neural Information Processing Systems 8, MIT press
%   [2] J. A. Palmer, S. Makeig, K. Kreutz-Delgado, and B. D. Rao, "Newton Method for the ICA Mixture Model".
%       In Proceedings of the 33rd IEEE International Conference on Acoustics and Signal Processing (ICASSP 2008), Las Vegas, NV, pp. 1805-1808, 2008.
%   [3] Bell, A. J., and Sejnowski, T. J. "An information-maximization approach to blind separation and blind deconvolution."
%       Neural Comput. 7, 6 (1995), 1129?1159.
%   [4] A. Hyvaerinen. "Fast and Robust Fixed-Point Algorithms for Independent Component Analysis."
%       IEEE Transactions on Neural Networks 10(3):626-634, 1999.
%   [5] Francis R. Bach, Michael I. Jordan. "Kernel Independent Component Analysis",
%       Journal of Machine Learning Research, 3, 1-48, 2002
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-04-17

global bcilab;

varargin = hlp_sort_namevalues(varargin{:});
if ~exp_begindef('filter') return; end
if ~exist('variant','var') || isempty(variant)
    variant = 'amica'; end
if ~exist('state','var')
    state = []; end

% determine the cleaning defaults first, depending on the cleaning level
clean_opts = hlp_varargin2struct(varargin,'cleaning_level','default');
switch clean_opts.cleaning_level
    case 'off'
        clean_opts = hlp_varargin2struct({},'clean_drifts',[], 'clean_windows',[], 'clean_channels',[], 'clean_peaks',[], 'clean_shaping',[]);
    case 'strong'
        clean_opts = hlp_varargin2struct({},'clean_drifts',{[0.5 2],'hp'}, 'clean_windows',{0.2}, 'clean_channels',{0.6}, 'clean_peaks',[], 'clean_shaping',[],'do_reject',1);
    case 'hardcore'
        clean_opts = hlp_varargin2struct({},'clean_drifts',{[0.5 2],'hp'}, 'clean_windows',{0.275}, 'clean_channels',{0.625}, 'clean_peaks',[], 'clean_shaping',[],'do_reject',1,'numrej',5,'rejsig',2.75,'rejstart',10,'rejint',10);
    case 'superhardcore'
        clean_opts = hlp_varargin2struct({},'clean_drifts',{[0.5 2],'hp'}, 'clean_windows',{0.275}, 'clean_channels',{0.65}, 'clean_peaks',[], 'clean_shaping',[],'do_reject',1,'numrej',10,'rejsig',2.5,'rejstart',15,'rejint',15);
    case 'ultrahardcore'
        clean_opts = hlp_varargin2struct({},'clean_drifts',{[0.5 2],'hp'}, 'clean_windows',{0.275}, 'clean_channels',{0.675}, 'clean_peaks',[], 'clean_shaping',[],'do_reject',1,'numrej',10,'rejsig',2,'rejstart',15,'rejint',15, 'num_models',3);
    otherwise
        if isempty(clean_opts.cleaning_level) || strcmp(clean_opts.cleaning_level,'default')
            clean_opts = hlp_varargin2struct({},'clean_drifts',{[0.5 2],'hp'}, 'clean_windows',{0.15}, 'clean_channels',{0.6}, 'clean_peaks',[], 'clean_shaping',[]);
        else
            error('unrecognized cleaning_evel option');
        end
end

% read options
opts = hlp_varargin2struct(varargin, ...
    clean_opts, ...
    'transform',0, ...
    'cleaned_data',0, ...
    'max_iter',500, ...
    'resume',1);

if ~isempty(state)
    % online case: annotate the data
    eeg.icasphere = state.icasphere;
    eeg.icaweights = state.icaweights;
    eeg.icachansind = state.icachansind;
    eeg.icawinv = state.icawinv;
else
    % offline case: first pre-process the data (using a sequence of non-causal data reductions)
    
    % create an argument list that we hand down to the ICAs (except AMICA, which has its own custom argument list)
    arglist = hlp_struct2varargin(opts,'suppress',{'clean_drifts','clean_windows','clean_channels',...
        'clean_peaks','clean_shaping','transform','max_iter','cleaned_data','resume','cleaning_level', ...
        'do_reject','numrej','rejsig','rejstart','rejint','num_models'});
        
    % remove signal mean
    pre = flt_rmbase(eeg);

    % remove drifts using an FIR filter
    if ~isempty(opts.clean_drifts)
        pre = flt_fir(pre,[],opts.clean_drifts{1:2},'zero-phase'); end
        
    % remove extreme data periods using a signal power measure
    if ~isempty(opts.clean_windows)
        if ~iscell(opts.clean_windows{1})
            opts.clean_windows = {opts.clean_windows}; end
        % we support doing multiple passes of window cleaning (e.g. short extreme spikes first, followed by longer streteches of artifacts)
        for w = opts.clean_windows
            pre = flt_clean_windows(pre,w{1}{:}); end
    end
    
    % remove bad channels using a correlation measure
    if ~isempty(opts.clean_channels)
        pre = flt_clean_channels(pre,opts.clean_channels{:}); end
    
    % remove local peaks using windowed PCA
    if ~isempty(opts.clean_peaks)
        pre = flt_clean_peaks(pre,opts.clean_peaks{:}); end

    % run a signal-shaping final FIR filter
    if ~isempty(opts.clean_shaping)
        pre = flt_fir(pre,[],opts.clean_shaping{1:2},'zero-phase'); end

    % evaluate pre
    pre = exp_eval(pre);
    
    % run ICA on preprocessed data
    switch variant
        case 'amica'
            try
                addpath /data/common/amica
                
                % override some of the AMICA defaults...
                opts = hlp_varargin2struct(opts,'qsub','off', 'poll_interval',10, 'max_wait_time',60, 'numprocs',20, 'reduce_factor',0.75, ...
                    'outdir', [bcilab.paths.temp_path filesep 'amicaout-' num2str(hlp_fingerprint({pre,opts}))]);
                
                % create an AMICA-specific argument list (specifically, some of the default rejection options are retained in it
                arglist = hlp_struct2varargin(opts,'suppress',{'clean_drifts','clean_windows','clean_channels',...
                    'clean_peaks','clean_shaping','transform','cleaned_data','resume','cleaning_level', ...
                    'poll_interval','max_wait_time','reduce_factor','numprocs'});
                
                disp(['AMICA output is in ' opts.outdir]);
                % try to delete all files in this directory
                if ~opts.resume
                    try
                        files = dir(opts.outdir);
                        for f=1:length(files)
                            if ~isdir(files{f}.name)
                                delete(files{f}.name); end
                        end
                    catch, end
                end
                % run it, but capture the qsub id
                [conout,r] = evalc('runamica(pre.data,[],size(pre.data,1),size(pre.data,2), arglist{:},''numprocs'',opts.numprocs)');
                starttime = tic;
                if strcmp(opts.qsub,'on')
                    % running on the cluster... need to check for termination (by polling) 
                    maxiters = 0;
                    while maxiters < opts.max_iter
                        pause(opts.poll_interval);
                        % the # of completed iters are in the file out.txt; parse it!
                        try
                            t = fopen([opts.outdir filesep 'out.txt']);
                            while 1
                                line = fgetl(t);
                                if ~ischar(line)
                                    break; end
                                try
                                    strs = explode(line,' ');
                                    if strcmp(strs{1},'iter')
                                        maxiters = str2num(strs{2}); end
                                catch,end
                            end
                            fclose(t);
                        catch
                            try
                                fclose(t);
                            catch,end
                        end
                        % sometimes the # of requested jobs exceeds the cluster's free resources;
                        % in this case, we reduce that 
                        if maxiters == 0 && toc(starttime) > opts.max_wait_time
                            disp('Apparently, the AMICA jobs couldn''t be scheduled yet; restarting with ');
                            opts.numprocs = floor(opts.numprocs*opts.reduce_factor);
                            disp(['N = ' num2str(opts.numprocs)]);
                            try
                                % take apart the console output of AMICA to find the qsub id....
                                lines = explode(conout,13);
                                parts = explode(lines{3},'=');
                                id = str2num(parts{2});
                                if isempty(strfind(parts{1},'qsub id'))
                                    error('didn''t find the qsub id...'); end
                                res = system(['ssh juggling qdel ' num2str(id)]);
                                if res ~= 0
                                    error('couldn''t qdel...'); end
                            catch
                                disp('couldn''t delete the previous jobs; please delete them manually...');
                            end
                            if opts.numprocs <= 4
                                disp('Could not acquire cluster resources; falling back to local computation.');
                                % try it locally...
                                r = runamica(pre.data,[],size(pre.data,1),size(pre.data,2), arglist{:},'numprocs',opts.numprocs,'qsub','off');
                                % ... and leave the while loop with a valid r
                                break;
                            else
                                [conout,r] = evalc('runamica(pre.data,[],size(pre.data,1),size(pre.data,2), arglist{:},''numprocs'',opts.numprocs)');
                                starttime = tic;
                            end
                        end
                    end
                    % load the results
                    if ~isfield(r,'W')
                        r = loadmodout(opts.outdir); end
                end
                if isempty(r)
                    error; end
            catch
                
                % run the pure MATLAB variant
                disp('note: runamica failed to run, falling back to MATLAB variant; available parameters: ');
                disp('      num_models, num_mix_comps, max_iter, update_alpha, update_mu, update_sbeta, minlrate, do_newton');
                ao = hlp_varargin2struct(arglist,'num_models',1,'num_mix_comps',3,'max_iter',opts.max_iter,'update_alpha',1,'update_mu',1,'update_sbeta',1,'minlrate',1e-8,'do_newton',1);
                [r.W r.A r.c r.LL r.LLt r.mod_prob] = amica(pre.data, ao.num_models, ao.num_mix_comps, ao.max_iter, ao.update_alpha|ao.update_mu|ao.update_sbeta, ao.minlrate, 50, ao.do_newton, 1);
                % recompute some missing parameters
                r.S = eye(size(pre.data,1));
                r.v = zeros(size(r.LLt));
                for m = 1:size(r.v,1)
                    r.v(m,:) = 1./sum(exp(bsxfun(@minus,r.LLt,r.LLt(m,:))),1); end
            end
            
            % translate properties from r (the result) to pre (the dataset)
            pre.icasphere  = r.S;                
            if size(r.W,3) > 1
                % we have multiple models...
                % take the most likely model as weights and sphere
                [x,bestmod] = max(r.mod_prob); %#ok<ASGLU>
                pre.icaweights = r.W(:,:,bestmod);
                % but retain all the rest in .bcilab.amica
                if exist('mask','var')
                    r.sample_mask = mask; end
                eeg.bcilab.amica = r;
            else
                pre.icaweights = r.W;
            end
            
        case 'infomax'
            % by default, we use extended Infomax
            pre = pop_runica(pre,'extended',3,arglist{:});
        case 'kernelica'
            % precondition using FastICA
            [dummy,unmixing] = fastica(pre.data,arglist{:}); %#ok<ASGLU>
            % fine-tune using KernelICA
            pre.icaweights = kernel_ica_options(pre.data,'W0',unmixing,'contrast','kgv',arglist{:});
            pre.icasphere = eye(size(pre.data,1));
        otherwise
            if isequal(1,strmatch('brainica',variant))
                % we have one of the brainica variants
                if isfield(pre.etc,'baseline')
                    arglist = [arglist {'basewnd', pre.etc.baseline}]; end
                [pre.icaweights,pre.icasphere] = feval(variant,pre.data,pre.srate,'extended',3,'chanlocs',pre.chanlocs,arglist{:});
            else
                % let pop_runica handle all the rest
                pre = pop_runica(pre,'icatype',variant,arglist{:}); 
            end
    end
    % add channel indices, if necessary
    if isempty(pre.icachansind)
        pre.icachansind = 1:size(pre.data,1); end    
    % add weight inverse (mixing matrix), if necessary
    if isempty(pre.icawinv)
        pre.icawinv = pinv(pre.icaweights*pre.icasphere); end;
    if opts.cleaned_data
        eeg = pre;
    else
        % carry over the fields to the output dataset
        eeg.icaweights = pre.icaweights;
        eeg.icasphere = pre.icasphere;
        eeg.icawinv = pre.icawinv;
        eeg.icachansind = pre.icachansind;
        if ~isempty(opts.clean_channels)
            % if channel rejection was used, we remember from which channels the ica map was derived
            [dummy,idxeeg,idxpre] = intersect({eeg.chanlocs.labels},{pre.chanlocs.labels}); %#ok<ASGLU>
            [dummy,idxpre_order] = sort(idxpre);  %#ok<ASGLU>
            eeg.icachansind = idxeeg(idxpre_order);
        end
    end
end

% retain the ICA decomposition in the state
state = struct('icasphere',{eeg.icasphere}, 'icaweights',{eeg.icaweights}, 'icawinv',{eeg.icawinv}, 'icachansind',{eeg.icachansind});

% transform the data, if necessary
if opts.transform
    eeg.data = (eeg.icaweights*eeg.icasphere)*eeg.data(eeg.icachansind,:);     
    eeg.chanlocs = struct('labels',cellfun(@num2str,num2cell(1:eeg.nbchan,1),'UniformOutput',false));
    eeg.etc.implicit_ica = state;
    eeg.icaweights = [];
    eeg.icasphere = [];
    eeg.icawinv = [];
    eeg.icachansind = [];
end

exp_enddef;