function S = champagne(B,L,varargin)
% Source-Activation = champagne(Measured-Signal,Lead-Field-Matrix)
% Obtain a sparse reconstruction of the source activations (according to the LFM) that explain the measured signal, according to the 
% Champagne [1] algorithm.
%
% In:
%   Measured-Signal   : A [#Channels x #Samples] matrix of observed data (e.g., a time segment of EEG/MEG).
%                       Note: #Samples should ideally be > 3x #Channels.
%
%   Lead-Field-Matrix : A [#Channels x #Orientation-Axes x #Source-Locations] matrix which describes the forward projection
%                       from each possible source, along each of the allowed orientation axes, onto the channels.
%                       The #Orientation-Axes can be 1 if the orientations are known, or 2 for two degrees of freedom (free planar rotations)
%                       or 3 for full 3 degrees of freedom. 
%                       Note: it is not clear to me if it makes mathematical sense to orient these in a small cone, to encode a certain
%                       knowledge about the most likely directions.
%
%   Options           : optional name-value pairs
%   
% Out:
%   Source-Activation : A [#Orientation-Axes x #Time-Points x #Source-Locations] matrix, sparse along the Source-Locations dimension,
%                       of time courses (and, implicitly, orientations) of the active sources.
%
% References:
%   [1] David P. Wipf et al., "Robust Bayesian Estimation of the location, orientation, and time course of multiple correlated neural sources using MEG"
%       NeuroImage 49 641-655, 2010
%

% read options
o = hlp_varargin2struct(varargin,'maxiter',100);

[d_b,d_t] = size(B);        % #Channels, #Samples
[d_b,d_c,d_s] = size(L);    % #Channels, #Orientation-Axes, #Sources

% reduce the signal to one that yields the same covariance
B_rank = rank(B);
B_tilde = [B zeros(d_b,B_rank - mod(B_rank,d_t))];
B_tilde = reshape(B_tilde,[1 1 

% TODO: estimate noise covariance Sigma_epsilon
Sig_eps = 0;

% initialize the Gamma's (we use a 3d matrix here)
Gamma = rempat(diag(ones(1,d_c)),[1 1 d_s]);

% TODO: preallocate X
% X = 


% for each iteration...
for iter = 1:o.maxiter
    % collect the Gamma's into a sparse block-diagonalized form
    for i=1:d_s
        tmp{i} = Gamma(:,:,i); end    
    Gamma_blk = sparse(blkdiag(tmp{i}));
    
    % compute Sigma_b
    Sig_b = L*Gamma_blk*L';
    inv_Sig_b = pinv(Sig_b);
    % update X
    for i=1:d_c
        X(:,:,i) = Gamma(:,:,i)*L(:,:,i)'*inv_Sig_b
    end
end