function res = test_errors_classify(varargin)
% quick test for error classification
%
% ad hoc with default channel subset
%  res = test_errors('day',5, 'channels','sel', 'compare',{{'odderr.corr','err.corr','odd.regular','odd.odd','err.err'},'odderr.err'})
% with ICA rejection matrix
%  res = test_errors('day',5, 'projection',my_icamatrix, 'compare',{{'odderr.corr','err.corr','odd.regular','odd.odd','err.err'},'odderr.err'})
% with r-square windows:
%  res = test_errors('wnds',[0.18 0.3; 0.26 0.34; 0.34 0.38; 0.375 0.485], 'channels','sel', 'compare',{{'odderr.corr','err.corr','odd.regular','odd.odd','err.err'},'odderr.err'})
%
% ... to plot classifier weights for each time window: vis_ldaweights(res.subj(2).cla);
%
%
%%%%%  res = test_errors_classify('day',6, 'channels','sel', 'compare',{{'odderr.corr','odd.oddrlr','err.rooterr'},'odderr.err'}, 'classifier', cla1)


global res;
clear res;

% --- read the options & apply defaults ---
opts = hlp_varargin2struct(varargin, ...
    ... % -- data selection
    'day', 5, ...                                % the day of the study to be analyzed; also sets the defaults for several subsequent params
    'subjects', [], ...                          % the subject selection (default: all subjects for that day)
    'compare', {'odderr.corr','odderr.err'}, ... % the data sets which should be compared...
    'projection', [], ...                        % an optional projection transform on the data (e.g. the ICA component rejection matrix);
    ...                                          % can also be a cell array of such matrices, one per subject
    ... % -- feature params
    'wnds', [0.25 0.3;0.3 0.35;0.35 0.4; 0.4 0.45;0.45 0.5;0.5 0.55;0.55 0.6], ... % the window segments
    'classifier', 'quick', ...                   % the classifier model
    'filter', {'fft-nownd', [0.1 15]}, ...       % the preproc filter setup
    'window', [0 0.8], ...                       % the time window for classification
    'channels', [], ...                          % channel subset selection (default: all channels for that day)
    'samplerate', [], ...                        % sampling rate (default: depending on day)
    'boostdata', 0, ...                          % whether the data amounts should be boosted to yield equal class priors
    ...                                          % if this is a cell array, data amounts are boosted as if a cost structure was specified
    'universal', [] ...                          % if 1, give (pseudo-) universal classification results
    );


% --- define some constants ---
% subjects, by day
subjects_d{4} = 1:17;
subjects_d{5} = 1:8;
subjects_d{6} = 1:1;
% data set types and associated file names, by day
data_files_d{4} = struct('noerr','Block2-Train-NoError-Kurz','odderr','Block3-Game-Error-NoBCI','bcii','Block4-Game-BCI-Indiv','bciu','Block5-Game-BCI-Univ');
data_files_d{5} = struct('noerr','d5-NoError','odderr','d5-Error','odd','d5-Oddball','err','d5-Root');
data_files_d{6} = struct('noerr','Teil3','odderr','Teil3','odd','Teil3','err','Teil3');

% channel set, by day
channels_d{4} = struct('all',1:32, 'sel', {{'F3' 'FC2' 'FC4' 'FC6' 'FT8' 'Cz' 'C4' 'CP5' 'CP3' 'CP1' 'CPz' 'CP4' 'CP6' 'P7' 'Pz' 'P4' 'PO3' 'POz' 'PO4' 'C3'}});
channels_d{5} = struct('all',1:96, 'sel', {{'F3' 'FC2' 'FC4' 'FC6' 'FT8' 'Cz' 'C4' 'CP5' 'CP3' 'CP1' 'CPz' 'CP4' 'CP6' 'P7' 'Pz' 'P4' 'PO3' 'POz' 'PO4' 'C3'}});
channels_d{6} = struct('all',1:32, 'sel',1:32);
% sample rate, by day
srate_d = {200,200,200,200,200,500};

% marker sets, by event category
markers = struct('corr', {{'S 11' 'S 17' 'S 18'}}, 'err', {{'S 12','S 13'}}, 'regular', 'S  1', 'odd', 'S  2','oddrlr',{{'S 27','S 28'}}, 'rooterr',{{'S 29','S 30'}});
% predefined classifiers
classifiers = struct('quick', {{{'lda' 0.015}}}, 'good', {{'logreg'}}, 'slow', {hlp_search('qda-full',15)}, 'qda', {{'qda'}});


% --- assign missing options ---
if isempty(opts.subjects)
    opts.subjects = subjects_d{opts.day}; end
if ischar(opts.channels) && strcmp(opts.channels,'all')
    opts.channels = channels_d{opts.day}.all; end
if ischar(opts.channels) && strcmp(opts.channels,'sel')
    opts.channels = channels_d{opts.day}.sel; end
if isempty(opts.channels)
    opts.channels = channels_d{opts.day}.all; end
if isempty(opts.samplerate)
    opts.samplerate = srate_d{opts.day}; end
if ischar(opts.classifier)
    opts.classifier = classifiers.(opts.classifier); end
if isnumeric(opts.projection) || (iscell(opts.projection) && ~isempty(opts.projection{1}) && isnumeric(opts.projection{1}))
    opts.projection = repmat({opts.projection}, [1,max(opts.subjects)]); end

% prepare for an intense computation
%env_cachemode('eager');
warning off MATLAB:nearlySingularMatrix
warning off MATLAB:singularMatrix
warning off MATLAB:log:logOfZero

for s = opts.subjects
    try
        % get the available data file names for the chosen day
        files = data_files_d{opts.day};

        % interpret opts.compare and load the appropriate sets for each class
        disp(['subject ' num2str(s) ': preprocessing...']);
        setidx = 1;
        for i=1:2
            data_selections{i} = {};
            if ~iscell(opts.compare{i})
                opts.compare{i} = {opts.compare{i}}; end

            % since there may be multiple data selections to be grouped into the same class, we have to iteratively merge them
            for k = opts.compare{i}
                [setname,mrkname] = strtok(k{1},'.'); mrkname = mrkname(2:end);
                raw = io_quickload(['data:/Projekte/Nature-Error/Datasets/Day' num2str(opts.day) '/' utl_vpstring(s) '/' files.(setname)]);
                preproc = bci_preproc(raw,@para_simple_train, 'srate',opts.samplerate, 'events',{markers.(mrkname)}, 'window',opts.window, 'filter',opts.filter, 'channels',opts.channels, 'project', opts.projection{s});
                preproc = set_pickepos(preproc, 1:length(preproc.classidx)*0.45);
                % quick hack: remember the set index per trial, using the classidx
                preproc.classidx = ones(1,length(preproc.classidx))*setidx; setidx = setidx+1;
                % another hack: align the reflatencies of the data sets, so the subsequent chrono CV cannot produce artifacts
                preproc.reflatencies = preproc.reflatencies - min(preproc.reflatencies);
                preproc.reflatencies = preproc.reflatencies/max(preproc.reflatencies);
                % merge into selection
                data_selections{i} = set_merge(data_selections{i},preproc);
                clear raw preproc
            end

        end

        if (opts.boostdata)
            if iscell(opts.boostdata)
                % boost the data to reflect a cost structure
                for c=1:2
                    data_selections{c} = set_pickepos(data_selections{c}, floor(1:1/opts.boostdata{c}:length(data_selections{c}.classidx))); end
            else
                % boost the data to equalize class priors
                classratio = length(data_selections{1}.classidx)/length(data_selections{2}.classidx);
                if (classratio < 0.7) || ((1/classratio) < 0.7)
                    if classratio < 1 boost_class = 1; else boost_class = 2; end
                    data_selections{boost_class} = set_pickepos(data_selections{boost_class}, floor(1:min(1/classratio,classratio):length(data_selections{boost_class}.classidx)));
                end
            end
        end

        % combine the classes into a single set
        [data{s},I] = set_sortepos(set_combine(data_selections{1}, data_selections{2}));
        % and remember the set index per trial, in the final data set (so we can later compare it with the crossval results)
        setidx = [data_selections{1}.classidx data_selections{2}.classidx];
        setidx = setidx(I);
        clear data_selections

        if isempty(opts.universal)

            % classify
            if isstruct(opts.classifier)
                [subj(s).err,subj(s).pred,subj(s).StdDev,subj(s).fprate,subj(s).fnrate,subj(s).stats] = bci_classify(opts.classifier, data{s})

                % also calc the error per set index
                trialerr = [subj(s).stats.errors];
            else
                disp(['subject ' num2str(s) ': training...']);
                % train a classifier
                trainopts = {'trainfun',@para_simple_train, 'trainset',data{s}, 'nestedfoldness',{'chron' 5 5}};
                [subj(s).err,subj(s).cla,st,subj(s).stats] = bci_train(trainopts, [], {opts.wnds}, opts.classifier)

                % also calc the error per set index
                trialerr = [subj(s).stats.folderr{:}];
            end

            for settype = unique(setidx)
                % count the number of errors for that set
                subj(s).setsize(settype) = sum(setidx==settype)/length(setidx);
                subj(s).seterr(settype) = mean(trialerr(setidx==settype));
            end

            % store statistics
            try res.subj(s) = subj(s); disp(subj(s).err); disp(subj(s).seterr); catch,end
            clear data
        end
    catch
        disp(['error at subject ' num2str(s)]);
        env_handleerror(lasterror);
    end
end

if opts.universal
    disp(['universal classifier: training...']);
    % train a universal classifier
    trainopts = {'trainfun',@para_simple_train, 'trainset',set_sortepos(set_merge(data{:})), 'nestedfoldness',{'chron' 10 5}};
    [err,cla,st,stats] = bci_train(trainopts, [], {opts.wnds}, opts.classifier);
    % and store statistics
    res.univ = struct('err',err,'cla',cla,'stats',stats);
end