  function eeg = flt_clean(eeg, varargin)
% Clean the dataset by using either one of the custom settings, or by overwriting some or all the settings in one
% of the custom settings. This function makes use of other cleaning functions, to which arguments can be sent. For example,
% flt_clean_windows removes all metadata by default, e.g. events and ICA decompositions. To avoid this, 
% ''clean_windows', {[],[],[],true}' must be given in the arguments to flt_clean.
% In:
%   eeg  : a continuous dataset
%   Options... : optional name-value pairs controlling the preprocessing, decomposition and postprocessing (specific to the chosen variant)
%
%                'cleaning_level': degree of data cleaning, can be 'default', 'strong', 'hardcore', 'superhardcore' (default: []='default')
%                                  sets the default values for the below clean_*** options (one can then deviate from the defaults set by the 
%                                  cleaning_level, by specifying the parameter of interest with a custom value)
%
%                All of the following rejection options can be disabled by setting them to [].
%
%                'clean_drifts'  : reject drifts from the model-source data using a non-causal FIR filter specified as {Freqs, Mode} as in flt_fir. 
%                                  (default: {[0.5 2],'hp'})
%                'clean_windows' : reject windows with abnormal (extreme) activity from the model-source data using a ranked window power statistic,
%                                  specified as in flt_clean_windows (default: {0.15}); multiple passes can be issued by specifying a cell array of options cell arrays.
%                'clean_channels': reject broken channels from the model-source data using a correlation measure, specified as in flt_clean_channels
%                                  (default: {0.6})
%                'clean_peaks'   : reject local peaks from the model-source data using a subspace projection method, specified as in flt_clean_peaks
%                                  (default: [])
%                'clean_shaping' : dampen strong high-frequency activity by applying a low-pass FIR filter, specified as in flt_fir
%                                  (default: [])
% Out:
%   Result-Set : continuous cleaned dataset.
%


% determine the cleaning defaults first, depending on the cleaning level
clean_opts = hlp_varargin2struct(varargin,'cleaning_level','default');
switch clean_opts.cleaning_level
    case 'off'
        clean_opts = hlp_varargin2struct({},'clean_drifts',[], 'clean_windows',[], 'clean_channels',[], 'clean_peaks',[], 'clean_shaping',[]);
    case 'strong'
        clean_opts = hlp_varargin2struct({},'clean_drifts',{[0.5 2],'hp'}, 'clean_windows',{0.2}, 'clean_channels',{0.6}, 'clean_peaks',[], 'clean_shaping',[],'do_reject',1);
    case 'hardcore'
        clean_opts = hlp_varargin2struct({},'clean_drifts',{[0.5 2],'hp'}, 'clean_windows',{0.275}, 'clean_channels',{0.625}, 'clean_peaks',[], 'clean_shaping',[],'do_reject',1,'numrej',5,'rejsig',2.75,'rejstart',10,'rejint',10);
    case 'superhardcore'
        clean_opts = hlp_varargin2struct({},'clean_drifts',{[0.5 2],'hp'}, 'clean_windows',{0.275}, 'clean_channels',{0.65}, 'clean_peaks',[], 'clean_shaping',[],'do_reject',1,'numrej',10,'rejsig',2.5,'rejstart',15,'rejint',15);
    case 'ultrahardcore'
        clean_opts = hlp_varargin2struct({},'clean_drifts',{[0.5 2],'hp'}, 'clean_windows',{0.275}, 'clean_channels',{0.675}, 'clean_peaks',[], 'clean_shaping',[],'do_reject',1,'numrej',10,'rejsig',2,'rejstart',15,'rejint',15, 'num_models',3);
    otherwise
        if isempty(clean_opts.cleaning_level) || strcmp(clean_opts.cleaning_level,'default')
            clean_opts = hlp_varargin2struct({},'clean_drifts',{[0.5 2],'hp'}, 'clean_windows',{0.15}, 'clean_channels',{0.6}, 'clean_peaks',[], 'clean_shaping',[]);
        else
            error('unrecognized cleaning_level option');
        end
end


% read options
opts = hlp_varargin2struct(varargin, ...
    clean_opts, ...
    'transform',0, ...
    'cleaned_data',0, ...
    'max_iter',500, ...
    'resume',1);

        
    % remove signal mean
    pre = flt_rmbase(eeg);

    % remove drifts using an FIR filter
    if ~isempty(opts.clean_drifts)
        pre = flt_fir(pre,[],opts.clean_drifts{1:2},'zero-phase'); end
        
    % remove extreme data periods using a signal power measure
    if ~isempty(opts.clean_windows)
        if ~iscell(opts.clean_windows{1})
            opts.clean_windows = {opts.clean_windows}; end
        % we support doing multiple passes of window cleaning (e.g. short extreme spikes first, followed by longer streteches of artifacts)
        for w = opts.clean_windows
            pre = flt_clean_windows(pre,w{1}{:}); end
    end
    
    % remove bad channels using a correlation measure
    if ~isempty(opts.clean_channels)
        pre = flt_clean_channels(pre,opts.clean_channels{:}); end
    
    % remove local peaks using windowed PCA
    if ~isempty(opts.clean_peaks)
        pre = flt_clean_peaks(pre,opts.clean_peaks{:}); end

    % run a signal-shaping final FIR filter
    if ~isempty(opts.clean_shaping)
        pre = flt_fir(pre,[],opts.clean_shaping{1:2},'zero-phase'); end

    % evaluate pre
    eeg = exp_eval(pre);