function model = ml_trainrvm(trials,targets,varargin)
% Model = ml_trainrvm(Trials, Targets, Options...)
% Learn a probabilistic inear or non-linear predictive model, for classification or regression, using the Relevance Vector Machine.
%
% The Relevance Vector Machine [1] is a Bayesian equivalent to the popular Support Vector Machines (SVMs) [2]. RVMs can be used for general-purpose
% linear or non-linear (kernelized) classification or regression, and produce state-of-the-art results in most cases. Various kernels are supplied,
% where the rbf kernel is usually the best initial choice. In the non-linear case, the kernel scaling parameter should be found via parameter search,
% which can be time-consuming. In contrast to SVMs, Relevance Vector Machines give probablistic outputs, which can be practical when multiple uncertain
% predictions are to be fused, etc. In the future, an implementation of RVMs using convex optimization will be provided, which is assumed to give better
% optimality guarantees than the current implementation [3,4,5].
%
% RVMs (as well as kernel SVMs) are the most versatile general-purpose classifiers currently available in the toolbox, and are a good default choice 
% if very little is known about the data. For best results, care must be taken to search over the respective regularization parameter(s) appropriately.
% If more is known about the structure of the data (e.g. that it should be linearly separable, or that that sparsity can be exploited across features),
% specialized methods may be more appropriate (and give similar results faster or reach better performance by overfitting less strongly).
%
% In:
%   Trials       : training data, as in ml_train
%
%   Targets      : target variable, as in ml_train
%
%   Options  : optional name-value parameters to control the training details:
%              'type': classification (c) / regression (r) (default: 'c')
%
%              'kernel': one of several kernel types:
%                         * 'linear':   Linear
%                         * 'rbf':      Gaussian / radial basis functions (default)
%                         * 'laplace':  Laplacian
%                         * 'poly':		Polynomial
%                         * 'cauchy':	Cauchy
%
%              'gamma': scaling parameter of the kernel (for regularization), default: 1
%                       reasonable range: 2.^(-16:2:4)
%
%              'degree': degree of the polynomial kernel, if used (default: 3)
%
%              'bias': whether to add a bias to the data (default: 1)
%
%              misc options:
%              'iterations': Number of interations to run for
%              'time':       time limit to run for, e.g. '1.5 hours', '30 minutes', '1 second'
%              'fixednoise': whether the gaussian noise is to be fixed 0/1
%              'beta':       (Gaussian) noise precision (inverse variance)
%              'noisestd':   (Gaussian) noise standard deviation
%              'scaling':    pre-scaling of the data (see hlp_findscaling for options) (default: 'std') 
%              'diagnosticlevel':   verbosity level, 0-4
%
% Out:
%   Model   : the computed model...
%             classes indicates the class labels which the model predicts
%             additional parameters determine a posterior distribution over the weights
%
% Examples:
%   model = ml_trainrvm(trials,labels): learn a sparse Bayesian model
%   
% See also:
%   ml_predictrvm()
%
% References:
%  [1] Vladimir Vapnik. "The Nature of Statistical Learning Theory." 
%      Springer-Verlag, 1995
%  [2] Michael E. Tipping and Alex Smola, "Sparse Bayesian Learning and the Relevance Vector Machine". 
%      Journal of Machine Learning Research 1: 211?244. (2001)
%  [3] Michael E. Tipping and A. C. Faul. "Fast marginal likelihood maximisation for sparse Bayesian models."
%      In C. M. Bishop and B. J. Frey (Eds.), Proceedings of the Ninth International Workshop on Artificial Intelligence and Statistics, Key West, FL, Jan 3-6 (2003)
%  [4] David P. Wipf and Srikantan Nagarajan, "A New View of Automatic Relevance Determination,"
%      In J.C. Platt, D. Koller, Y. Singer, and S. Roweis, editors, Advances in Neural Information Processing Systems 20, MIT Press, 2008.
%  [5] David P. Wipf and Srikantan Nagarajan, "Sparse Estimation Using General Likelihoods and Non-Factorial Priors," 
%      In Advances in Neural Information Processing Systems 22, 2009.
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-06

% read inputs
opts = hlp_varargin2struct(varargin, ...
        'kernel','rbf', ...
        'gamma',1, ...
        'degree',3, ...
        'type','c', ...
        'bias',1, ...
        'iterations',10000,...
        'time','10000 seconds',...
        'fixednoise',false,...
        'beta',[],...
        'noisestd',[],...
        'scaling','std',...
        'diagnosticlevel',0);
    
if opts.type == 'c'
    opts.likelihood = 'bernoulli';
else
    opts.likelihood = 'gaussian';
end

% remap targets for classification
if strcmp(opts.type,'c')
    classes = unique(targets);
    if length(classes) > 2
        % multiclass case: use the voter
        model = ml_trainvote(trials,targets,@ml_trainrvm,@ml_predictrvm,varargin{:});
        return;
    end
    % remap target labels to 0/1
    targets(targets==classes(1)) = 0;
    targets(targets==classes(2)) = 1;
else
    classes = [];
end

% prescale the data
sc_info = hlp_findscaling(trials,opts.scaling);
trials = hlp_applyscaling(trials,sc_info);

% kernelize the data
basis = trials;
trials = utl_kernelize(trials,basis,opts.kernel,opts.gamma,opts.degree);
% add a bias
if opts.bias
    trials = [ones(size(trials,1),1) trials]; end

% run the RVM
args1 = hlp_struct2varargin(opts,'restrict',{'iterations','time','diagnosticlevel','fixednoise','freebasis'});
args2 = hlp_struct2varargin(opts,'restrict',{'beta','noisestd','relevant','weights','alpha'});
[param, hyperparam, diag] = SparseBayes(opts.likelihood,trials,targets,SB2_UserOptions(args1{:}),SB2_ParameterSettings(args2{:}));

% preselect relevant basis vectors
if ~strcmp(opts.kernel,'linear')
    if opts.bias
        feature_sel = param.Relevant-1;
        if feature_sel(1) == 0
            % bias was relevant, remove it from the basis vector selection (it will be added after kernelization)
            feature_sel = feature_sel(2:end);
        else
            % bias was not relevant, forget about it
            opts.bias = 0;
        end
    else
        feature_sel = param.Relevant;
    end
    basis = basis(feature_sel,:);
else
    feature_sel = param.Relevant;
end

model = struct('sc_info',{sc_info},'classes',{classes},'basis',{basis},'param',{param},'hyperparam',{hyperparam},...
               'diag',{diag},'type',{opts.type},'feature_sel',{feature_sel},'bias',{opts.bias}, ...
               'kernel',{opts.kernel},'gamma',{opts.gamma},'degree',{opts.degree});
