function result = para_bandpower(data, mode, varargin)
% Result = para_bandpower(Input-Data, Operation-Mode, Options...)
% Paradigm for oscillatory processes, using per-channel logarithmic bandpower (in a time/frequency window); relies on a spatial filter to be effective.
%
% The logarithmic Bandpower estimates ("log-BP") paradigm is based on the design of the original Graz Brain-Computer Interface [1], which used
% lateralized motor imagery for control. The features exploited by this paradigm in its original form are Event-Related Synchronization and 
% Desynchronization [2] localized in the motor cortex, but the paradigm is not restricted to these applications.
%
% Generally, log-BP can be used as a simple method to operate on oscillatory processes, either in relation to events, or asynchronously. 
% The paradigm is simple in that it does not capture any complex time variations in the oscillations detected, does not capture interactions
% between multiple frequency bands, and does not contain data-adaptive signal processing or feature extraction (the machine learning component is
% adaptive, though).
%
% The paradigm is implemented as a standard sequence of signal (pre-)processing (spatial/spectral filtering), feature extraction, and machine learning.
% The defining property of the paradigm is that it extracts, per trial, the per-channel log-variance log(var(X)) as features of the signal. The resulting 
% feature vectors are then passed along to the learner component. By default, the paradigm uses a non-adaptive spatial filter, the surface Laplacian, and
% a non-adaptive spectral filter. In place of these, a spatial transformation into Indepentent Components and/or adaptive spectral filtering can be used.
% Since the feature space created by it is typically relatively low-dimensional, almost any classifier in the toolbox can be used with log-BP, including
% non-linear ones such as SVMperf and QDA. Aside from these optional modifications, the parameters that almost certainly need to be tuned to the 
% task at hand are the length of the data epoch and the choice of a frequency band (defaulting to motor imagery time scales and frequency ranges),
% both of which can also be found via a small parameter search.
% 
% The most closely related (more advanced) paradigms are para_csp (Common Spatial Patterns, CSP), para_multiband_csp (Multiband-CSP),
% and para_speccsp (Spectrally weighted CSP), which add adaptively learned spatial filters in one (CSP) or more (Multiband-CSP) bands, and 
% adaptively learned spatio-spectral filters (Spec-CSP), respectively. An even more advanced paradigm is para_dal (Dual-Augmented Lagrange), 
% which in addition allows to express temporal structure (time variation) in the process of interest.
%
% Some application areas include detection of major brain rhythm modulations (e.g. alpha, beta), for example related to relaxation, 
% aspects of workload, and motor imagery. See also [3].
%
% Simple Example: Consider a calibration dataset in which a subject is in varying degrees of relaxedness throughout different time periods. Markers 'relaxed',
% 'non-relaxed' and 'stressed' have been inserted in 2 second intervals into these data ranges. The goal is to learn a model which can predict these
% states, using alpha rhythm features.
%  
%   calib = io_loadset('datasets/demos/relaxation.eeg')
%   [loss,model,stats] = bci_train({'data',calib, 'paradigm',@para_bandpower}, 'events',{'relaxed','non-relaxed','stressed'}, ...
%       'spectrum',[7 15], 'epoch',[-2 2])
%
% Advanced Example: On the same dataset, the relevant information shall be derived using a better (adaptive) spatial filter (ICA) and a better classifier 
% (logistic regression) than the defaults. The classifier shall be sparse, since we assume that only few of the independent components in the data
% will carry relevant information. A Bayesian variant of the classifier will be used (using Automatic Relevance Determination) to avoid the need
% for time-consuming regularization. As an alternative, the l1-regularized variant of the classifier could be used.
%
%   calib = io_loadset('datasets/demos/relaxation.eeg')
%   [loss,model,stats] = bci_train({'data',calib, 'paradigm',@para_bandpower}, 'events',{'relaxed','non-relaxed','stressed'}, ...
%       'spectrum',[7 15], 'epoch',[-2 2], 'ica','on', 'learner',{'logreg',[],'variant','vb-ard'})
%
%
% In:
%   Input-Set       : some data set
%
%   Operation-Mode  : 'preprocessing', to pre-process the Input-Set according to the paradigm (and parameters)
%                     'learning', to learn a predictive model from a pre-processed Input-Set (and parameters)
%                     'prediction', to make predictions given a pre-processed Input-Set and a Model
%
%   User Options... : optional parameters of the paradigm; name-value pairs 
%                     --- for preprocessing (see flt_default)
%                     * 'laplace': surface laplacian (default: 1)
%                     * 'spectrum':  spectral filter (default: [7 30])
%                     * 'srate':   sampling rate, in Hz (default: 200)
%                     * 'epoch':  epoch window, in seconds relative to event (default: [0.5 3.5])
%                     * 'events':  event types to determine epochs and the target variable (user-specified)
%                     * others, by default turned off: see flt_default
%                     --- for learning
%                     * 'learner': learning function to be used on the features; (default: 'qda')
%                     --- for prediction
%                     * 'model': the detector model (as produced during the learning)
%
% Out:
%   Result : depends on the Operation-Mode; 
%            * for 'preprocessing', this is the preprocessed dataset
%            * for 'learning', this is the learned model
%            * for 'prediction', this is the predictions produced by the model, given the data
%
% References:
%  [1] G. Pfurtscheller, C. Neuper, "Motor imagery and direct brain-computer communication"
%      Proceedings of the IEEE, Vol. 89, No. 7. (07 August 2002), pp. 1123-1134.
%  [2] Pfurtscheller, G., and da Silva, L. "Event-related EEG/MEG synchronizaion and desynchronization: basic principles." 
%      Clin Neurophysiol 110 (1999), 1842-1857.
%  [3] Buzsaki, G., "Rhythms of the brain"
%      Oxford University Press US, 2006
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-29

result = para_skeleton(data, mode, hlp_varargin2struct(varargin, ...
	... % preprocessing defaults:
    'laplace','on', 'spectrum',[7 30], 'epoch',[0.5 3.5], 'srate',200, ...
    ... % learning defaults:
    'featureextract',@(D,opts)reshape(log(var(D.data,0,2)),[size(D.data,1),size(D.data,3)])', 'learner','lda'));
